package ru.microem.reader.usb_bluetooth_test;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.navigation.NavDestination;
import ru.microem.reader.usb_bluetooth_test.ui.discover.DiscoverFragment;
import ru.microem.reader.usb_bluetooth_test.ui.reader.ReaderFragment;
import ru.microem.reader.usb_bluetooth_test.ui.tags.TagsFragment;
import ru.microem.reader.usb_bluetooth_test.ui.tcl.TclFragment;
import ru.microem.reader.usb_bluetooth_test.ui.sam.SamFragment;
import ru.microem.reader.usb_bluetooth_test.ui.mifare_classic.MifareClassicFragment;
import ru.microem.reader.usb_bluetooth_test.ui.mifare_ultralight.MifareUltralightFragment;
import ru.microem.reader.usb_bluetooth_test.ui.mifare_plus.MifarePlusFragment;
import ru.microem.reader.usb_bluetooth_test.ui.icode.ICodeFragment;
import ru.microem.readerlibrary.UemReader;

import android.app.Activity;
//import android.app.Fragment;
//import android.support.v4.app.Fragment;
import android.content.Intent;
import android.net.Uri;
import android.os.Bundle;

import com.google.android.material.floatingactionbutton.FloatingActionButton;
import com.google.android.material.snackbar.Snackbar;

import android.os.Handler;
import android.view.View;

import androidx.navigation.NavController;
import androidx.navigation.Navigation;
import androidx.navigation.ui.AppBarConfiguration;
import androidx.navigation.ui.NavigationUI;

import com.google.android.material.navigation.NavigationView;

import androidx.drawerlayout.widget.DrawerLayout;

import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.widget.Toolbar;

import android.view.Menu;
import android.widget.TextView;
import android.widget.Toast;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

public class MainActivity extends AppCompatActivity {

    private AppBarConfiguration mAppBarConfiguration;

    public DiscoverFragment mDiscoverFragment = null;
    public ReaderFragment mReaderFragment = null;
    public TagsFragment mTagsFragment = null;
    public MifareClassicFragment mMfcFragment = null;
    public MifareUltralightFragment mMifareUltralightFragment = null;
    public MifarePlusFragment mMifarePlusFragment = null;
    public ICodeFragment mICodeFragment = null;
    public TclFragment mTclFragment = null;
    public SamFragment mSamFragment = null;

    public UemReader mUemR = null;

    public byte[][] mCommands = new byte[][]{
            new byte[] {},
            new byte[] {0x05, 0x01},
            new byte[] {0x07, 0x01, 0x01, 0x00},
            new byte[] {0x64},
            new byte[] {0x05, 0x02},
            new byte[] {0x07, 0x02, 0x01, 0x00},
            new byte[] {0x10},
            new byte[] {0x51, 0x40},
            new byte[] {0x30, 0x00, 0x00, 0x00},
            new byte[] {0x04, (byte)0x80, 0x01},
            new byte[] {0x75, 0x40, 0x55, 0x00, 0x00}
    };

    public int mActiveMode = UemReader.CommandsCards.UEM_CARD_STANDARD_ISO14443A;

    public byte[] mUID = null;

    @Override
    protected void onPause() {
        mUemR.stopReceivingBroadcasts();
        /*Toast.makeText(this,
                "onPause" , Toast.LENGTH_LONG)
                .show();*/
        super.onPause();
    }

    @Override
    protected void onResume() {
        super.onResume();
        mUemR.startReceivingBroadcasts();
        /*Toast.makeText(this,
                "onResume" , Toast.LENGTH_LONG)
                .show();*/
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        Toolbar toolbar = findViewById(R.id.toolbar);
        setSupportActionBar(toolbar);

        mUemR = new UemReader(this);

        final Activity mMainActivity = this;
        mUemR.Bluetooth.setCallbacks(new UemReader.BluetoothCallbacks() {
            @Override
            public void onBluetoothReaderFound(String readerName) {
                Toast.makeText(mMainActivity,
                        "Bluetooth Reader Found "+readerName , Toast.LENGTH_LONG)
                        .show();
                mUemR.Bluetooth.stopDiscovery();
                //mUemR.connect(readerName);
                mDiscoverFragment.addReader(readerName);
            }

            @Override
            public void onBluetoothReaderConnected() {
                mUemR.Reader.beep(3);
            }

            @Override
            public void onBluetoothResponse(byte[] response) {
                if (response != null) {
                    final String bufReceived = UemReader.Helper.byteArrayToHexString(response);
                    mMainActivity.runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(getApplicationContext(),
                                    bufReceived, Toast.LENGTH_LONG)
                                    .show();
                        }
                    });
                }
                //mUemR.disconnect();
            }
        });
        mUemR.USB.setCallbacks(new UemReader.UsbCallbacks() {
            @Override
            public void onUsbReaderAttached(final String readerName) {
                /*getSupportFragmentManager().beginTransaction()
                        .replace(R.id.nav_host_fragment, new DiscoverFragment())
                        .commit();*/
                //onSupportNavigateUp();
                NavController navController = Navigation.findNavController(mMainActivity, R.id.nav_host_fragment);
                navController.navigate(R.id.nav_discover);
                final Handler handler = new Handler();
                handler.postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(mMainActivity,
                                "ACTION_USB_DEVICE_ATTACHED "+readerName , Toast.LENGTH_LONG)
                                .show();
                        try {
                            mUemR.connect(readerName);
                    /*Toast.makeText(getApplicationContext(),
                            mUemR.getLastError()+readerName, Toast.LENGTH_LONG)
                            .show();*/
                        } catch (final Exception e) {
                            //mMainActivity.runOnUiThread(new Runnable() {
                            //    @Override
                            //    public void run() {
                            Toast.makeText(getApplicationContext(),
                                    e.getMessage(), Toast.LENGTH_LONG)
                                    .show();
                            //    }
                            //});
                        }
                        //final String[] rdrs = mUemR.USB.discover();
                /*mMainActivity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(getApplicationContext(),
                                readerName, Toast.LENGTH_LONG)
                                .show();
                    }
                });*/
                        //if (!mUemR.connect(readerName)) {
                    /*mMainActivity.runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(getApplicationContext(),
                                    uemR.getLastError(), Toast.LENGTH_LONG)
                                    .show();
                        }
                    });*/
                        //}

                        try {
                            mUemR.Reader.beep(1);
                        } catch (final Exception e) {
                            //mMainActivity.runOnUiThread(new Runnable() {
                            //@Override
                            //public void run() {
                            Toast.makeText(getApplicationContext(),
                                    e.getMessage(), Toast.LENGTH_LONG)
                                    .show();
                            //    }
                            //});
                        }
                        try {
                            mUemR.disconnect();
                        } catch (final Exception e) {
                            Toast.makeText(getApplicationContext(),
                                    e.getMessage(), Toast.LENGTH_LONG)
                                    .show();
                        }

                        mDiscoverFragment.addReader(readerName);
                    }
                }, 100);

            }
            @Override
            public void onUsbReaderDetached(String readerName) {
                Toast.makeText(mMainActivity,
                        "ACTION_USB_DEVICE_DETACHED "+readerName , Toast.LENGTH_LONG)
                        .show();
            }
        });

        //mUemR.startReceivingBroadcasts();

        FloatingActionButton fab = findViewById(R.id.fab);
        fab.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {

                Intent i = new Intent(Intent.ACTION_SEND);
                //i.setType("message/rfc822");
                //i.setData(Uri.parse("mailto:"));
                i.setDataAndType(Uri.parse("mailto:support@microem.ru"), "message/rfc822");
                i.putExtra(Intent.EXTRA_EMAIL  , new String[]{"support@microem.ru"});
                i.putExtra(Intent.EXTRA_SUBJECT, "Ваше приложение под Android: тест библиотеки считывателя");
                i.putExtra(Intent.EXTRA_TEXT   , "Уважаемая поддержка!");
                i.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
                try {
                    startActivity(Intent.createChooser(i, "Отправить email (send email)..."));
                } catch (android.content.ActivityNotFoundException ex) {
                    Toast.makeText(mMainActivity, "Не установлены почтовые клиенты. There are no email clients installed.", Toast.LENGTH_SHORT).show();
                }
                Snackbar.make(view, "Trying to send email to support...", Snackbar.LENGTH_LONG)
                        .setAction("Action", null).show();


            }
        });




        DrawerLayout drawer = findViewById(R.id.drawer_layout);
        NavigationView navigationView = findViewById(R.id.nav_view);
        // Passing each menu ID as a set of Ids because each
        // menu should be considered as top level destinations.
        mAppBarConfiguration = new AppBarConfiguration.Builder(
                R.id.nav_discover,
                R.id.nav_reader,
                R.id.nav_tags,
                R.id.nav_mifare_classic,
                R.id.nav_mifare_ultralight,
                R.id.nav_mifare_plus,
                R.id.nav_icode,
                R.id.nav_sam,
                R.id.nav_tcl)
                .setDrawerLayout(drawer)
                .build();
        NavController navController = Navigation.findNavController(this, R.id.nav_host_fragment);
        NavigationUI.setupActionBarWithNavController(this, navController, mAppBarConfiguration);
        NavigationUI.setupWithNavController(navigationView, navController);
        navController.addOnDestinationChangedListener(new NavController.OnDestinationChangedListener() {
            @Override
            public void onDestinationChanged(@NonNull NavController controller,
                                             @NonNull NavDestination destination, @Nullable Bundle arguments) {
                if(destination.getId() == R.id.nav_discover) {

                } else {

                }
            }
        });
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.main, menu);
        return true;
    }

    @Override
    public boolean onSupportNavigateUp() {
        NavController navController = Navigation.findNavController(this, R.id.nav_host_fragment);
        return NavigationUI.navigateUp(navController, mAppBarConfiguration)
                || super.onSupportNavigateUp();
    }

    @Override
    public void onRequestPermissionsResult(int requestCode,
                                           String permissions[], int[] grantResults) {

        int index = 0;
        Map<String, Integer> PermissionsMap = new HashMap<String, Integer>();
        for (String permission : permissions){
            PermissionsMap.put(permission, grantResults[index]);
            Toast.makeText(this, "p "+permission, Toast.LENGTH_SHORT).show();
            index++;
        }

        /*if((PermissionsMap.get(ACCESS_FINE_LOCATION) != 0)
                || PermissionsMap.get(READ_SMS) != 0){
            Toast.makeText(this, "Location and SMS permissions are a must", Toast.LENGTH_SHORT).show();
            finish();
        }
        else
        {
            h.postDelayed(r, 1500);
        }*/
    }

    public static void appendLog(TextView log, String commandName, String commandResult) {
        appendLog(log, commandName, commandResult, null);
    }

    public static void appendLog(TextView log, String commandName, String commandResult, String responseData) {
        if (responseData != null) {
            log.setText(responseData + "\n" + log.getText());
        }
        if (commandResult != null) {
            log.setText(commandResult + "\n" + log.getText());
        }
        if (commandName != null) {
            log.setText((new SimpleDateFormat("yyyy.MM.dd HH:mm:ss > ")).format(new Date()) + commandName + ":\n" + log.getText());
        }
    }

    public void onSendCommand(View v) {
        mDiscoverFragment.onSendCommand(v);
    }

    public void onDisconnect(View v) {
        mDiscoverFragment.onDisconnect(v);
    }

    public void onBeginDiscovery(View v) {
        mDiscoverFragment.onBeginDiscovery(v);
    }

    public void onReaderCommand(View v) {
        mReaderFragment.onReaderCommand(v);
    }

    public void onTagsCommands(View v) {
        mTagsFragment.onTagsCommands(v);
    }

    public void onSamCommands(View v) {
        mSamFragment.onSamCommands(v);
    }

    public void onTclRATSPPS(View v) {
        mTclFragment.onTclRATSPPS(v);
    }

    public void onTclTransceiveAPDU(View v) {
        mTclFragment.onTclTransceiveAPDU(v);
    }

    public void onMfcCommand(View v) {
        mMfcFragment.onMfcCommand(v);
    }

    public void onMfulCommand(View v) {
        mMifareUltralightFragment.onMfulCommand(v);
    }

    public void onMfpCommand(View v) {
        mMifarePlusFragment.onMfpCommand(v);
    }

    public void onICodeCommand(View v) {
        mICodeFragment.onICodeCommand(v);
    }
}
