package ru.microem.reader.usb_bluetooth_test.ui.discover;

import android.content.Context;
import android.content.SharedPreferences;
import android.graphics.Color;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.RadioButton;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

import java.util.ArrayList;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.fragment.app.Fragment;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModelProviders;
import ru.microem.reader.usb_bluetooth_test.MainActivity;
import ru.microem.reader.usb_bluetooth_test.R;
import ru.microem.readerlibrary.UemReader;

public class DiscoverFragment extends Fragment implements AdapterView.OnItemSelectedListener {

    private DiscoverViewModel discoverViewModel;

    private View mRoot = null;

    private ArrayList<String> mListBLE = null;

    public View onCreateView(@NonNull LayoutInflater inflater,
                             ViewGroup container, Bundle savedInstanceState) {
        discoverViewModel =
                ViewModelProviders.of(this).get(DiscoverViewModel.class);
        View root = inflater.inflate(R.layout.fragment_discover, container, false);
        final TextView textView = root.findViewById(R.id.text_discover);
        discoverViewModel.getText().observe(this, new Observer<String>() {
            @Override
            public void onChanged(@Nullable String s) {
                textView.setText(s);
            }
        });

        mRoot = root;

        ((MainActivity)getActivity()).mDiscoverFragment = this;

        Spinner spinner = (Spinner) root.findViewById(R.id.cb_command);
        spinner.setOnItemSelectedListener(this);

        final ListView listBLE = root.findViewById(R.id.bt_list);

        mListBLE = new ArrayList<String>();

        ArrayAdapter<String> adapter = new ArrayAdapter<String>(getActivity().getApplicationContext(),
                android.R.layout.simple_list_item_1, android.R.id.text1, mListBLE){

            @Override
            public View getView(int position, View convertView, ViewGroup parent) {
                View view =super.getView(position, convertView, parent);

                TextView textView=(TextView) view.findViewById(android.R.id.text1);

                /*YOUR CHOICE OF COLOR*/
                textView.setTextColor(Color.BLUE);

                return view;
            }
        };

        listBLE.setAdapter(adapter);
        listBLE.setOnItemClickListener(new AdapterView.OnItemClickListener() {

            @Override
            public void onItemClick(AdapterView<?> parent, View view,
                                    int position, long id) {

                // ListView Clicked item index
                int itemPosition     = position;

                // ListView Clicked item value
                String  itemValue    = (String) listBLE.getItemAtPosition(position);

                // Show Alert
                Toast.makeText(getActivity().getApplicationContext(),
                        "Connecting " +itemValue+ " ..." , Toast.LENGTH_LONG)
                        .show();

                ((MainActivity)getActivity()).mUemR.disconnect();
                ((MainActivity)getActivity()).mUemR.connect(itemValue);

                if (((MainActivity)getActivity()).mUemR.getInterfaceType().equals(UemReader.UEM_INTERFACE_USB)) {
                    ((MainActivity)getActivity()).mUemR.Reader.beep(3);
                    //mUemR.disconnect();
                }
            }

        });

        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        ((Spinner)root.findViewById(R.id.cb_command)).setSelection(sharedPref.getInt("cb_command", 0));
        ((EditText)root.findViewById(R.id.customCmd)).setText(sharedPref.getString("customCmd", ""));
        ((TextView)root.findViewById(R.id.outtext)).setText(sharedPref.getString("outtext", "Hello World!"));

        return root;
    }

    public void onPause() {
        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        sharedPref.edit().putInt("cb_command", ((Spinner)mRoot.findViewById(R.id.cb_command)).getSelectedItemPosition()).apply();
        sharedPref.edit().putString("customCmd", ((EditText)mRoot.findViewById(R.id.customCmd)).getText().toString()).apply();
        sharedPref.edit().putString("outtext", ((TextView)mRoot.findViewById(R.id.outtext)).getText().toString()).apply();

        super.onPause();
    }

    @Override
    public void onViewCreated(View view, @Nullable Bundle savedInstanceState) {

        /*Toast.makeText(getActivity().getApplicationContext(),
                "ReaderFragment onViewCreated", Toast.LENGTH_LONG)
                .show();*/


    }

    public void addReader(String value){
        boolean found = false;
        for (String string : mListBLE) {
            if(string.equals(value)){
                found = true;
                break;
            }
        }
        if (!found) {
            ListView listBLE = mRoot.findViewById(R.id.bt_list);
            mListBLE.add(value);
            ((ArrayAdapter<String>) listBLE.getAdapter()).notifyDataSetChanged();
        }
    }

    public void onItemSelected(AdapterView<?> parent, View view,
                               int pos, long id) {
        // parent.getItemAtPosition(pos)

            TextView textView = (TextView) mRoot.findViewById(R.id.customCmd);
            String cmd;
            try {
                cmd = UemReader.Helper.byteArrayToHexString((((MainActivity)getActivity()).mCommands[pos]));
            } catch (Exception e) {
                System.out.println("error " + e.getMessage());
                return;
            }
            textView.setText(cmd);
    }

    public void onNothingSelected(AdapterView<?> parent) {
        // Another interface callback
    }

    /** Called when the user touches the button */
    public void onSendCommand(View view) {
        switch(view.getId()) {
            case R.id.send:
                TextView textView = (TextView) mRoot.findViewById(R.id.customCmd);
                byte[] cmd = UemReader.Helper.hexStringToByteArray(textView.getText().toString());
                byte[] res = ((MainActivity)getActivity()).mUemR.Reader.directIOReader(cmd);
                if (res != null) {
                    TextView resView = (TextView) mRoot.findViewById(R.id.outtext);
                    resView.setText("status: " + ((MainActivity)getActivity()).mUemR.getLastErrorCode() + " result: " + UemReader.Helper.byteArrayToHexString(res));
                } else {
                    ((MainActivity)getActivity()).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(getActivity().getApplicationContext(),
                                    ((MainActivity)getActivity()).mUemR.getLastError(), Toast.LENGTH_LONG)
                                    .show();
                        }
                    });
                }
                /*EditText textView = (EditText) findViewById(R.id.customCmd);
                if (gattClient != null && characteristicFIFO != null && (characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) > 0) {
                    characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
                    String s = (String)textView.getText().toString();
                    characteristicFIFO.setValue(wrapCommand(hexStringToByteArray(s)));
                    gattClient.writeCharacteristic(characteristicFIFO);
                }*/
            break;
            default:
        }
    }

    public void onDisconnect(View view) {
        switch(view.getId()) {
            case R.id.disconnect:
                ((MainActivity)getActivity()).mUemR.disconnect();
                break;
            default:
        }
    }

    public void onBeginDiscovery(View view) {
        switch(view.getId()) {
            case R.id.discover:
                mListBLE.clear();
                ListView listBLE = mRoot.findViewById(R.id.bt_list);
                ((ArrayAdapter<String>) listBLE.getAdapter()).notifyDataSetChanged();
                ((MainActivity)getActivity()).mUemR.disconnect();
                ((MainActivity)getActivity()).mUemR.Reader.setEncryptionMode(false);
                final String[] arReaders = ((MainActivity)getActivity()).mUemR.USB.discover();
                if (arReaders.length > 0) {
                    if (((MainActivity)getActivity()).mUemR.connect(arReaders[0])) {
                        ((MainActivity)getActivity()).runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                Toast.makeText(getActivity().getApplicationContext(),
                                        arReaders[0], Toast.LENGTH_LONG)
                                        .show();
                            }
                        });
                                /*byte[] recBytes = uemR.DirectIO_Reader(new byte[]{0x05, 0x02});
                                if (recBytes != null) {
                                    final String bufReceived = UemReader.Helper.byteArrayToHexString(recBytes);
                                    mMainActivity.runOnUiThread(new Runnable() {
                                        @Override
                                        public void run() {
                                            Toast.makeText(getApplicationContext(),
                                                    bufReceived, Toast.LENGTH_LONG)
                                                    .show();
                                        }
                                    });
                                }*/
                        ((MainActivity)getActivity()).mUemR.Reader.beep(3);
                        ((MainActivity)getActivity()).mUemR.disconnect();
                        addReader(arReaders[0]);
                    }
                }

                ((MainActivity)getActivity()).mUemR.Bluetooth.startDiscovery();
                break;
            default:
        }
    }
}