package ru.microem.reader.usb_bluetooth_test.ui.mifare_classic;

import android.content.Context;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.RadioButton;
import android.widget.Spinner;
import android.widget.TextView;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;
import androidx.fragment.app.Fragment;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModelProviders;
import ru.microem.reader.usb_bluetooth_test.MainActivity;
import ru.microem.reader.usb_bluetooth_test.R;
import ru.microem.readerlibrary.UemReader;

public class MifareClassicFragment extends Fragment {

    private MifareClassicViewModel mifareClassicViewModel;

    private View mRoot = null;

    public View onCreateView(@NonNull LayoutInflater inflater,
                             ViewGroup container, Bundle savedInstanceState) {
        mifareClassicViewModel =
                ViewModelProviders.of(this).get(MifareClassicViewModel.class);
        View root = inflater.inflate(R.layout.fragment_mifare_classic, container, false);
        final TextView textView = root.findViewById(R.id.text_mifare_classic);
        mifareClassicViewModel.getText().observe(this, new Observer<String>() {
            @Override
            public void onChanged(@Nullable String s) {
                textView.setText(s);
            }
        });

        mRoot = root;

        ((MainActivity)getActivity()).mMfcFragment = this;

        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        ((EditText)root.findViewById(R.id.mfc_sector)).setText(sharedPref.getString("mfc_sector", "0"));
        ((EditText)root.findViewById(R.id.mfc_block)).setText(sharedPref.getString("mfc_block", "1"));
        ((EditText)root.findViewById(R.id.mfc_key)).setText(sharedPref.getString("mfc_key", "FFFFFFFFFFFF"));
        ((CheckBox)root.findViewById(R.id.mfc_use_sam)).setChecked(sharedPref.getBoolean("mfc_use_sam", false));
        ((EditText)root.findViewById(R.id.mfc_sam_key_no)).setText(sharedPref.getString("mfc_sam_key_no", "11"));
        ((EditText)root.findViewById(R.id.mfc_sam_key_ver)).setText(sharedPref.getString("mfc_sam_key_ver", "0"));
        ((RadioButton)root.findViewById(R.id.mfc_key_a)).setChecked(sharedPref.getBoolean("mfc_key_a", true));
        ((RadioButton)root.findViewById(R.id.mfc_key_b)).setChecked(sharedPref.getBoolean("mfc_key_b", false));
        ((EditText)root.findViewById(R.id.mfc_data)).setText(sharedPref.getString("mfc_data", "00000000000000000000000000000000"));
        ((TextView)root.findViewById(R.id.mfc_log)).setText(sharedPref.getString("mfc_log", ""));

        return root;
    }

    public void onPause() {
        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        sharedPref.edit().putString("mfc_sector", ((EditText)mRoot.findViewById(R.id.mfc_sector)).getText().toString()).apply();
        sharedPref.edit().putString("mfc_block", ((EditText)mRoot.findViewById(R.id.mfc_block)).getText().toString()).apply();
        sharedPref.edit().putString("mfc_key", ((EditText)mRoot.findViewById(R.id.mfc_key)).getText().toString()).apply();
        sharedPref.edit().putBoolean("mfc_use_sam", ((CheckBox)mRoot.findViewById(R.id.mfc_use_sam)).isChecked()).apply();
        sharedPref.edit().putString("mfc_sam_key_no", ((EditText)mRoot.findViewById(R.id.mfc_sam_key_no)).getText().toString()).apply();
        sharedPref.edit().putString("mfc_sam_key_ver", ((EditText)mRoot.findViewById(R.id.mfc_sam_key_ver)).getText().toString()).apply();
        sharedPref.edit().putBoolean("mfc_key_a", ((RadioButton)mRoot.findViewById(R.id.mfc_key_a)).isChecked()).apply();
        sharedPref.edit().putBoolean("mfc_key_b", ((RadioButton)mRoot.findViewById(R.id.mfc_key_b)).isChecked()).apply();
        sharedPref.edit().putString("mfc_data", ((EditText)mRoot.findViewById(R.id.mfc_data)).getText().toString()).apply();
        sharedPref.edit().putString("mfc_log", ((TextView)mRoot.findViewById(R.id.mfc_log)).getText().toString()).apply();

        super.onPause();
    }

    public void onMfcCommand(View view) {
        EditText editSector = mRoot.findViewById(R.id.mfc_sector);
        int sector = Integer.parseInt(editSector.getText().toString());
        EditText editBlock = mRoot.findViewById(R.id.mfc_block);
        int block = Integer.parseInt(editBlock.getText().toString());
        EditText editKey = mRoot.findViewById(R.id.mfc_key);
        byte[] key = UemReader.Helper.hexStringToByteArray(editKey.getText().toString());
        CheckBox checkUseSAM = mRoot.findViewById(R.id.mfc_use_sam);
        boolean useSAM = checkUseSAM.isChecked();
        EditText editSAMKeyNo = mRoot.findViewById(R.id.mfc_sam_key_no);
        int samKeyNo = Integer.parseInt(editSAMKeyNo.getText().toString());
        EditText editSAMKeyVer = mRoot.findViewById(R.id.mfc_sam_key_ver);
        int samKeyVer = Integer.parseInt(editSAMKeyVer.getText().toString());
        RadioButton radioKeyA = mRoot.findViewById(R.id.mfc_key_a);
        boolean useKeyA = radioKeyA.isChecked();
        EditText editData = mRoot.findViewById(R.id.mfc_data);
        byte[] data = UemReader.Helper.hexStringToByteArray(editData.getText().toString());
        TextView log = mRoot.findViewById(R.id.mfc_log);
        switch(view.getId()) {
            case R.id.mfc_authenticate:
                if (!useSAM) {
                    ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.authenticateWithPlainKey(
                            sector,
                            useKeyA ? UemReader.CommandsCards.CommandsMifareClassic.UEM_MIFARE_KEY_TYPE_A : UemReader.CommandsCards.CommandsMifareClassic.UEM_MIFARE_KEY_TYPE_B,
                            key);
                    ((MainActivity) getActivity()).appendLog(log, "Authenticate", String.format("%s, Status: 0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                    ));
                } else {
                    ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.authenticateFirstSAM(
                            sector,
                            useKeyA ? UemReader.CommandsCards.CommandsMifareClassic.UEM_MIFARE_KEY_TYPE_A : UemReader.CommandsCards.CommandsMifareClassic.UEM_MIFARE_KEY_TYPE_B,
                            samKeyNo,
                            samKeyVer
                    );
                    ((MainActivity) getActivity()).appendLog(log, "Authenticate SAM:", String.format("%s, Status: 0x%02x, SW1=0x%02x SW2=0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode(),
                            ((MainActivity) getActivity()).mUemR.SAM.SW[0],
                            ((MainActivity) getActivity()).mUemR.SAM.SW[1]
                    ));
                }
                break;
            case R.id.mfc_read:
                if (!useSAM) {
                    byte[] res = ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.readBlock(sector, block);
                    ((MainActivity) getActivity()).appendLog(log, "Read", String.format("%s, Status: 0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                            ), ((MainActivity) getActivity()).mUemR.getLastError().equals(UemReader.UEM_STATUS_SUCCESS)?String.format("Data: %s",
                            UemReader.Helper.byteArrayToHexString(res)
                            ):null
                    );
                } else {
                    byte[] res = ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.readBlockSAM(sector, block);
                    ((MainActivity) getActivity()).appendLog(log, "Read SAM", String.format("%s, Status: 0x%02x, SW1=0x%02x SW2=0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode(),
                            ((MainActivity) getActivity()).mUemR.SAM.SW[0],
                            ((MainActivity) getActivity()).mUemR.SAM.SW[1]
                            ), ((MainActivity) getActivity()).mUemR.getLastError().equals(UemReader.UEM_STATUS_SUCCESS)?String.format("Data: %s",
                            UemReader.Helper.byteArrayToHexString(res)):null
                    );
                }
                break;
            case R.id.mfc_write:
                if (!useSAM) {
                    ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.writeBlock(sector, block, data);
                    ((MainActivity) getActivity()).appendLog(log, "Write", String.format("%s, Status: 0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                    ));
                } else {
                    ((MainActivity) getActivity()).mUemR.Cards.MifareClassic.writeBlockSAM(sector, block, data);
                    ((MainActivity) getActivity()).appendLog(log, "Write SAM:", String.format("%s, Status: 0x%02x, SW1=0x%02x SW2=0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode(),
                            ((MainActivity) getActivity()).mUemR.SAM.SW[0],
                            ((MainActivity) getActivity()).mUemR.SAM.SW[1]
                    ));
                }
                break;
            default:
        }
    }
}