package ru.microem.reader.usb_bluetooth_test.ui.tcl;

import android.content.Context;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.EditText;
import android.widget.RadioButton;
import android.widget.TextView;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;
import androidx.fragment.app.Fragment;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModelProviders;
import ru.microem.reader.usb_bluetooth_test.MainActivity;
import ru.microem.reader.usb_bluetooth_test.R;
import ru.microem.readerlibrary.UemReader;

public class TclFragment extends Fragment {

    private TclViewModel tclViewModel;

    private View mRoot = null;

    public View onCreateView(@NonNull LayoutInflater inflater,
                             ViewGroup container, Bundle savedInstanceState) {
        tclViewModel =
                ViewModelProviders.of(this).get(TclViewModel.class);
        View root = inflater.inflate(R.layout.fragment_tcl, container, false);
        final TextView textView = root.findViewById(R.id.text_tcl);
        tclViewModel.getText().observe(this, new Observer<String>() {
            @Override
            public void onChanged(@Nullable String s) {
                textView.setText(s);
            }
        });

        mRoot = root;

        ((MainActivity)getActivity()).mTclFragment = this;

        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        ((EditText)root.findViewById(R.id.tcl_apdu)).setText(sharedPref.getString("tcl_apdu", "00A4040C07a000000247100100"));

        return root;
    }

    public void onPause() {
        SharedPreferences sharedPref = getActivity().getPreferences(Context.MODE_PRIVATE);
        sharedPref.edit().putString("tcl_apdu", ((EditText)mRoot.findViewById(R.id.tcl_apdu)).getText().toString()).apply();

        super.onPause();
        //Log.d(LOG_TAG, "Fragment1 onDestroy");
    }

    /*public void onStop() {
        super.onStop();
        //Log.d(LOG_TAG, "Fragment1 onDestroy");
    }

    public void onDestroyView() {
        super.onDestroyView();
        //Log.d(LOG_TAG, "Fragment1 onDestroyView");
    }*/

    /*public void onDestroy() {
        super.onDestroy();
        //Log.d(LOG_TAG, "Fragment1 onDestroy");
    }

    public void onDetach() {
        super.onDetach();
        //Log.d(LOG_TAG, "Fragment1 onDetach");
    }*/

    public void onTclTransceiveAPDU(View view) {
        switch(view.getId()) {
            case R.id.tcl_send_apdu:
                EditText send = mRoot.findViewById(R.id.tcl_apdu);
                byte[] res = ((MainActivity) getActivity()).mUemR.Cards.exchangeISO14443TCL(UemReader.Helper.hexStringToByteArray(send.getText().toString()));
                TextView log = mRoot.findViewById(R.id.tcl_log);
                ((MainActivity) getActivity()).appendLog(log, "Transceive APDU", String.format("%s, Status: 0x%02x",
                        ((MainActivity) getActivity()).mUemR.getLastError(),
                        ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                        ), ((MainActivity) getActivity()).mUemR.getLastError().equals(UemReader.UEM_STATUS_SUCCESS)?String.format("Response: %s",
                        UemReader.Helper.byteArrayToHexString(res)):null
                );
                break;
            default:
        }
    }

    public void onTclRATSPPS(View view) {
        switch(view.getId()) {
            case R.id.tcl_rats_pps:
                TextView log = mRoot.findViewById(R.id.tcl_log);
                if (((MainActivity) getActivity()).mActiveMode == UemReader.CommandsCards.UEM_CARD_STANDARD_ISO14443A) {
                    byte[] atr = ((MainActivity) getActivity()).mUemR.Cards.ISO14443A.activateTCLdeprecated();
                    ((MainActivity) getActivity()).appendLog(log, "RATS+PPS", String.format("%s, Status: 0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                            ), ((MainActivity) getActivity()).mUemR.getLastError().equals(UemReader.UEM_STATUS_SUCCESS)?String.format("ATR: %s",
                                UemReader.Helper.byteArrayToHexString(atr)
                            ):null
                    );
                }
                if (((MainActivity) getActivity()).mActiveMode == UemReader.CommandsCards.UEM_CARD_STANDARD_ISO14443B) {
                    /*CLSCRF_Attrib_B( pParent->m_pReader,
                            m_pCard->UID,
                            0x00,                                     // Param1
//                                 PICC -> PCD    PCD -> PICC
                            ( DSI << 6 ) | ( DRI << 4 ) | cPCD_FSDI,  // Param2
                            1,    // PCD compliant with ISO/IEC 14443-4  Param3
                            m_pCard->CID,                             // Param4
                            RecvBuffer,
                            BUF_SIZE,
                            &RecvLength,
                                  &MbliCid );*/
                    UemReader.CommandsCards.CommandsISO14443B.AttribISO14443B res = ((MainActivity) getActivity()).mUemR.Cards.ISO14443B.attribDeprecated(
                            ((MainActivity) getActivity()).mUID,
                            (byte)0x00,
                            (byte)0x00,
                            (byte)0x01,
                            (byte)0x00,
                            null);
                    ((MainActivity) getActivity()).appendLog(log, "Attrib", String.format("%s, Status: 0x%02x",
                            ((MainActivity) getActivity()).mUemR.getLastError(),
                            ((MainActivity) getActivity()).mUemR.getLastErrorCode()
                            ), ((MainActivity) getActivity()).mUemR.getLastError().equals(UemReader.UEM_STATUS_SUCCESS) ? String.format("hlBuf: %s mbliCid: %s\n",
                                    UemReader.Helper.byteArrayToHexString(res.higherLayerBuffer != null ? res.higherLayerBuffer : new byte[0]),
                                    String.valueOf(res.mbliCid)
                            ) : null
                    );
                }

                break;
            default:
        }
    }
}