/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * mifare_classic.cc
 * Copyright (C) 2021 MicroEM JSC <support@microem.ru>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name ``MicroEM JSC'' nor the name of any other
 *    contributor may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * CLESCAR IS PROVIDED BY MicroEM JSC ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL MicroEM JSC OR ANY OTHER CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "mifare_classic.h"

extern PVOID m_pReader;

Gtk::Entry* edt_mifare_key;
Gtk::RadioButton* rad_mifare_key_a;
Gtk::RadioButton* rad_mifare_key_b;
Gtk::Button* btn_mifare_load_key;
Gtk::SpinButton* spin_mifare_sector;
Gtk::Button* btn_mifare_auth;
Gtk::SpinButton* spin_mifare_block;
Gtk::Button* btn_mifare_read;
Gtk::Entry* edt_mifare_data;
Gtk::Button* btn_mifare_write;

extern BYTE card_uid[12];
extern int card_uid_length;

void on_btn_mifare_load_key_clicked()
{
	LONG Status;
	BYTE Key[6] = {0};

	Glib::ustring ustrKey = edt_mifare_key->get_text();
	std::string strKey = std::string(ustrKey);
	char* cKey = (char*)(strKey.c_str());
	hex2bin(cKey, Key, 6);

	printf("CLSCRF_MifareStandard_WriteKeyToE2: ");
	Status = CLSCRF_MifareStandard_WriteKeyToE2(m_pReader, 
	                                            0x60 + rad_mifare_key_b->get_active(), 
	                                            0x00, 
	                                            Key);
	if( Status == 0 )
		printf("OK\n");
    else
        printf("Error 0x%08X\n", (UINT)Status);
}

void on_btn_mifare_auth_clicked()
{
	LONG Status;
	
	if (card_uid_length == 0)
		return;

	printf("CLSCRF_MifareStandard_AuthE2Ex: ");
	Status = CLSCRF_MifareStandard_AuthE2Ex(m_pReader, 
	                                        0x60 + rad_mifare_key_b->get_active(), 
	                                        card_uid, 
	                                        0x00, 
	                                        spin_mifare_sector->get_value_as_int());
	if( Status == 0 )
		printf("OK\n");
    else
        printf("Error 0x%08X\n", (UINT)Status);
}

void on_btn_mifare_read_clicked()
{
	LONG Status;
	BYTE Data[16] = {0};
	char out[256] = {0};

	printf("CLSCRF_MifareStandard_Read: ");
	Status = CLSCRF_MifareStandard_Read(m_pReader,
                                        spin_mifare_sector->get_value_as_int(),
                                        spin_mifare_block->get_value_as_int(),
                                        Data);
    if( Status == 0 )
	{
		printf("OK\n");
		bin2hex(Data, out, 16);
		edt_mifare_data->set_text(out);
	}
    else
	{
        printf("Error 0x%08X\n", (UINT)Status);
		edt_mifare_data->set_text("");
	}

}

void on_btn_mifare_write_clicked()
{
	LONG Status;
	BYTE Data[16] = {0};

	Glib::ustring ustrData = edt_mifare_data->get_text();
	std::string strData = std::string(ustrData);
	char* cData = (char*)(strData.c_str());
	hex2bin(cData, Data, 16);

	printf("CLSCRF_MifareStandard_Write: ");
	Status = CLSCRF_MifareStandard_Write(m_pReader,
                                         spin_mifare_sector->get_value_as_int(),
                                         spin_mifare_block->get_value_as_int(),
                                         Data);
	if( Status == 0 )
		printf("OK\n");
    else
        printf("Error 0x%08X\n", (UINT)Status);
}
