/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * search.cc
 * Copyright (C) 2021 MicroEM JSC <support@microem.ru>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name ``MicroEM JSC'' nor the name of any other
 *    contributor may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * CLESCAR IS PROVIDED BY MicroEM JSC ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL MicroEM JSC OR ANY OTHER CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "search.h"
#include <iostream>

bool stop_search = false;
bool pause_search = true;
bool search_paused = false;

std::mutex m_Mutex;

extern Glib::Dispatcher m_DispatcherNoCards;
extern Glib::Dispatcher m_DispatcherNewCard;

Gtk::ListBox* list_cards;
Gtk::CheckButton* chk_enable_beep;

char card_info[256] = {0};

BYTE card_uid[12] = {0};
int card_uid_length = 0;

void search_thread_entry(PVOID pReader, Gtk::ListBox* list_cards)
{
	ULONG Status;
	BYTE ATQ[2];
	BYTE SAK;
	BYTE UID[10];
	DWORD i, n;
	char Err;
	
    // Do something with that, I guess

	while (!stop_search)
	{
		if (pause_search)
			search_paused = true;
		while (pause_search && (!stop_search))
		{
		}
		if (stop_search)
			break;
		if (search_paused)
		{
			search_paused = false;
			Status = CLSCRF_Mfrc_On( pReader );
		}
		//std::cout << "Searching..." << std::endl;
		//Sleep( 1000 );
		
		CLSCRF_Mfrc_Rf_Off_On( pReader, 1 );

		//list_cards->children()->removeremove_all();
		/*for (i = 0; i < list_cards->get_children().size(); i++)
		{
			Gtk::Widget* element = list_cards->get_children()[i];
  			list_cards->remove(*element);
			delete element;
		}*/
		
		//printf("Activate_Idle_A\n");
		Status = CLSCRF_Activate_Idle_A( pReader,
		                                 ATQ, // 2 bytes
		                                 &SAK, // pbSAK 1 byte
		                                 UID, // pbUID max 10 bytes
		                                 &n );
		if( Status == 0 )
		{
			std::lock_guard<std::mutex> lock(m_Mutex);
			if (chk_enable_beep->get_active())
				CLSCRF_Sound( pReader, 1 );
		    int pos = snprintf(card_info, sizeof(card_info) - pos, "ATQ=%02X%02X SAK=%02X UID=", ATQ[1], ATQ[0], SAK );
		    for( i = 0; i < n; i++ )
		        pos += snprintf(card_info + pos, sizeof(card_info) - pos, "%02X", UID[i] );
			
		    //sprintf("\n");
		    //CLSCRF_Halt_A( m_pReader );
			//Gtk::Label* label = new Gtk::Label();
			//label->set_text(card_info);
			//list_cards->add(*label);
			//list_cards->insert(*label, -1);
			//list_cards->show_all();

			card_uid_length = n;
			memcpy(card_uid, UID, card_uid_length);
			
			m_DispatcherNoCards.emit();
			m_DispatcherNewCard.emit();
		}
		else
		{
			card_uid_length = 0;
			
			m_DispatcherNoCards.emit();
				
		    //printf("Sorry, error 0x%08X\n", Status );
		    if( Status == 0x80100001 )
		    {
		        CLSCRF_GetLastInternalError( pReader, (BYTE*)&Err );
		        printf("Internal Error = %d\n", Err );
		    }
		}

	}
}