package com.example.uemreaderlibrary;

import android.Manifest;
import android.app.Activity;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothManager;
import android.bluetooth.BluetoothProfile;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanResult;
import android.bluetooth.le.ScanSettings;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbEndpoint;
import android.hardware.usb.UsbInterface;
import android.hardware.usb.UsbManager;
import android.os.Build;
import android.util.Log;
import android.widget.ArrayAdapter;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;
import java.util.Vector;

import static android.bluetooth.BluetoothDevice.PHY_LE_1M_MASK;
import static android.bluetooth.BluetoothDevice.PHY_LE_CODED_MASK;
import static android.bluetooth.BluetoothDevice.TRANSPORT_LE;
import static android.content.ContentValues.TAG;
import static android.content.Context.BLUETOOTH_SERVICE;

public class UemReader {
    public static final String UEM_INTERFACE_UNKNOWN = "Unknown Interface";
    public static final String UEM_INTERFACE_USB = "USB Interface";
    public static final String UEM_INTERFACE_BLUETOOTH = "Bluetooth Interface";

    public static final String UEM_STATUS_SUCCESS = "OK";
    public static final String UEM_STATUS_OPERATION_PENDING = "Operation in progress";
    public static final String UEM_ERROR_READER_INCORRECT_NAME = "Incorrect reader name";
    public static final String UEM_ERROR_READER_CONNECTION_FAILED = "Failed to connect to the reader";
    public static final String UEM_ERROR_READER_NOT_CONNECTED = "Reader not connected";
    public static final String UEM_ERROR_READER_ALREADY_CONNECTED = "Reader already connected";
    public static final String UEM_ERROR_READER_INCORRECT_RESPONSE = "Incorrect reader response";
    public static final String UEM_ERROR_READER_RESPONSE_FAILURE = "Reader not responding";

    private Context mContext = null;

    //private HashMap<Object, String> mConnectedReaders = new HashMap<Object, String>();
    private String mInterfaceType = null;

    private String mLastError = UEM_STATUS_SUCCESS;

    private Object mConnection = null;

    public UemReader(Context context) {
        if (context == null)
            throw new NullPointerException("Parameter is null!");
        this.mContext = context;
    }

    public String GetLastError() {
        return mLastError;
    }

    public class CommandsReader {
        public byte[] DirectIO(byte[] command) {
            return DispatchCommand(command);
        }

        public byte Sound(int nBeepCount) {
            byte res[] = DispatchCommand(new byte[]{0x05, (byte)nBeepCount});
            if (res != null)
                return res[0];
            return 0x00;
        }
    }

    public CommandsReader Reader = new CommandsReader();



    /*private byte[] DispatchCommand(Object oReader, byte[] command) {
        String interfaceType = mConnectedReaders.get(oReader);
        if (interfaceType == null) {
            throw new IllegalArgumentException("Reader is not connected!");
        }
        if (interfaceType.equals(UEM_INTERFACE_USB)) {
            byte[] resp = this.USB.Tranceive((UsbDeviceConnection)oReader, command);
            if (resp[0] != command[0]) {
                lastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            return Arrays.copyOfRange(resp, 0, resp.length-2);
        }
        return null;
    }*/

    private byte[] DispatchCommand(byte[] command) {
        //String interfaceType = mConnectedReaders.get(oReader);

        if (mInterfaceType == null) {
            throw new IllegalArgumentException("Reader is not connected!");
        }
        if (mInterfaceType.equals(UEM_INTERFACE_USB)) {
            byte[] resp = this.USB.Transceive(command);
            if (resp == null) {
                return null;
            }
            if (resp.length == 0 || resp[0] != command[0]) {
                mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                return null;
            }
            return Arrays.copyOfRange(resp, 1, resp.length);
        }
        if (mInterfaceType.equals(UEM_INTERFACE_BLUETOOTH)) {
            boolean res = this.Bluetooth.SendCommand(command);
            if (res) {
                mLastError = UEM_STATUS_OPERATION_PENDING;
                return null;
            }
        }
        return null;
    }

    public boolean Connect(String readerName) {
        return this.USB.Connect(readerName) || this.Bluetooth.Connect(readerName);
    }

    public boolean Disconnect() {
        if (mInterfaceType.equals(UEM_INTERFACE_USB))
            return this.USB.Disconnect();
        if (mInterfaceType.equals(UEM_INTERFACE_BLUETOOTH))
            return this.Bluetooth.Disconnect();
        return false;
    }

    public interface Callbacks {
        void onBluetoothReaderFound(String readerName);
        void onBluetoothReaderConnected();
        void onBluetoothResponse(byte[] response);
    }

    public class InterfaceUSB {
        private HashMap<String, UsbDevice> mReadersList = new HashMap<String, UsbDevice>();

        public String[] Discover() {
            UsbManager usbManager = (UsbManager)mContext.getSystemService(Context.USB_SERVICE);
            mReadersList = usbManager.getDeviceList();
            ArrayList<String> arReaders = new ArrayList<String>();

            for (UsbDevice usbDevice: mReadersList.values()) {
                if ((usbDevice.getVendorId() == 49745) && (usbDevice.getProductId() == 4874)) {
                    arReaders.add(usbDevice.getDeviceName());
                    //if (mDiscoveryCallback != null)
                    //    mDiscoveryCallback.onNewReaderFound(usbDevice.getDeviceName());
                }
            }

            Object[] objReaders = arReaders.toArray();
            String[] strReaders = Arrays.copyOf(objReaders, objReaders.length, String[].class);
            return strReaders;
        }

        private UsbEndpoint mEndpointBulkIn = null;
        private UsbEndpoint mEndpointBulkOut = null;

        /*public UsbDeviceConnection Connect(String readerName) {
            if (!readersList.containsKey(readerName)) {
                lastError = UEM_ERROR_READER_INCORRECT_NAME;
                return null;
            }

            UsbManager usbManager = (UsbManager) getSystemService(Context.USB_SERVICE);
            UsbDevice reader = readersList.get(readerName);
            UsbInterface interfaceUEM = reader.getInterface(0);
            endpointBulkIn = interfaceUEM.getEndpoint(0);
            endpointBulkOut = interfaceUEM.getEndpoint(1);
            UsbDeviceConnection connection = usbManager.openDevice(reader);
            if (connection == null) {
                lastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return null;
            }

            connection.claimInterface(interfaceUEM, true);
            mConnectedReaders.put(connection, UEM_INTERFACE_USB);
            interfaceType = UEM_INTERFACE_USB;

            return connection;
        }*/

        private boolean Connect(String readerName) {
            if (mConnection != null) {
                mLastError = UEM_ERROR_READER_ALREADY_CONNECTED;
                return false;
            }
            if (!mReadersList.containsKey(readerName)) {
                mLastError = UEM_ERROR_READER_INCORRECT_NAME;
                return false;
            }

            UsbManager usbManager = (UsbManager)mContext.getSystemService(Context.USB_SERVICE);
            UsbDevice reader = mReadersList.get(readerName);
            UsbInterface interfaceUEM = reader.getInterface(0);
            mEndpointBulkIn = interfaceUEM.getEndpoint(0);
            mEndpointBulkOut = interfaceUEM.getEndpoint(1);
            UsbDeviceConnection connection = usbManager.openDevice(reader);
            if (connection == null) {
                mLastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return false;
            }

            connection.claimInterface(interfaceUEM, true);
            mConnection = connection;
            mInterfaceType = UEM_INTERFACE_USB;

            return true;
        }

        /*public boolean Disconnect(UsbDeviceConnection connection) {
            if (connection == null)
                return false;

            connection.close();
            mConnectedReaders.remove(connection);
            connection = null;

            return true;
        }*/

        private boolean Disconnect() {
            if (mConnection == null)
                return false;

            ((UsbDeviceConnection)mConnection).close();
            mConnection = null;
            mInterfaceType = null;

            return true;
        }

        /*public byte[] Tranceive(UsbDeviceConnection connection, byte[] sendBuffer) {
            if (connection == null || sendBuffer == null)
                throw new NullPointerException("Parameter is null!");
            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            int TIMEOUT = 0;
            connection.bulkTransfer(endpointBulkOut, sendBuffer, sendBuffer.length, TIMEOUT);

            byte[] receiveBuffer = new byte[256];
            final int nReceived = connection.bulkTransfer(endpointBulkIn, receiveBuffer, receiveBuffer.length, TIMEOUT);

            if (nReceived < 0) {
                lastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                return null;
            }

            byte[] response = UemReader.Helper.unwrapResponse(Arrays.copyOf(receiveBuffer, nReceived));

            return response;
        }*/

        private byte[] Transceive(byte[] sendBuffer) {
            if (mConnection == null || sendBuffer == null)
                throw new NullPointerException("Parameter is null!");

            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            int TIMEOUT = 0;
            byte[] command = UemReader.Helper.wrapCommand(sendBuffer);
            ((UsbDeviceConnection)mConnection).bulkTransfer(mEndpointBulkOut, command, command.length, TIMEOUT);

            byte[] receiveBuffer = new byte[256];
            final int nReceived = ((UsbDeviceConnection)mConnection).bulkTransfer(mEndpointBulkIn, receiveBuffer, receiveBuffer.length, TIMEOUT);

            if (nReceived < 0) {
                mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                return null;
            }

            byte[] response = UemReader.Helper.unwrapResponse(Arrays.copyOf(receiveBuffer, nReceived));

            return response;
        }
    }

    public InterfaceUSB USB = new InterfaceUSB();

    public static class Helper {
        public static byte[] hexStringToByteArray(String s) {
            int len = s.length();
            byte[] data = new byte[len / 2];

            for (int i = 0; i < len; i += 2)
                data[i / 2] = (byte) ((Character.digit(s.charAt(i), 16) << 4) + Character.digit(s.charAt(i + 1), 16));

            return data;
        }

        final private static char[] hexArray = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

        public static String byteArrayToHexString(byte[] bytes) {
            char[] hexChars = new char[bytes.length * 2];
            int v;

            for (int j = 0; j < bytes.length; j++) {
                v = bytes[j] & 0xFF;
                hexChars[j * 2] = hexArray[v >>> 4];
                hexChars[j * 2 + 1] = hexArray[v & 0x0F];
            }

            return new String(hexChars);
        }

        private static byte[] CRC16(byte[] buf, int start, int count) {
            int crc = 0xFFFF;

            for (int pos = start; pos < start + count; pos++) {
                crc ^= (int) buf[pos] & 0xFF;   // XOR byte into least sig. byte of crc

                for (int i = 8; i != 0; i--) {    // Loop over each bit
                    if ((crc & 0x0001) != 0) {      // If the LSB is set
                        crc >>= 1;                    // Shift right and XOR 0x8408
                        crc ^= 0x8408;
                    } else                            // Else LSB is not set
                        crc >>= 1;                    // Just shift right
                }
            }

            return new byte[]{(byte) ((crc ^ 0xFFFF) & 0xFF), (byte) (((crc ^ 0xFFFF) >> 8) & 0xFF)};
        }

        private static byte[] byteStuff(byte[] data) {
            ByteArrayOutputStream stuffedData = new ByteArrayOutputStream();
            for (byte dataByte : data) {
                if ((dataByte & 0xFF) < 0xFD) {
                    stuffedData.write(dataByte);
                } else {
                    stuffedData.write(0xFF);
                    stuffedData.write(0xFF - (dataByte & 0xFF));
                }
            }
            return stuffedData.toByteArray();
        }

        private static byte[] unByteStuff(byte[] rawData) {
            ByteArrayOutputStream unStuffedData = new ByteArrayOutputStream();
            boolean fInvertNext = false;
            for (byte dataByte : rawData) {
                if ((dataByte & 0xFF) == 0xFF) {
                    fInvertNext = true;
                    continue;
                }
                if (fInvertNext) {
                    unStuffedData.write(0xFF - (dataByte & 0xFF));
                    fInvertNext = false;
                } else {
                    unStuffedData.write(dataByte & 0xFF);
                }
            }
            return unStuffedData.toByteArray();
        }

        private static byte nCommand = 0x00;

        public static byte[] wrapCommand(byte[] data) {
            ByteArrayOutputStream rawData = new ByteArrayOutputStream();

            rawData.write(0x00);
            rawData.write(nCommand++);
            rawData.write(data, 0, data.length);

            byte[] FCS = CRC16(rawData.toByteArray(), 0, rawData.size());
            rawData.write(FCS, 0, FCS.length);

            byte[] tmpData = byteStuff(rawData.toByteArray());
            rawData = new ByteArrayOutputStream();
            rawData.write(0xFD);
            rawData.write(tmpData, 0, tmpData.length);
            rawData.write(0xFE);

            return rawData.toByteArray();
        }

        public static byte[] unwrapResponse(byte[] rawData) {
            rawData = unByteStuff(rawData);
            assert ((rawData[0] & 0xFF) == 0xFD);
            assert ((rawData[rawData.length - 1] & 0xFF) == 0xFE);
            byte[] FCS = CRC16(rawData, 1, rawData.length - 4);
            assert ((FCS[0] & 0xFF) == (rawData[rawData.length - 3] & 0xFF));
            assert ((FCS[1] & 0xFF) == (rawData[rawData.length - 2] & 0xFF));

            return Arrays.copyOfRange(rawData, 3, rawData.length - 3);
        }
    }

    public class InterfaceBluetooth {
        private BluetoothAdapter bluetoothAdapter = null;
        private BluetoothLeScanner bluetoothLeScanner = null;
        private BluetoothManager bluetoothManager = null;
        private BluetoothScanCallback  bluetoothScanCallback = null;
        private BroadcastReceiver broadcastReceiver = null;
        IntentFilter filterBluetoothStateChanged = null;
        //private BluetoothGatt gattClient;

        private BluetoothGattCharacteristic characteristicFIFO;
        private BluetoothGattCharacteristic characteristicCREDITS;

        final UUID SERVICE_UUID_SERIAL = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d701");
        final UUID CHARACTERISTIC_UUID_FIFO = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d703");
        final UUID CHARACTERISTIC_UUID_CREDITS = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d704");

        int finalMTU = 0;

        byte mLastCommandCode = 0x00;

        boolean mScanAfterBTEnable = false;

        private Callbacks mBluetoothCallbacks = null;

        InterfaceBluetooth() {
            bluetoothScanCallback = new BluetoothScanCallback();
            broadcastReceiver = new BroadcastReceiver() {
                @Override
                public void onReceive(Context context, Intent intent) {
                    Object action = intent.getAction();
                    if (action == BluetoothAdapter.ACTION_STATE_CHANGED) {
                        int state = intent.getIntExtra(BluetoothAdapter.EXTRA_STATE, BluetoothAdapter.ERROR);
                        switch (state) {
                            case BluetoothAdapter.STATE_ON: {
                                if (mScanAfterBTEnable) {
                                    mScanAfterBTEnable = false;
                                    StartScanner();
                                }
                            }
                            case BluetoothAdapter.STATE_OFF: {
                                //bluetoothAdapter.enable();
                            }
                        }
                    }
                }
            };
        }

        public void RegisterCallbacks(Callbacks callbacks) {
            this.mBluetoothCallbacks = callbacks;
        }

        private void StartScanner() {
            if (!bluetoothAdapter.isEnabled())
                return;
            if (bluetoothLeScanner == null)
                bluetoothLeScanner = bluetoothAdapter.getBluetoothLeScanner();

            ScanFilter.Builder builder = new ScanFilter.Builder();
                            /*switch(scanType) {
                                case SCAN_BY_NAME:
                                    builder.setDeviceName(data);
                                    break;
                            }*/
            Vector<ScanFilter> filter = new Vector<ScanFilter>();
            filter.add(builder.build());
            ScanSettings.Builder builderScanSettings = new ScanSettings.Builder();
            //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_BALANCED);
            //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_POWER);
            builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY);
            builderScanSettings.setReportDelay(0);
            bluetoothLeScanner.stopScan(bluetoothScanCallback);
            bluetoothLeScanner.flushPendingScanResults(bluetoothScanCallback);
            bluetoothLeScanner.startScan(filter, builderScanSettings.build(),bluetoothScanCallback);
            //bluetoothLeScanner.startScan(bluetoothScanCallback);
        }

        private void CheckPermissions() {
            if(Build.VERSION.SDK_INT > Build.VERSION_CODES.LOLLIPOP){
                int permissionCheck = mContext.checkSelfPermission("Manifest.permission.ACCESS_FINE_LOCATION");
                permissionCheck += mContext.checkSelfPermission("Manifest.permission.ACCESS_COARSE_LOCATION");
                if (permissionCheck != 0) {

                    ((Activity)mContext).requestPermissions(new String[]{Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION}, 1001); //Any number
                }
            }else{
                Log.d(TAG, "checkBTPermissions: No need to check permissions. SDK version < LOLLIPOP.");
            }
        }

        private void Initialize() {
            CheckPermissions();
            bluetoothManager = (BluetoothManager)mContext.getSystemService(BLUETOOTH_SERVICE);
            bluetoothAdapter = BluetoothAdapter.getDefaultAdapter();//bluetoothManager.getAdapter();

            //bluetoothScanCallback = new BluetoothScanCallback();
            filterBluetoothStateChanged = new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED);
            ((Activity)mContext).registerReceiver(broadcastReceiver, filterBluetoothStateChanged);
        }

        public boolean TurnRadioOn() {
            if (bluetoothAdapter.isEnabled())
                return true;

            mScanAfterBTEnable = true;

            Intent enableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            //startActivityForResult(enableIntent, REQUEST_ENABLE_BLUETOOTH);
            //Intent discoverableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_DISCOVERABLE);
            //discoverableIntent.putExtra(BluetoothAdapter.EXTRA_DISCOVERABLE_DURATION, 60);
            //startActivity(discoverableIntent);
            ((Activity)mContext).startActivityForResult(enableIntent, 0);

            return false;
        }

        public void TurnRadioOff() {
            if (bluetoothAdapter.isEnabled()) {
                bluetoothAdapter.disable();
            }
        }

        public void StartDiscovery() {
            Log.i(TAG,"startScan()");
            _bleDevices.clear();
            if (bluetoothManager == null) {
                Initialize();
                //((Activity)mContext).registerReceiver(broadcastReceiver, filterBluetoothStateChanged);
                //bluetoothAdapter.disable();
            } else {
                StopDiscovery();
                //((Activity)mContext).registerReceiver(broadcastReceiver, filterBluetoothStateChanged);
            }
            if (TurnRadioOn())
                StartScanner();
        }

        public void StopDiscovery() {
            //Disconnect();
            if (bluetoothLeScanner != null) {
                try {
                    ((Activity) mContext).unregisterReceiver(broadcastReceiver);
                } catch (Exception e) {};
                bluetoothLeScanner.stopScan(bluetoothScanCallback); // stop scan
                bluetoothLeScanner.flushPendingScanResults(bluetoothScanCallback);
                //bluetoothLeScanner = null;
            }
            //bluetoothManager = null;
            //bluetoothAdapter = null;
            //filterBluetoothStateChanged = null;
        }

        private boolean Connect(String readerName) {
            if (mConnection != null) {
                mLastError = UEM_ERROR_READER_ALREADY_CONNECTED;
                return false;
            }
            if (!_bleDevices.containsKey(readerName)) {
                mLastError = UEM_ERROR_READER_INCORRECT_NAME;
                return false;
            }

            return ConnectDevice(_bleDevices.get(readerName));
        }

        // BLUETOOTH CONNECTION
        private boolean ConnectDevice(BluetoothDevice device) {
            if (device == null) Log.i(TAG,"Device is null");
            GattClientCallback gattClientCallback = new GattClientCallback();
            //gattClient = device.connectGatt(this,false,gattClientCallback);

            if (((BluetoothGatt)mConnection) != null && bluetoothManager.getConnectedDevices(BluetoothProfile.GATT).size() > 0)
                return false;

            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                mConnection = device.connectGatt(mContext, false, gattClientCallback, TRANSPORT_LE,PHY_LE_CODED_MASK ^ PHY_LE_1M_MASK);
            } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                mConnection = device.connectGatt(mContext, false, gattClientCallback, TRANSPORT_LE);
            } else {
                mConnection = device.connectGatt(mContext, false, gattClientCallback);
            }

            mInterfaceType = UEM_INTERFACE_BLUETOOTH;

            return true;
        }

        private boolean Disconnect() {
            if (mConnection == null)
                return false;

            ((BluetoothGatt)mConnection).disconnect();
            ((BluetoothGatt)mConnection).close();
            mConnection = null;
            mInterfaceType = null;

            return true;
        }

        private HashMap<String, BluetoothDevice> _bleDevices = new HashMap<String, BluetoothDevice>();

        // BLE Scan Callbacks
        private class BluetoothScanCallback extends ScanCallback {

            @Override
            public void onScanResult(int callbackType, ScanResult result) {
                Log.i(TAG, "onScanResult");
                if (result.getDevice().getName() != null){
                    if (result.getDevice().getName().contains("ANNA-") || result.getDevice().getName().contains("uEM Reader")) {
                        if(!_bleDevices.containsKey(result.getDevice().getName())) {
                            _bleDevices.put(result.getDevice().getName(), result.getDevice());
                            /*ListView listBLE = findViewById(R.id.bleList);
                            _listBLE.add(result.getDevice().getName());
                            ((ArrayAdapter<String>) listBLE.getAdapter()).notifyDataSetChanged();*/
                            if (mBluetoothCallbacks != null)
                                mBluetoothCallbacks.onBluetoothReaderFound(result.getDevice().getName());
                        }
                        // When find your device, connect.
                        //connectDevice(result.getDevice());
                        //bluetoothLeScanner.stopScan(bluetoothScanCallback); // stop scan
                    }
                }
            }

            @Override
            public void onBatchScanResults(List<ScanResult> results) {
                Log.i(TAG, "onBathScanResults");

                //bluetoothLeScanner.startScan(bluetoothScanCallback);
            }

            @Override
            public void onScanFailed(int errorCode) {
                Log.i(TAG, "ErrorCode: " + errorCode);
            }
        }

        // Bluetooth GATT Client Callback
        private class GattClientCallback extends BluetoothGattCallback {
            @Override
            public void onMtuChanged(BluetoothGatt gatt, int mtu, int status) {
                super.onMtuChanged(gatt, mtu, status);
                finalMTU = mtu;
                gatt.discoverServices();
            }

            @Override
            public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
                super.onConnectionStateChange(gatt, status, newState);
                Log.i(TAG,"onConnectionStateChange");

                if (status == BluetoothGatt.GATT_FAILURE) {
                    Log.i(TAG, "onConnectionStateChange GATT FAILURE");
                    return;
                } else if (status != BluetoothGatt.GATT_SUCCESS) {
                    /*final int s = status;
                    ((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "Failed to change connection state... "+String.valueOf(s), Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                    Log.i(TAG, "onConnectionStateChange != GATT_SUCCESS");
                    return;
                }

                if (newState == BluetoothProfile.STATE_CONNECTED) {
                    Log.i(TAG, "onConnectionStateChange CONNECTED");
                    /*((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "BLE Device connected!", Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                    gatt.requestMtu(512);
                } else if (newState == BluetoothProfile.STATE_DISCONNECTED) {
                    Log.i(TAG, "onConnectionStateChange DISCONNECTED");
                    /*((Activity)mContext).runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            Toast.makeText(mContext,
                                    "BLE Device disconnected!", Toast.LENGTH_LONG)
                                    .show();
                        }});*/
                }
            }

            @Override
            public void onServicesDiscovered(BluetoothGatt gatt, int status) {
                super.onServicesDiscovered(gatt, status);
                Log.i(TAG,"onServicesDiscovered");
                if (status != BluetoothGatt.GATT_SUCCESS) return;

                // Reference your UUIDs
                characteristicFIFO = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_FIFO);
                characteristicCREDITS = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_CREDITS);

                String UUID_MASK = "0000%s-0000-1000-8000-00805f9b34fb";
                String CLIENT_CHARACTERISTIC_CONFIG = String.format(UUID_MASK, "2902");

                if ((characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_NOTIFY) > 0) {
                    BluetoothGattDescriptor descriptor = characteristicFIFO.getDescriptor(UUID.fromString(CLIENT_CHARACTERISTIC_CONFIG));
                    if (descriptor == null || gatt == null) return;

                    // Set write type to WRITE_TYPE_DEFAULT to fix an issue seen on some phone models and
                    // Android versions that enable notification with a write_command message instead of a
                    // write_request message (as specified by the Bluetooth specification).
                    characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT);
                    descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
                    gatt.setCharacteristicNotification(characteristicFIFO, true);
                    gatt.writeDescriptor(descriptor);
                }
            }

            @Override
            public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                super.onCharacteristicRead(gatt, characteristic, status);
                Log.i(TAG,"onCharacteristicRead");
            }

            @Override
            public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                super.onCharacteristicWrite(gatt, characteristic, status);
                Log.i(TAG,"onCharacteristicWrite");
            }

            @Override
            public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
                super.onCharacteristicChanged(gatt, characteristic);
                Log.i(TAG,"onCharacteristicChanged");
                // Here you can read the characteristc's value
                /*final BluetoothGattCharacteristic ch = characteristic;
                ((Activity)mContext).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        TextView textView = (TextView) findViewById(R.id.outtext);
                        String out = new String(ch.getValue());
                        try {
                            out = bytesToHex(unwrapResponse(ch.getValue()));
                        } catch (Exception e) {
                            System.out.println("Error " + e.getMessage());
                            return;
                        }
                        textView.setText(out);
                    }
                });*/

                if (characteristic != characteristicFIFO) {
                    return;
                }

                mLastError = UEM_STATUS_SUCCESS;

                byte[] response = UemReader.Helper.unwrapResponse(characteristic.getValue());

                if (response == null || response.length <= 0) {
                    mLastError = UEM_ERROR_READER_RESPONSE_FAILURE;
                }

                if (response[0] != mLastCommandCode) {
                    mLastError = UEM_ERROR_READER_INCORRECT_RESPONSE;
                }

                response = Arrays.copyOfRange(response, 1, response.length);

                mBluetoothCallbacks.onBluetoothResponse(response);
            }

            @Override
            public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorRead(gatt, descriptor, status);
                Log.i(TAG,"onDescriptorRead");
            }

            @Override
            public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorWrite(gatt, descriptor, status);
                Log.i(TAG,"onDescriptorWrite");

                mBluetoothCallbacks.onBluetoothReaderConnected();
            }

        }

        private boolean SendCommand(byte[] sendBuffer) {
            if (sendBuffer == null)
                throw new NullPointerException("Parameter is null!");

            if (sendBuffer.length == 0)
                throw new IllegalArgumentException("Incorrect parameter!");

            if (((BluetoothGatt)mConnection) == null) {
                mLastError = UEM_ERROR_READER_NOT_CONNECTED;
                return false;
            }

            if (characteristicFIFO == null || (characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) <= 0) {
                mLastError = UEM_ERROR_READER_CONNECTION_FAILED;
                return false;
            }

            mLastCommandCode = sendBuffer[0];

            byte[] rawCommand = UemReader.Helper.wrapCommand(sendBuffer);

            characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
            characteristicFIFO.setValue(rawCommand);
            ((BluetoothGatt)mConnection).writeCharacteristic(characteristicFIFO);

            return true;
        }
    }

    public InterfaceBluetooth Bluetooth = new InterfaceBluetooth();
}
