// Transfer.cpp: implementation of the CTransfer class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Transfer.h"
#include "MfErrNo.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
//                      class CTransceiveBuffer 
//////////////////////////////////////////////////////////////////////

CTransceiveBuffer::CTransceiveBuffer()
{
    m_Count = 0;
}

CTransceiveBuffer::~CTransceiveBuffer()
{
}

bool CTransceiveBuffer::AddByte( unsigned char Value )
{
    if( m_Count + 1 > TX_BUF_SIZE )
        return false;

    m_Buf[ m_Count ] = Value;
    m_Count++;

    return true;
}

bool CTransceiveBuffer::AddBulk( const unsigned char *Data, DWORD Size )
{
    bool bRet = false;

    if( m_Count + Size > TX_BUF_SIZE )
        Size = TX_BUF_SIZE - m_Count;
    else
        bRet = true;

    memcpy( &m_Buf[ m_Count ], Data, Size );
    m_Count += Size;

    return bRet;
}

bool CTransceiveBuffer::GetBulk( int nIndex, unsigned char *Data, DWORD *pSize )
{
    bool bRet = false;
    DWORD CpySize = *pSize;

    if( nIndex + *pSize > m_Count )
        CpySize = TX_BUF_SIZE - nIndex;

    if( CpySize && CpySize == *pSize )
        bRet = true;

    if( CpySize )
        memcpy( Data, &m_Buf[ nIndex ], CpySize );

    *pSize = CpySize;
    return bRet;
}

BYTE CTransceiveBuffer::operator []( int nIndex ) const
{
    if( nIndex > TX_BUF_SIZE )
        return 0;
    else
        return m_Buf[ nIndex ];
}

//static CTransfer* lpTr;

// 10:10:50.646 R     00 B6 AB F4 B9 49 4E 63 2A 10 B4 8E
//                 FD 00 B6 AB F4 B9 49 4E 63 2A 10 B4 8E 00 FE 
//////////////////////////////////////////////////////////////////////
// Construction/Destruction     class CTransfer
//////////////////////////////////////////////////////////////////////
CTransfer::CTransfer()
{
//#if ADDRESS_RS_485
    m_IO_Addr = 0;
//#endif
    m_Cmd_ID = 0;
    m_Timeout = DEFAULT_IO_TIMEOUT;
    m_PortNumber = -1;
    m_Baudrate = 0;
}

CTransfer::~CTransfer()
{
    Close();
}

bool CTransfer::Open( DWORD LogLevel, 
                      int Port, 
                      int Baud )
{
    RX_Reset();
    //UART_RX_ENABLE();
    bool bResult = false;
    m_PortNumber = Port;    // from 0
    m_Baudrate = Baud;      // 0 -> USB, else -> COM-port

    if( m_Baudrate )
    {
        m_ComPort.SetLogLevel( LogLevel );

	    bResult = m_ComPort.Open( Port, 
                                  Baud,
                                  8,//8,
                                  'N',//'e',// e -; o -
                                  "1" );//1
        if( bResult )
            m_EvIOComplete = CreateEvent( NULL, true, false, NULL );
    }
    else
    {
        m_USB.SetLogLevel( LogLevel );
        bResult = m_USB.Open( Port );
    }

    return bResult;
}

bool CTransfer::Close( void )
{
    bool bResult = false;

/*
    else
    {
        int b = 0;
        int a = 1 / b;
    }
*/
    if( m_ComPort.IsOpened() )
    {
        bResult = m_ComPort.Close();
        CloseHandle( m_EvIOComplete );
    }

    if( m_USB.IsOpened() )
        bResult = m_USB.Close();

    return bResult;
}

void CTransfer::RX_Reset( void )
{
    m_state.n = 0;  //  .. m_state.b.RX_stuff = 0;
    m_RX_cnt = 0;
    m_RX_FCS = 0xFFFF;
}

void CTransfer::CalculateFCS( unsigned short &FCS, unsigned char b )
{
    unsigned char i;
    unsigned short w;

    w = ( b ^ FCS ) & 0xFF;
    i = 8;
    do{
        if( w & 1 ) 
        {
            w >>= 1;
            w ^= 0x8408;
        } 
        else 
        {
            w >>= 1;
        }
    } while( --i );
    FCS = w ^ ( FCS >> 8 );
}

void CTransfer::PostByte( unsigned char b )
{
    CalculateFCS( m_TX_FCS, b );
    ByteStuff( b );
}

void CTransfer::ByteStuff( unsigned char b )
{
    if( b < B_SPECIAL )
        m_TX_buf[m_TX_cnt++] = b;
    else                            // ByteStuffing
    {      
        m_TX_buf[m_TX_cnt++] = B_STUFF;
        m_TX_buf[m_TX_cnt++] = B_STUFF - b;
    }
}

char CTransfer::OnReceive( unsigned char inp )
{
    unsigned char b;
/*
	if( m_Timer )
	{
		KillTimer( NULL, m_Timer );
		m_Timer = 0;
	}
*/
    switch( inp )
    {
    case B_START:  // Frame begin
        RX_Reset();
        m_state.b.RX_start = 1;
        break;

    case B_STOP:  // Frame end
/*
         
   ,  ,    FCS. 
       . 
          
  ,     .
*/
        if( m_state.b.RX_start 
            && ( m_RX_cnt >= 4 ))// (1),  (1)  FCS(2) 
        {
            m_RX_FCS ^= 0xFFFF;
            if( m_RX_FCS == 0x0F47 )
            {
                return 0;
            }
            else
            {
/*
 NACK 1     FCS.   
  -      .
*/
//                Acknowledgment( NACK1_FCS_ERROR );
                return 2;
            }
        }
        else
        {
            RX_Reset();
            return 3;
        }
        break;

    case B_STUFF:  // Byte stuffing
        if( m_state.b.RX_start )
            m_state.b.RX_stuff = 1;
        break;

    default:    // Usual bytes
        if( ! m_state.b.RX_start )
            return 4;

        if( m_state.b.RX_stuff )
        {
            if( inp > 2 )       //  
            {
                RX_Reset();
                return 5;
            }
            m_state.b.RX_stuff = 0;
            b = B_STUFF - inp;
        }
        else
        {
            b = inp;
        }

//#if ADDRESS_RS_485
        if( m_state.b.RX_addr )
        {
            m_RX_buf[m_RX_cnt++] = b;
        }
        else
        {
            if( m_IO_Addr && m_IO_Addr != b )
            {
                RX_Reset();
                return 7;           //     
            }
            m_state.b.RX_addr = 1;
        }
//#else
//            m_RX_buf[m_RX_cnt++] = b;
//#endif
/*
         
        .
*/
        if( m_RX_cnt < RX_BUF_SIZE )
        {
            CalculateFCS( m_RX_FCS, b );
        }
        else
        {
            RX_Reset();
            return 6;
        }

    }
    return 1;
}

bool CTransfer::OnRX( int Port, BYTE *RXBuff, int Size )
{
    char Error;
    for( int i = 0; i < Size; i++ )
    {
        Error = OnReceive( RXBuff[i] );
        if( Error > 1 )
        {
//            if( m_lpfnError ) 
//                (*m_lpfnError)( Port, SERN_EVENT, Error, m_NotifContext );
            OnError( Port, 1, Error );
            return true;
        }
    }
    if( ! Error )
    {
//        if( m_lpfnRX ) 
//            (*m_lpfnRX)( Port, m_RX_buf, m_RX_cnt - 2, m_NotifContext );
        SetEvent( m_EvIOComplete );
//        m_state.b.RX_OK = 1;
        return true;
    }
    else
        return false;
}

void CTransfer::OnError( int Port, UINT Oper, DWORD Err )
{
//    if( m_lpfnError )
//        (*m_lpfnError)( Port, Oper, Err, m_NotifContext );
//        str.Format( " COM%d   OK", Port+1 );
//	CString str, so, se;
    if( !( (Oper == 1) && (Err == 1) ) )
    {
/*
        switch( Oper )
        {
        case SERN_EVENT:
            so = "EVENT";
            break;

        case SERN_READ:
            so = "READ";
            break;

        case SERN_WRITE:
            so = "WRITE";
            break;

        default:
            so = "NO";
        }

        switch( Err )
        {
        case CE_RXOVER:
            se = "CE_RXOVER";
            break;

        case CE_OVERRUN:
            se = "CE_OVERRUN";
            break;

        case CE_RXPARITY:
            se = "CE_RXPARITY";
            break;

        case CE_FRAME:
            se = "CE_FRAME";
            break;

        case CE_TXFULL:
            se = "CE_TXFULL";
            break;

        case CE_MODE:
            se = "CE_MODE";
            break;

        case ERROR_TIMEOUT:
            se = "ERROR_TIMEOUT";
            break;

        default:
            se.Format( "0x%08X", Err );
        }
//        str.Format( " COM%d   %s       %s", Port+1, so, se );
*/
        m_Error = Err;
//        m_state.b.RX_Err = 1;
    }
//    CCardPage* pPage = (CCardPage*)GetActivePage();
//    pPage->m_StatusBar.SetWindowText( str );
//    m_pCurPage->m_StatusBar.SetWindowText( str );
}
/*
//void CTransfer::SendCommand( unsigned char Oper, unsigned char Size, unsigned char *Param )
void CTransfer::Transmit( unsigned char *Data, int Size )
{
//    m_Cmd_op = Data[0];
    m_TX_cnt = 0;
    m_TX_FCS = 0xffff;
    m_TX_buf[m_TX_cnt++] = B_START;  // Frame begin
    PostByte( m_Cmd_ID++ );
    for( int i = 0; i < Size; i++ )
        PostByte( Data[i] );
    m_TX_FCS ^= 0xffff;
    ByteStuff( m_TX_FCS & 0xFF );
    ByteStuff( ( m_TX_FCS >> 8 ) & 0xFF );
    m_TX_buf[m_TX_cnt++] = B_STOP;  // Frame end
    m_ComPort.Write( m_TX_buf, m_TX_cnt );

}
*/
//////////////////////////////////////////////////////////////////////
LONG CTransfer::SingleTransceive()
{
    m_Error = 0;
    LONG Status = 0;

    if( ! IsOpened() )
    {
        Status = SCARD_E_NO_READERS_AVAILABLE;
        return Status;
    }

    if( m_TrBuf.GetLength() == 0 )
    {
        Status = SCARD_E_NOT_TRANSACTED;
//        memcpy( pbRecvBuffer, &Status, *pcbRecvLength );
        return Status;
    }

    m_RcBuf.Clear();

    if( m_Baudrate )
    {
        m_TX_cnt = 0;
        m_TX_FCS = 0xffff;
        m_TX_buf[m_TX_cnt++] = B_START;  // Frame begin
//#if ADDRESS_RS_485
        PostByte( m_IO_Addr );
//#endif
        PostByte( m_Cmd_ID++ );
        DWORD Count = m_TrBuf.GetLength();
        for( DWORD i = 0; i < m_TrBuf.GetLength(); i++ )
            PostByte( m_TrBuf[i] );
        m_TX_FCS ^= 0xffff;
        ByteStuff( m_TX_FCS & 0xFF );
        ByteStuff( ( m_TX_FCS >> 8 ) & 0xFF );
        m_TX_buf[m_TX_cnt++] = B_STOP;  // Frame end

        m_ComPort.Write( m_TX_buf, m_TX_cnt );


        DWORD Wait_Result = WaitForSingleObject( m_EvIOComplete, m_Timeout );
        switch( Wait_Result )
        {
        case WAIT_ABANDONED:
            m_Error = ERROR_INVALID_HANDLE_STATE;   // == 0x00000649
            break;

        case WAIT_OBJECT_0:
            ResetEvent( m_EvIOComplete );
            break;

        case WAIT_TIMEOUT:
            m_Error = ERROR_TIMEOUT;                // == 0x000005B4
            break;

        }

/*
        m_Error = ERROR_TIMEOUT;

        BYTE RXBuff[300];

        while( m_ComPort.Read( RXBuff, &Count ) && Count == 1 )
        {
            if( OnRX( m_PortNumber, RXBuff, Count ) )
                break;
        }
*/
        if( m_Error )               // ( COM-port & Protocol ) Errors (4 bytes)
        {
            Status = m_Error;
        } 
        else
        {
            if( m_RX_buf[2] )       // ( Reader & Card ) Errors (1 byte)
            {
                Status = SCARD_F_INTERNAL_ERROR;
                m_RcBuf.AddByte( m_RX_buf[2] );
            }
            else
            {
                m_RcBuf.AddBulk( &m_RX_buf[2], m_RX_cnt - 4 );
            }
        }
    }
    else
    {
        m_TX_cnt = (BYTE)m_TrBuf.GetLength();
        DWORD Count = m_TX_cnt;
        m_TrBuf.GetBulk( 0, m_TX_buf, &Count );

        m_Error = m_USB.Transfer( m_TX_buf, m_TX_cnt, m_RX_buf, RX_BUF_SIZE, &m_RX_cnt );

        if( m_Error )               // ( COM-port & Protocol ) Errors (4 bytes)
        {
            Status = m_Error;
        } 
        else
        {
            if( m_RX_buf[0] )       // ( Reader & Card ) Errors (1 byte)
            {
                Status = SCARD_F_INTERNAL_ERROR;
                m_RcBuf.AddByte( m_RX_buf[0] );
            }
            else
            {
                m_RcBuf.AddBulk( m_RX_buf, m_RX_cnt );
            }
        }
    }
   

    return Status;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void CTransfer::SetTimeout( DWORD NewTimeout )
{
    if( NewTimeout )
        m_Timeout = NewTimeout; 
    else
        m_Timeout = DEFAULT_IO_TIMEOUT; 
}