// SAM_MifareClassic.cpp:      .
//

#include "stdafx.h"
#include "..\..\Clscrfl.h"
#include "..\..\Clscrfl_SAM.h"

LPVOID	g_pReader = NULL;
DWORD	g_dwActivationPause = 10;
DWORD	dwCyclePause = 1000; 
BYTE	g_UID[10] = { 0 };
DWORD	g_dwUIDLength = sizeof(g_UID);
BYTE	g_ucSectorNumber = 0x09;
BYTE	g_ucBlockNumber = 0x00;
BYTE	g_ucSAMKeyNumber = 0x11;
BYTE	g_ucSAMKeyVersion = 0x00;
BYTE	g_WriteData[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };

//   SAM-   g_ucSAMKeyNumber   g_ucSAMKeyVersion      PICC / Mifare
//  A0A1A2A3A4A50000B0B1B2B3B4B50000,  A0A1A2A3A4A5 -  A, B0B1B2B3B4B5 -  B.

LONG PrepareReader()
{
	LONG Status;

	Status = CLSCRF_Create(&g_pReader);
	if (Status != SCARD_S_SUCCESS) return Status;

	Status = CLSCRF_OpenUSB(g_pReader);
	if (Status != SCARD_S_SUCCESS) return Status;

	Status = CLSCRF_Mfrc_On(g_pReader);
	if (Status != SCARD_S_SUCCESS) return Status;

	DWORD dwState = 0;

	Status = CLSCRF_GetState(g_pReader, &dwState);
	if (Status != SCARD_S_SUCCESS) return Status;

	if ((dwState & 0xFFUL) != 0x00)
	{
		Status = CLSCRF_Mfrc_Set_Rf_Mode(g_pReader, 0x00);
		if (Status != SCARD_S_SUCCESS) return Status;
	}

	return SCARD_S_SUCCESS;
}

LONG FinishWork()
{
	LONG Status = SCARD_S_SUCCESS;

	Status = CLSCRF_Mfrc_Off(g_pReader);
	if (Status != SCARD_S_SUCCESS) return Status;

	Status = CLSCRF_Close(g_pReader);
	if (Status != SCARD_S_SUCCESS) return Status;

	Status = CLSCRF_Destroy(&g_pReader);

	return SCARD_S_SUCCESS;
}

LONG ActivateCard()
{
	LONG Status = SCARD_S_SUCCESS;

	Status = CLSCRF_Mfrc_Rf_Off_On(g_pReader, 10);
	if (Status != SCARD_S_SUCCESS) return Status;

	Sleep(g_dwActivationPause);

	BYTE ATQ[2] = { 0 }, SAK = 0;

	Status = CLSCRF_Activate_Idle_A(g_pReader, ATQ, &SAK, g_UID, &g_dwUIDLength);
	if (Status != SCARD_S_SUCCESS) return Status;

	return SCARD_S_SUCCESS;
}

LONG AuthenticateMifareClassic()
{
	BYTE ucSW1, ucSW2;
	BYTE PCDcap2[6] = { 0 };
	BYTE Div[256] = { 0 };

	printf(("\n"));

	printf(("CLSCRF_SAM_MifareAuthenticate...\t"));
	LONG Status = CLSCRF_SAM_MifareAuthenticate(
		g_pReader,
		0x00,
		0x01,
		&g_UID[g_dwUIDLength - 4],
		g_ucSAMKeyNumber,
		g_ucSAMKeyVersion,
		0x0A,
		g_ucSectorNumber,
		&ucSW1,
		&ucSW2
		);
	printf(("0x%08X\n"), Status);

	if (Status != SCARD_S_SUCCESS)
	{
		printf(("\tAuthentication failed.\t\t"));

		return SCARD_E_UNEXPECTED;
	}

	printf(("\tAuthentication successful!\t"));

	return Status;
}

LONG ReadMifareClassic()
{
	BYTE ucSW1, ucSW2;
	BYTE Data[64] = { 0 };

	LONG Status = CLSCRF_SAM_MifareRead(
		g_pReader,
		0x00,
		(DWORD)g_ucSectorNumber,
		(DWORD)g_ucBlockNumber,
		Data,
		&ucSW1,
		&ucSW2
		);

	if (Status != SCARD_S_SUCCESS)
		return Status;

	printf("Data = ");
	for (int i = 0; i < 16; i++)
		printf("%02X", Data[i]);
	printf("\n");

	return Status;
}

LONG WriteMifareClassic()
{
	BYTE ucSW1, ucSW2;

	LONG Status = CLSCRF_SAM_MifareWrite(
		g_pReader,
		0x00,
		(DWORD)g_ucSectorNumber,
		(DWORD)g_ucBlockNumber,
		g_WriteData,
		&ucSW1,
		&ucSW2
		);

	return Status;
}

int _tmain(int argc, _TCHAR* argv[])
{
	LONG Status = SCARD_S_SUCCESS;

	printf(("Preparing reader...\t\t\t"));
	Status = PrepareReader();
	printf(("0x%08X\n"), Status);

	if (Status != SCARD_S_SUCCESS)
	{
		FinishWork();

		return 0;
	}

	printf(("\nInsert a card...\t\t\t"));
	while (true)
	{
		Status = ActivateCard();

		if (Status != SCARD_S_SUCCESS)
		{
			Sleep(dwCyclePause);

			continue;
		}

		printf(("Card found!\n"));

		printf(("Resetting SAM...\t\t\t"));
		Status = CLSCRF_SAM_Reset(g_pReader);
		printf(("0x%08X\n"), Status);

		if (Status != SCARD_S_SUCCESS) 
			return Status;

		printf(("Authenticating MifareClassic...\t"));
		Status = AuthenticateMifareClassic();
		printf(("0x%08X\n"), Status);

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\nInsert a card...\t\t\t"));

			continue;
		}

		printf(("Reading MifareClassic...\t\t"));
		Status = ReadMifareClassic();
		printf(("ReadMifareClassic:\t\t\t0x%08X\n"), Status);

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\nInsert a card...\t\t\t"));

			continue;
		}

		printf(("Writing MifareClassic...\t\t"));
		Status = WriteMifareClassic();
		printf(("0x%08X\n"), Status);

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\nInsert a card...\t\t\t"));

			continue;
		}

		printf(("Waiting %ums\n"), dwCyclePause);
		Sleep(dwCyclePause);

		printf(("\n\nInsert a card...\t\t\t"));
	}

	FinishWork();

	return 0;
}

