/*
 * ru_microem_clscrfl_ClscrflJNI.cpp
 *
 *  Created on: 13 июл. 2021 г.
 *      Author: thermalview
 */

//#define JNICALL
//#define jlong long

#include "Clscrfl.h"
#include "Clscrfl_SAM.h"
#include <jni.h>
//#include "ru_microem_clscrfl_ClscrflJNI.h"
#include <iostream>
#include <unistd.h>

#ifdef __cplusplus
extern "C" {
#endif

static bool fVerboseOutput = false;

JNIEXPORT void JNICALL Java_ru_microem_clscrfl_Clscrfl_jniSetVerboseOutput
  (JNIEnv * env, jobject thisObject, jboolean enabled)
{
	fVerboseOutput = enabled;
}


JNIEXPORT jlong JNICALL Java_ru_microem_clscrfl_Clscrfl_jniCreateReader
  (JNIEnv * env, jobject thisObject)
{
//	if (fVerboseOutput)
//	{
//		std::cout << "Clscrfl: CLSCRF_Create" << std::endl;
//	}

	PVOID pReader = NULL;

	CLSCRF_Create( &pReader );

	return (long)pReader;
}

JNIEXPORT void JNICALL Java_ru_microem_clscrfl_Clscrfl_jniDestroyReader
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	CLSCRF_Destroy( (LPVOID*)&reader );
}


JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniOpenUSB
  (JNIEnv * env, jobject thisObject, jlong reader, jint index)
{
	int res = CLSCRF_OpenUSB((LPVOID)reader, (DWORD)index, 0);

	return res;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniClose
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	int res = CLSCRF_Close((LPVOID)reader);

	return res;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniGetLastInternalError
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	BYTE err = 0;
	CLSCRF_GetLastInternalError((LPVOID)reader, &err);

	return (int)(err & 0xFF);
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniGetState
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	DWORD dwState = 0;
	int res = CLSCRF_GetState((LPVOID)reader, &dwState);
	if (res != 0)
	{
		return res;
	}
	return (int)dwState;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniRadioChipPowerUp
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	int res = CLSCRF_Mfrc_On((LPVOID)reader);

	return res;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniRadioChipPowerDown
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	int res = CLSCRF_Mfrc_Off((LPVOID)reader);

	return res;
}

JNIEXPORT jobject JNICALL Java_ru_microem_clscrfl_Clscrfl_jniGetVersion
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	BYTE Version[6] = {0};
	//std::cout << "1" << std::endl;
    jclass clazz = env->FindClass("ru/microem/clscrfl/Clscrfl$ReaderVersion");
    //std::cout << "2" << std::endl;
    jmethodID constructor = env->GetMethodID(clazz, "<init>", "(Lru/microem/clscrfl/Clscrfl;)V");
    usleep(1); // If not, we get SIGSEGV on -o > 0
    //std::cout << "3" << std::endl;
    jobject obj = env->NewObject(clazz, constructor);
    //std::cout << "4" << std::endl;
    jfieldID generalErrorField = env->GetFieldID(clazz, "generalError", "I");
    //std::cout << "5" << std::endl;
    jfieldID innerErrorField = env->GetFieldID(clazz, "innerError", "I");
    //std::cout << "6" << std::endl;
    jfieldID versionField = env->GetFieldID(clazz, "version", "[B");
    //std::cout << "7" << std::endl;
	long res = CLSCRF_Get_Mfrc_Version((LPVOID)reader, Version);
	//std::cout << "8" << std::endl;
	if (res == 0)
	{
		env->SetIntField(obj, generalErrorField, 0);
		//std::cout << "9" << std::endl;
		env->SetIntField(obj, innerErrorField, 0);
		//std::cout << "10" << std::endl;
		jbyteArray jniVersion = env->NewByteArray(6);
		//std::cout << "11" << std::endl;
		env->SetByteArrayRegion(jniVersion, 0, 6, (jbyte*)Version);
		//std::cout << "12" << std::endl;
		env->SetObjectField(obj, versionField, jniVersion);
		//std::cout << "13" << std::endl;
	}
	else
	{
		env->SetIntField(obj, generalErrorField, res);
		//std::cout << "14" << std::endl;
		BYTE err = 0;
		CLSCRF_GetLastInternalError((LPVOID)reader, &err);
		//std::cout << "15" << std::endl;
		env->SetIntField(obj, innerErrorField, err & 0xFF);
		//std::cout << "16" << std::endl;
	}

    // return object
    return obj;
}

JNIEXPORT jobject JNICALL Java_ru_microem_clscrfl_Clscrfl_jniGetSerial
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	BYTE Serial[4] = {0};

    jclass clazz = env->FindClass("ru/microem/clscrfl/Clscrfl$ReaderSerial");

    jmethodID constructor = env->GetMethodID(clazz, "<init>", "(Lru/microem/clscrfl/Clscrfl;)V");
    usleep(1); // If not, we get SIGSEGV on -o > 0
    jobject obj = env->NewObject(clazz, constructor);

    jfieldID generalErrorField = env->GetFieldID(clazz, "generalError", "I");
    jfieldID innerErrorField = env->GetFieldID(clazz, "innerError", "I");
    jfieldID serialField = env->GetFieldID(clazz, "serial", "[B");

	long res = CLSCRF_Get_Mfrc_Serial_Number((LPVOID)reader, Serial);

	if (res == 0)
	{
		env->SetIntField(obj, generalErrorField, 0);
		env->SetIntField(obj, innerErrorField, 0);
		jbyteArray jniSerial = env->NewByteArray(4);
		env->SetByteArrayRegion(jniSerial, 0, 4, (jbyte*)Serial);
		env->SetObjectField(obj, serialField, jniSerial);
	}
	else
	{
		env->SetIntField(obj, generalErrorField, res);
		BYTE err = 0;
		CLSCRF_GetLastInternalError((LPVOID)reader, &err);
		env->SetIntField(obj, innerErrorField, err & 0xFF);
	}

    // return object
    return obj;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniResetRadioField
  (JNIEnv * env, jobject thisObject, jlong reader, jint delay)
{
	int res = CLSCRF_Mfrc_Rf_Off_On((LPVOID)reader, (USHORT)delay);

	return res;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniBeep
  (JNIEnv * env, jobject thisObject, jlong reader, jint count)
{
	int res = CLSCRF_Sound((LPVOID)reader, (BYTE)count);

	return res;
}


JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniLed
  (JNIEnv * env, jobject thisObject, jlong reader, jint color, jint count, jint postColor)
{
	int res = CLSCRF_Led((LPVOID)reader, (BYTE)color, (BYTE)count, (BYTE)postColor);

	return res;
}

JNIEXPORT jobject JNICALL Java_ru_microem_clscrfl_Clscrfl_jniActivateIdleA
  (JNIEnv * env, jobject thisObject, jlong reader)
{
	BYTE ATQ[2] = {0};
	BYTE SAK = 0;
	BYTE UID[10] = {10};
	DWORD dwUIDLength = 10;

    jclass clazz = env->FindClass("ru/microem/clscrfl/Clscrfl$ISO14443ACard");

    jmethodID constructor = env->GetMethodID(clazz, "<init>", "(Lru/microem/clscrfl/Clscrfl;)V");
    usleep(1); // If not, we get SIGSEGV on -o > 0
    jobject obj = env->NewObject(clazz, constructor);

    jfieldID generalErrorField = env->GetFieldID(clazz, "generalError", "I");
    jfieldID innerErrorField = env->GetFieldID(clazz, "innerError", "I");
    jfieldID uidField = env->GetFieldID(clazz, "UID", "[B");
    jfieldID atqField = env->GetFieldID(clazz, "ATQ", "[B");
    jfieldID sakField = env->GetFieldID(clazz, "SAK", "B");

	long res = CLSCRF_Activate_Idle_A((LPVOID)reader, ATQ, &SAK, UID, &dwUIDLength);

	if (res == 0)
	{
		env->SetIntField(obj, generalErrorField, 0);
		env->SetIntField(obj, innerErrorField, 0);
		jbyteArray jniUID = env->NewByteArray(dwUIDLength);
		env->SetByteArrayRegion(jniUID, 0, dwUIDLength, (jbyte*)UID);
		env->SetObjectField(obj, uidField, jniUID);
		jbyteArray jniATQ = env->NewByteArray(2);
		env->SetByteArrayRegion(jniATQ, 0, 2, (jbyte*)ATQ);
		env->SetObjectField(obj, atqField, jniATQ);
		env->SetByteField(obj, sakField, SAK);
	}
	else
	{
		env->SetIntField(obj, generalErrorField, res);
		BYTE err = 0;
		CLSCRF_GetLastInternalError((LPVOID)reader, &err);
		env->SetIntField(obj, innerErrorField, err & 0xFF);
	}

    // return object
    return obj;

	/*//printf("Trying to Find Class.\n");
    jclass clazz = env->FindClass("ru/microem/clscrfl/Clscrfl$ISO14443ACard");
    if (clazz == NULL) {
	   printf("Find Class Failed.\n");
	} else {
	   printf("Found class.\n");
	}
    printf("Trying to Find constructor %s.\n", "(Lru/microem/clscrfl/Clscrfl;)V");
    // Get the method id of an empty constructor in clazz
    jmethodID constructor = env->GetMethodID(clazz, "<init>", "(Lru/microem/clscrfl/Clscrfl;)V");//, "()V");//
    if (constructor == NULL) {
	   printf("Find constructor Failed.\n");
	} else {
	   printf("Found constructor.\n");
	}

    // Create an instance of clazz
    return env->NewObject(clazz, constructor);*/
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniMifareClassicLoadKey
  (JNIEnv * env, jobject thisObject, jlong reader, jint recordNumber, jint keyType, jbyteArray key)
{
    int len = env->GetArrayLength(key);
    if (len != 6)
    {
    	throw std::invalid_argument( "key len != 6" );
    	return -1;
    }
    BYTE buf[6] = {0};
    env->GetByteArrayRegion(key, 0, 6, reinterpret_cast<jbyte*>(buf));

	int res = CLSCRF_MifareStandard_WriteKeyToE2((LPVOID)reader, (BYTE)keyType, (BYTE)recordNumber, buf);

	return res;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniMifareClassicAuthenticate
  (JNIEnv * env, jobject thisObject, jlong reader, jint sectorIndex, jint recordNumber, jint keyType, jbyteArray UID)
{
    int len = env->GetArrayLength(UID);
    if (len < 4)
    {
    	throw std::invalid_argument( "UID len < 4" );
    	return -1;
    }
    BYTE buf[4] = {0};
    env->GetByteArrayRegion(UID, 0, 4, reinterpret_cast<jbyte*>(buf));

    int res = CLSCRF_MifareStandard_AuthE2Ex((LPVOID)reader, (BYTE)keyType, buf, (BYTE)recordNumber, (BYTE)sectorIndex);

	return res;
}

JNIEXPORT jobject JNICALL Java_ru_microem_clscrfl_Clscrfl_jniMifareClassicRead
  (JNIEnv * env, jobject thisObject, jlong reader, jint sectorIndex, jint blockIndex)
{
	BYTE Data[16] = {0};

    jclass clazz = env->FindClass("ru/microem/clscrfl/Clscrfl$MifareBlock");

    jmethodID constructor = env->GetMethodID(clazz, "<init>", "(Lru/microem/clscrfl/Clscrfl;)V");
    usleep(1); // If not, we get SIGSEGV on -o > 0
    jobject obj = env->NewObject(clazz, constructor);

    jfieldID generalErrorField = env->GetFieldID(clazz, "generalError", "I");
    jfieldID innerErrorField = env->GetFieldID(clazz, "innerError", "I");
    jfieldID dataField = env->GetFieldID(clazz, "data", "[B");

	long res = CLSCRF_MifareStandard_Read((LPVOID)reader, (BYTE)sectorIndex, (BYTE)blockIndex, Data);

	if (res == 0)
	{
		env->SetIntField(obj, generalErrorField, 0);
		env->SetIntField(obj, innerErrorField, 0);
		jbyteArray jniData = env->NewByteArray(16);
		env->SetByteArrayRegion(jniData, 0, 16, (jbyte*)Data);
		env->SetObjectField(obj, dataField, jniData);
	}
	else
	{
		env->SetIntField(obj, generalErrorField, res);
		BYTE err = 0;
		CLSCRF_GetLastInternalError((LPVOID)reader, &err);
		env->SetIntField(obj, innerErrorField, err & 0xFF);
	}

    // return object
    return obj;
}

JNIEXPORT jint JNICALL Java_ru_microem_clscrfl_Clscrfl_jniMifareClassicWrite
  (JNIEnv * env, jobject thisObject, jlong reader, jint sectorIndex, jint blockIndex, jbyteArray data)
{
    int len = env->GetArrayLength(data);
    if (len != 16)
    {
    	throw std::invalid_argument( "data len != 16" );
    	return -1;
    }
    BYTE buf[16] = {0};
    env->GetByteArrayRegion(data, 0, 16, reinterpret_cast<jbyte*>(buf));

	int res = CLSCRF_MifareStandard_Write((LPVOID)reader, (BYTE)sectorIndex, (BYTE)blockIndex, buf);

	return res;
}

#ifdef __cplusplus
}
#endif
