package ru.microem.clscrfl;

import java.io.File;
import java.net.URISyntaxException;
import java.security.CodeSource;

public class Clscrfl {
	protected long refReader = 0;
	protected boolean verboseOutput = false;
	
	private boolean isInterfaceOpened = false;
	
    public boolean isOpened() {
		return isInterfaceOpened;
	}

	public boolean isVerboseOutput() {
		return verboseOutput;
	}

	public void setVerboseOutput(boolean verboseOutput) {
		this.verboseOutput = verboseOutput;
		jniSetVerboseOutput(verboseOutput);
	}

	static {
    	CodeSource codeSource = Clscrfl.class.getProtectionDomain().getCodeSource();
    	try {
    		File jarFile = new File(codeSource.getLocation().toURI().getPath());
    		File parentDir = jarFile.getParentFile();
    		File soFile = new File(parentDir, "libClscrflJNI.so");
            //System.loadLibrary("Clscrfl");
    		String libPath = soFile.getPath();
    		//System.out.println("Loading libClscrflJNI: " + libPath);
    		System.load(libPath);
		} catch (URISyntaxException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }
    
    /*public static void main(String[] args) {
    	try {
			Clscrfl clscrfl1 = new Clscrfl();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
    }*/
    
	private static final char[] HEX_ARRAY = "0123456789ABCDEF".toCharArray();
	public static String bytesToHex(byte[] bytes) {
	    char[] hexChars = new char[bytes.length * 2];
	    for (int j = 0; j < bytes.length; j++) {
	        int v = bytes[j] & 0xFF;
	        hexChars[j * 2] = HEX_ARRAY[v >>> 4];
	        hexChars[j * 2 + 1] = HEX_ARRAY[v & 0x0F];
	    }
	    return new String(hexChars);
	}
	
	public static byte[] hexToBytes(String s) {
		if (s == null) {
			return null;
		}
	    int len = s.length();
	    if (len < 2) {
	    	return null;
	    }
	    byte[] data = new byte[(int)(len / 2)];
	    for (int i = 0; i < len; i += 2) {
	        data[(int)(i / 2)] = (byte) ((Character.digit(s.charAt(i), 16) << 4)
	                             + Character.digit(s.charAt(i+1), 16));
	    }
	    return data;
	}
	
	public class ClscrflException extends Exception {
	    /**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		private final int generalError;
		private final int innerError;

	    public ClscrflException(String message, int generalError, int innerError) {
	        super(message);
	        this.generalError = generalError;
	        this.innerError = innerError;
	    }

	    public int getGeneralError() {
	        return generalError;   
	    }
	    
	    public int getInnerError() {
	        return innerError;   
	    }
	    
	    public String getFullMessage() {
	    	return getMessage() + 
	    			" GeneralError=0x" + Integer.toHexString(getGeneralError()) + 
	    			" InnerError=0x" + Integer.toHexString(getInnerError());
	    }
	}
	
	public Clscrfl() throws ClscrflException {
		refReader = jniCreateReader();
		
		if (refReader == 0) {
			throw new ClscrflException("Couldn't create reader object", 0, 0);
		}
	}
	
	private void CheckStatus(int res) throws ClscrflException {
		if (res < 0 || res > 0x0000FFFF) {
			int err = jniGetLastInternalError(refReader);
			throw new ClscrflException("Error code returned!", (int) res, err);
		}
	}
	
	private native void jniSetVerboseOutput(boolean enabled);
	
	private native long jniCreateReader();
	
	private native void jniDestroyReader(long reader);
	
	private native int jniOpenUSB(long reader, int index);
	
	public void OpenUSB(int index) throws ClscrflException {
		int res = jniOpenUSB(refReader, index);
		if (res != 0) {
			throw new ClscrflException("Failed to open an interface", res, 0);
		} else {
			isInterfaceOpened = true;
		}
	}
	
	private native int jniClose(long reader);
	
	public void Close() throws ClscrflException {
		int res = jniClose(refReader);
		if (res != 0) {
			throw new ClscrflException("Couldn't create reader object", res, 0);
		} else {
			isInterfaceOpened = false;
		}
	}
	
	private native int jniGetLastInternalError(long reader);
	
	private native int jniGetState(long reader);
	
	public int getState() throws ClscrflException {
		int res = jniGetState(refReader);
		CheckStatus(res);
		return res;
	}
	
	private native int jniRadioChipPowerUp(long reader);
	
	public void radioChipPowerUp() throws ClscrflException {
		int res = jniRadioChipPowerUp(refReader);
		CheckStatus(res);
	}
	
	private native int jniRadioChipPowerDown(long reader);
	
	public void radioChipPowerDown() throws ClscrflException {
		int res = jniRadioChipPowerDown(refReader);
		CheckStatus(res);
	}
	
	public class ReaderVersion {
		public int generalError = 0;
		public int innerError = 0;
		public byte[] version = null;
	}
	
	private native ReaderVersion jniGetVersion(long reader);
	
	public byte[] getVersion() throws ClscrflException {
		ReaderVersion version = jniGetVersion(refReader);
		if (version.generalError != 0 && version.innerError != 0) {
			throw new ClscrflException("jniGetVersion: Error code returned!", version.generalError, version.innerError);
		}
		return version.version;
	}
	
	public class ReaderSerial {
		public int generalError = 0;
		public int innerError = 0;
		public byte[] serial = null;
	}
	
	private native ReaderSerial jniGetSerial(long reader);
	
	public byte[] getSerial() throws ClscrflException {
		ReaderSerial serial = jniGetSerial(refReader);
		if (serial.generalError != 0 && serial.innerError != 0) {
			throw new ClscrflException("jniGetSerial: Error code returned!", serial.generalError, serial.innerError);
		}
		return serial.serial;
	}
	
	private native int jniResetRadioField(long reader, int delay);
	
	public void resetRadioField(int delay) throws ClscrflException {
		int res = jniResetRadioField(refReader, delay);
		CheckStatus(res);
	}
	
	private native int jniBeep(long reader, int count);
	
	public void beep(int count) throws ClscrflException {
		int res = jniBeep(refReader, count);
		CheckStatus(res);
	}
	
	private native int jniLed(long reader, int color, int count, int postColor);
	
	public void led(int color, int count, int postColor) throws ClscrflException {
		int res = jniLed(refReader, color, count, postColor);
		CheckStatus(res);
	}
	
	public class ISO14443ACard {
		public int generalError = 0;
		public int innerError = 0;
		public byte[] UID = null;
		public byte[] ATQ = null;
		public byte SAK = 0;
	}
	
	private native ISO14443ACard jniActivateIdleA(long reader) ;
	
	public ISO14443ACard activateIdleA() throws ClscrflException {
		ISO14443ACard card = jniActivateIdleA(refReader);
		if (card.generalError != 0 && card.innerError != 0) {
			throw new ClscrflException("jniActivateIdleA: Error code returned!", card.generalError, card.innerError);
		}
		return card;
	}
	
	private native int jniMifareClassicLoadKey(long reader, int recordNumber, int keyType, byte[] key);
	
	public void mifareClassicLoadKeyA(int recordNumber, byte[] key) throws ClscrflException {
		int res = jniMifareClassicLoadKey(refReader, recordNumber, 0x00000060, key);
		CheckStatus(res);
	}
	
	public void mifareClassicLoadKeyB(int recordNumber, byte[] key) throws ClscrflException {
		int res = jniMifareClassicLoadKey(refReader, recordNumber, 0x00000061, key);
		CheckStatus(res);
	}
	
	private native int jniMifareClassicAuthenticate(long reader, int sectorIndex, int recordNumber, int keyType, byte[] UID);
	
	public void mifareClassicAuthenticateA(int sectorIndex, int recordNumber, byte[] UID) throws ClscrflException {
		int res = jniMifareClassicAuthenticate(refReader, sectorIndex, recordNumber, 0x00000060, UID);
		CheckStatus(res);
	}
	
	public void mifareClassicAuthenticateB(int sectorIndex, int recordNumber, byte[] UID) throws ClscrflException {
		int res = jniMifareClassicAuthenticate(refReader, sectorIndex, recordNumber, 0x00000061, UID);
		CheckStatus(res);
	}
	
	public class MifareBlock {
		public int generalError = 0;
		public int innerError = 0;
		public byte[] data = null;
	}
	
	private native MifareBlock jniMifareClassicRead(long reader, int sectorIndex, int blockIndex);
	
	public byte[] mifareClassicRead(int sectorIndex, int blockIndex) throws ClscrflException {
		MifareBlock block = jniMifareClassicRead(refReader, sectorIndex, blockIndex);
		if (block.generalError != 0 && block.innerError != 0) {
			throw new ClscrflException("jniMifareRead: Error code returned!", block.generalError, block.innerError);
		}
		return block.data;
	}
	
	private native int jniMifareClassicWrite(long reader, int sectorIndex, int blockIndex, byte[] data);
	
	public void mifareClassicWrite(int sectorIndex, int blockIndex, byte[] data) throws ClscrflException {
		int res = jniMifareClassicWrite(refReader, sectorIndex, blockIndex, data);
		CheckStatus(res);
	}
}

