#include "StdAfx.h"

#include "Clscrfl.h"

extern CLSCRF_CARD m_DefaultCard;

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Activate_Idle_A(IN     LPVOID  pReader,
OUT LPBYTE  pbATQ,
OUT LPBYTE  pbSAK,
OUT LPBYTE  pbUID,
IN OUT LPDWORD pdwUIDLength)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x43);

		DWORD Size;
		Status = pTr->SingleTransceive();


		if (Status == SCARD_S_SUCCESS)
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, pbATQ, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(3, pbSAK, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}

		if (Status == SCARD_S_SUCCESS)
		{
			*pdwUIDLength = 0; //kbhSCARD_E_INSUFFICIENT_BUFFER
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(4, (LPBYTE)pdwUIDLength, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}

		if (Status == SCARD_S_SUCCESS)
		{
			Size = *pdwUIDLength;
			if (!pTr->m_RcBuf.GetBulk(5, pbUID, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		ZeroMemory(pbATQ, 2);
		*pbSAK = 0;
		ZeroMemory(pbUID, 10);
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	if (Status == SCARD_S_SUCCESS)
	{
		m_DefaultCard.Timeout = 6800;     // Timeout = 64 ms
		m_DefaultCard.CardType = CLSCRF_ICC_TYPE_ISO14443A;
		m_DefaultCard.UID_Len = *pdwUIDLength;
		memcpy(&m_DefaultCard.UID, pbUID, *pdwUIDLength);
		memcpy(&m_DefaultCard.ATQ, pbATQ, 2);
		m_DefaultCard.ATQ.SAK = *pbSAK;
		//GetCID( &m_pCurCard->CID );
		//CLSCRF_Cards_Add( m_pCurCard );
		m_DefaultCard.Active = 1;
		m_DefaultCard.RfMode = 0x00;//RF_14443_A;
		m_DefaultCard.BlockNumberBit = 0;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Halt_A(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x1d);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Activate_Wakeup_A(IN     LPVOID pReader,
IN     LPBYTE pbUID,
IN     DWORD  UIDLength,
OUT LPBYTE pbATQ,
OUT LPBYTE pbSAK)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x44);
		pTr->m_TrBuf.AddByte((BYTE)UIDLength);
		pTr->m_TrBuf.AddBulk(pbUID, UIDLength);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, pbATQ, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(3, pbSAK, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		ZeroMemory(pbATQ, 2);
		*pbSAK = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443A_4_RATS(IN LPVOID pReader,
IN BYTE  ucCID,
OUT LPBYTE  pbATS)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x35);
		pTr->m_TrBuf.AddByte(ucCID);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			DWORD Size = 1;

			BYTE bStatusCode = 0x00;

			if (pTr->m_RcBuf.GetBulk(0, &bStatusCode, &Size))
			{
				if (bStatusCode == 0x00)
				{
					Size = 2;

					DWORD dwLen = 0;

					if (pTr->m_RcBuf.GetBulk(1, (PBYTE)&dwLen, &Size))
					{
						if (dwLen > 0)
						{
							BYTE ucLen = 0;
							Size = 1;

							pTr->m_RcBuf.GetBulk(3, &ucLen, &Size);

							if (ucLen > 0)
							{
								Size = ucLen;

								pTr->m_RcBuf.GetBulk(3, pbATS, &Size);

								memcpy(&m_DefaultCard.ATQ.ATS, pbATS, Size);
							}
						}
					}
				}
			}

		}
		else
		{
			BYTE bErr = 0x00;
			CLSCRF_GetLastInternalError(pReader, &bErr);
			bErr = bErr;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443A_4_PPS(IN LPVOID pReader,
IN BYTE  ucCID,
IN BYTE  ucBaudrate)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x36);
		pTr->m_TrBuf.AddByte(ucCID);
		pTr->m_TrBuf.AddByte(ucBaudrate);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443A_4_Activate(LPVOID  pReader,
IN BYTE ucCID,
IN BYTE ucBaudrate,
OUT LPBYTE  pbATS)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x2B);
		pTr->m_TrBuf.AddByte(ucCID);
		pTr->m_TrBuf.AddByte(ucBaudrate);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			DWORD Size = 1;

			BYTE bStatusCode = 0x00;

			if (pTr->m_RcBuf.GetBulk(0, &bStatusCode, &Size))
			{
				if (bStatusCode == 0x00)
				{
					Size = 2;

					DWORD dwLen = 0;

					if (pTr->m_RcBuf.GetBulk(1, (PBYTE)&dwLen, &Size))
					{
						if (dwLen > 0)
						{
							BYTE ucLen = 0;
							Size = 1;

							pTr->m_RcBuf.GetBulk(3, &ucLen, &Size);

							if (ucLen > 0)
							{
								Size = ucLen;

								pTr->m_RcBuf.GetBulk(3, pbATS, &Size);

								memcpy(&m_DefaultCard.ATQ.ATS, pbATS, Size);
							}
						}
					}
				}
			}

		}
		else
		{
			BYTE bErr = 0x00;
			CLSCRF_GetLastInternalError(pReader, &bErr);
			bErr = bErr;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

