#include "StdAfx.h"

#include "Clscrfl.h"

extern CLSCRF_CARD m_DefaultCard;

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Activate_Idle_B(IN     LPVOID  pReader,
IN     BYTE    bAfi,
IN     BYTE    bSni,
OUT LPDWORD pdwPUPI,
OUT LPDWORD pdwAppData,
OUT LPDWORD pdwProtInfo,
IN OUT LPDWORD pdwCardCount)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x56);
		pTr->m_TrBuf.AddByte(bAfi);
		pTr->m_TrBuf.AddByte(bSni);

		DWORD Ptr = 1;
		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			DWORD dwCount = pTr->m_RcBuf.GetLength() / 11;
			if (*pdwCardCount < dwCount)
			{
				return SCARD_E_INSUFFICIENT_BUFFER;
			}
			*pdwCardCount = dwCount;
			for (DWORD i = 0; i < *pdwCardCount; i++)
			{
				Size = 4;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwPUPI[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}

				Size = 4;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwAppData[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}

				Size = 3;
				pdwProtInfo[i] = 0;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwProtInfo[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}
			}
		}
	}
	else
	{
		*pdwCardCount = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Attrib_B(IN     LPVOID  pReader,
IN     LPBYTE  pbPUPI,
IN     BYTE    bParam1,
IN     BYTE    bParam2,
IN     BYTE    bParam3,
IN     BYTE    bParam4,
IN OUT LPBYTE  pbHigherLayerBuf,
IN     DWORD   dwHLBufSize,
IN OUT LPDWORD pdwHLInfLength,
OUT LPBYTE  pbMbliCid)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x54);
		pTr->m_TrBuf.AddByte(8 + (BYTE)*pdwHLInfLength);
		pTr->m_TrBuf.AddBulk(pbPUPI, 4);
		pTr->m_TrBuf.AddByte(bParam1);
		pTr->m_TrBuf.AddByte(bParam2);
		pTr->m_TrBuf.AddByte(bParam3);
		pTr->m_TrBuf.AddByte(bParam4);
		if (*pdwHLInfLength)
			pTr->m_TrBuf.AddBulk(pbHigherLayerBuf, *pdwHLInfLength);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			DWORD Size;
			*pdwHLInfLength = pTr->m_RcBuf.GetLength();
			if (*pdwHLInfLength > 1)
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(1, pbMbliCid, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
				else if ((*pbMbliCid & 0x0f) != bParam4)
					Status = SCARD_F_INTERNAL_ERROR;

				*pdwHLInfLength -= 2;
			}
			else
			{
				*pdwHLInfLength = 0;
				Status = SCARD_E_COMM_DATA_LOST;
			}

			if (Status == SCARD_S_SUCCESS && *pdwHLInfLength)
			{
				if (*pdwHLInfLength > dwHLBufSize)
				{
					*pdwHLInfLength = dwHLBufSize;
					Status = SCARD_E_INSUFFICIENT_BUFFER;
				}
				if (!pTr->m_RcBuf.GetBulk(2, pbHigherLayerBuf, pdwHLInfLength))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}
	else
	{
		*pdwHLInfLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Halt_B(IN LPVOID pReader,
IN LPBYTE pbPUPI)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x55);
		pTr->m_TrBuf.AddBulk(pbPUPI, 4);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Activate_Wakeup_B(IN     LPVOID  pReader,
IN     BYTE    bAfi,
IN     BYTE    bSni,
OUT LPDWORD pdwPUPI,
OUT LPDWORD pdwAppData,
OUT LPDWORD pdwProtInfo,
IN OUT LPDWORD pdwCardCount)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x57);
		pTr->m_TrBuf.AddByte(bAfi);
		pTr->m_TrBuf.AddByte(bSni);

		DWORD Ptr = 1;
		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pdwCardCount = pTr->m_RcBuf.GetLength() / 11;
			for (DWORD i = 0; i < *pdwCardCount; i++)
			{
				Size = 4;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwPUPI[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}

				Size = 4;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwAppData[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}

				Size = 3;
				pdwProtInfo[i] = 0;
				if (pTr->m_RcBuf.GetBulk(Ptr, (unsigned char*)&pdwProtInfo[i], &Size))
					Ptr += Size;
				else
				{
					Status = SCARD_E_COMM_DATA_LOST;
					break;
				}
			}
		}
	}
	else
	{
		*pdwCardCount = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}
