#include "StdAfx.h"

#include "Clscrfl.h"

LONG TranslateError(BYTE ucErrorCode)
{
	LONG Status;

	switch (ucErrorCode)
	{
	case 0x01:
		Status = SCARD_E_UNSUPPORTED_FEATURE;
		break;
	case 0x02:
		Status = SCARD_E_COMM_DATA_LOST;
		break;
	case 0x03:
		Status = SCARD_E_INVALID_PARAMETER;
		break;
	case 0x0F:
		Status = SCARD_E_UNEXPECTED;
		break;
	case 0x010:
		Status = SCARD_E_INVALID_PARAMETER;
		break;
	case 0x011:
		Status = SCARD_E_NO_ACCESS;
		break;
	case 0x012:
		Status = SCARD_E_NO_ACCESS;
		break;
	case 0x013:
		Status = SCARD_E_NOT_TRANSACTED;
		break;
	case 0x014:
		Status = SCARD_E_NOT_TRANSACTED;
		break;
	default:
		Status = SCARD_E_UNEXPECTED;
	}
	if ((ucErrorCode > 0xA0) && (ucErrorCode < 0xDF))
		Status = SCARD_E_UNEXPECTED;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Inventory_15693(IN     LPVOID  pReader,
IN     BYTE    bFlags,
IN     BYTE    bInventory,
IN     BYTE    bAfi,
IN     BYTE    bMaskLen,
IN     LPCBYTE pbMaskVal,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x31);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bInventory);
		pTr->m_TrBuf.AddByte(bAfi);
		pTr->m_TrBuf.AddByte(bMaskLen);
		int SendLength = (bMaskLen + 7) / 8;
		pTr->m_TrBuf.AddBulk(pbMaskVal, SendLength);

		Status = pTr->SingleTransceive();

		BYTE bErr = 0;
		if (Status != SCARD_S_SUCCESS)
			CLSCRF_GetLastInternalError(pReader, &bErr);

		if (
			(Status == SCARD_S_SUCCESS)
			|| (bErr == 0xEB/* Frame Collision */)
			/*||( bErr == 0xE8 )
			||( bErr == 0xF4 )*/
			)
		{
			//DWORD RecvLength = 0;
			DWORD Size = pTr->m_RcBuf.GetLength();
			if (Size > *pdwRecvLength)
			{
				*pdwRecvLength = 0;
				return SCARD_E_INSUFFICIENT_BUFFER;
			}

			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					return SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength)
			{
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
		else
			*pdwRecvLength = 0;
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_FindAllTags_15693(IN     LPVOID  pReader,
IN     BYTE    bFlags,
IN     BYTE    bAfi,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	/*
	BYTE bMaskLen = 0;
	UINT64 qwMaskVal = 0;
	std::list<UINT64> MaskList;
	std::list<BYTE> MaskLenList;
	UINT64 qwNewMask = 0;


	BYTE pbRecBuf[256] = {0};
	DWORD dwRecLen = sizeof(pbRecBuf);
	int iBufPos = 0, iTag = 0;
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufferSize = *pdwRecvLength;

	if (8 > BufferSize)
	return SCARD_E_INSUFFICIENT_BUFFER;

	while (true)
	{
	dwRecLen = sizeof(pbRecBuf);
	Status = CLSCRF_Inventory_15693(
	pReader,
	bFlags,
	0x01,
	bAfi,
	bMaskLen,
	(LPCBYTE)&qwMaskVal,
	pbRecBuf,
	&dwRecLen
	);

	iBufPos = 0;
	if (
	((Status == SCARD_F_INTERNAL_ERROR) && dwRecLen)
	||(Status == SCARD_S_SUCCESS)
	)
	for (int ts = 0; ts < 16; ts++)
	{
	switch (pbRecBuf[iBufPos++])
	{
	case 0x00:
	break;
	case 0xEB:
	qwNewMask = qwMaskVal;
	qwNewMask |= (((UINT64)ts) << bMaskLen);
	MaskList.push_back(qwNewMask);
	MaskLenList.push_back(bMaskLen+4);
	break;
	case 0xFF:
	break;
	default:
	;
	}
	if (pbRecBuf[iBufPos++])
	{
	if (UINT(iTag)*8 >= BufferSize)
	return SCARD_E_INSUFFICIENT_BUFFER;

	memcpy(&pbRecvBuffer[(iTag++)*8], &pbRecBuf[iBufPos+2], 8);
	iBufPos += pbRecBuf[iBufPos-1];
	}
	}
	else
	break;

	if (MaskList.empty())
	break;
	else
	{
	bMaskLen = MaskLenList.front();
	qwMaskVal = MaskList.front();
	MaskList.pop_front();
	MaskLenList.pop_front();
	}
	}

	*pdwRecvLength = iTag*8;
	*/
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x30);
		pTr->m_TrBuf.AddByte(0);
		pTr->m_TrBuf.AddByte(0);
		pTr->m_TrBuf.AddByte(bAfi);

		Status = pTr->SingleTransceive();

		BYTE bErr = 0;
		if (Status != SCARD_S_SUCCESS)
			CLSCRF_GetLastInternalError(pReader, &bErr);

		if (
			(Status == SCARD_S_SUCCESS)
			|| (bErr == 0xEB/* Frame Collision */)
			)
		{
			//DWORD RecvLength = 0;
			DWORD Size = pTr->m_RcBuf.GetLength();
			if (Size > *pdwRecvLength)
			{
				*pdwRecvLength = 0;
				return SCARD_E_INSUFFICIENT_BUFFER;
			}

			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					return SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength)
			{
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
		else
			*pdwRecvLength = 0;
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Stay_Quiet_15693(IN LPVOID pReader,
IN BYTE   bFlags,
IN BYTE   bStayQuiet,
IN LPBYTE pbUID)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x32);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bStayQuiet);
		pTr->m_TrBuf.AddBulk(pbUID, 8);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Select_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bSelect,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bSelect != 0x25)
			|| ((bFlags & 0x20) == 0)
			|| (pbUID == NULL))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = 10;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bSelect);
		pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ResetToReady_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bResetToReady,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bResetToReady != 0x26)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 10 : 2;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bResetToReady);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ReadSingleBlock_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bReadSingleBlock,
IN     LPBYTE pbUID,
IN     BYTE   bBlockNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbBlockSecurityStatus,
OUT LPBYTE pbData,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	BOOL fAddress = bFlags & 0x20;
	BOOL fOption = bFlags & 0x40;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bReadSingleBlock != 0x20)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bReadSingleBlock);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bBlockNumber);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 5)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	*pbFlags = 0x01;
	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		*pbBlockSecurityStatus = 0;
		if (*pbFlags & 0x01)
		{
			*pbErrorCode = pTr->m_RcBuf[4];
			Status = TranslateError(*pbErrorCode);

			ZeroMemory(pbData, ICODE_BLOCK_SIZE);
		}
		else
		{
			DWORD Expected = (fOption) ? 9 : 8;
			if (Size < Expected)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			if (Status == SCARD_S_SUCCESS)
			{
				DWORD Index = 4;
				if (fOption)
					*pbBlockSecurityStatus = pTr->m_RcBuf[Index++];
				Size = ICODE_BLOCK_SIZE;
				if (!pTr->m_RcBuf.GetBulk(Index, pbData, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_WriteSingleBlock_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bWriteSingleBlock,
IN     LPBYTE pbUID,
IN     BYTE   bBlockNumber,
IN     LPBYTE pbData,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bWriteSingleBlock != 0x21)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = ICODE_BLOCK_SIZE + ((fAddress) ? 11 : 3);
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bWriteSingleBlock);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bBlockNumber);
		pTr->m_TrBuf.AddBulk(pbData, ICODE_BLOCK_SIZE);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	*pbFlags = 0x01;
	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_LockBlock_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bLockBlock,
IN     LPBYTE pbUID,
IN     BYTE   bBlockNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bLockBlock != 0x22)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bLockBlock);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bBlockNumber);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_WriteMultipleBlocks_15693(IN		LPVOID pReader,
IN		BYTE   bFlags,
IN		BYTE   bWriteMultipleBlocks,
IN		LPBYTE pbUID,
IN		BYTE   bBlockNumber,
IN		BYTE   bBlocksCount,
IN		LPBYTE pbData,
OUT	LPBYTE pbFlags,
OUT	LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500 * bBlocksCount;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0x00;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bWriteMultipleBlocks != 0x24)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = 2 + ((fAddress) ? 8 : 0) + 2 + ICODE_BLOCK_SIZE * bBlocksCount;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);

		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bWriteMultipleBlocks);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bBlockNumber);
		pTr->m_TrBuf.AddByte((bBlocksCount - 1));
		pTr->m_TrBuf.AddBulk(pbData, ICODE_BLOCK_SIZE * bBlocksCount);
		//Timeout *= bBlocksCount;
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	*pbFlags = 0x01;
	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ReadMultipleBlocks_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bReadMultipleBlock,
IN     LPBYTE pbUID,
IN     BYTE   bFirstBlockNumber,
IN OUT LPBYTE pbNumberOfBlocks,
OUT LPBYTE pbFlags,
OUT LPBYTE pbBlockSecurityStatus,
OUT LPBYTE pbData,
OUT LPBYTE pbErrorCode)

{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG  Status = SCARD_S_SUCCESS;
	int   BlockCount = *pbNumberOfBlocks;// +1;
	ZeroMemory(pbBlockSecurityStatus, BlockCount);
	ZeroMemory(pbData, BlockCount * ICODE_BLOCK_SIZE);
	*pbErrorCode = 0;
	BOOL fAddress = bFlags & 0x20;
	BOOL fOption = bFlags & 0x40;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bReadMultipleBlock != 0x23)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD PieceCount = 0;
	DWORD SendLength = (fAddress) ? 12 : 4;
	DWORD BlockStatusSize = (fOption) ? 1 + ICODE_BLOCK_SIZE : ICODE_BLOCK_SIZE;
	int   Max_Blocks = (ICODE_PACKET_SIZE - 4) / BlockStatusSize;
	int   PieceFirst = bFirstBlockNumber;
	DWORD IndexIn = 0;
	DWORD IndexBSS = 0;
	DWORD IndexDat = 0;
	*pbNumberOfBlocks = 0;

	do
	{
		if (Status == SCARD_S_SUCCESS)
		{
			PieceCount = (BlockCount > Max_Blocks) ? Max_Blocks : BlockCount;
			pTr->m_TrBuf.Clear();
			pTr->m_TrBuf.AddByte(0x48);
			pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
			pTr->m_TrBuf.AddByte(bFlags);
			pTr->m_TrBuf.AddByte(bReadMultipleBlock);
			if (fAddress)
				pTr->m_TrBuf.AddBulk(pbUID, 8);
			pTr->m_TrBuf.AddByte((BYTE)PieceFirst);
			pTr->m_TrBuf.AddByte((BYTE)(PieceCount - 1));
			pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

			Status = pTr->SingleTransceive();
		}

		DWORD Size = 0;

		if (Status == SCARD_S_SUCCESS)
		{
			Size = pTr->m_RcBuf.GetLength();
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}

		//        DWORD Expected;

		*pbFlags = 0x01;
		if (Status == SCARD_S_SUCCESS)
		{
			*pbFlags = pTr->m_RcBuf[3];
			if (*pbFlags & 0x01)
			{
				//                if( *pbNumberOfBlocks )
				//                    *pbFlags = 0;
				//                else
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);

				break;
			}
			else
			{
				//                Expected = 4 + PieceCount * BlockStatusSize;
				//                if( Size < Expected )
				//                    Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
				if (Status == SCARD_S_SUCCESS)
				{
					IndexIn = 4;
					do
					{
						if (fOption)
							pbBlockSecurityStatus[IndexBSS++] = pTr->m_RcBuf[IndexIn++];
						Size = ICODE_BLOCK_SIZE;
						if (!pTr->m_RcBuf.GetBulk(IndexIn, &pbData[IndexDat], &Size))
						{
							Status = SCARD_E_COMM_DATA_LOST;
							break;
						}
						else
						{
							*pbNumberOfBlocks += 1;
							IndexDat += ICODE_BLOCK_SIZE;
							IndexIn += ICODE_BLOCK_SIZE;
						}
					} while (IndexIn < pTr->m_RcBuf.GetLength());
				}
			}
		}

		if (Status == SCARD_S_SUCCESS)
		{
			PieceFirst += PieceCount;
			BlockCount -= PieceCount;
		}
		else
			break;

	} while (BlockCount > 0);

	//if (Status == SCARD_S_SUCCESS)
	//	*pbNumberOfBlocks -= 1;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_WriteAFI_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bWriteAFI,
IN     LPBYTE pbUID,
IN     BYTE   bAFI,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bWriteAFI != 0x27)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bWriteAFI);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bAFI);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_LockAFI_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bLockAFI,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bLockAFI != 0x28)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 10 : 2;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bLockAFI);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_WriteDSFID_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bWriteDSFID,
IN     LPBYTE pbUID,
IN     BYTE   bDSFID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bWriteDSFID != 0x29)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bWriteDSFID);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddByte(bDSFID);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_LockDSFID_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bLockDSFID,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bLockDSFID != 0x2a)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 10 : 2;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bLockDSFID);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_GetSystemInfo_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bGetSystemInfo,
IN OUT LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbInfoFlags,
OUT LPBYTE pbDSFID,
OUT LPBYTE pbAFI,
OUT LPWORD pbMemorySize,
OUT LPBYTE pbICReference,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbDSFID = 0;
	*pbAFI = 0;
	*pbMemorySize = 0;
	*pbICReference = 0;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bGetSystemInfo != 0x2b)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 10 : 2;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bGetSystemInfo);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 5)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	DWORD Index = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			*pbErrorCode = pTr->m_RcBuf[4];
			Status = TranslateError(*pbErrorCode);
		}
		else
		{
			Index = 4;
			if (Status == SCARD_S_SUCCESS)
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(Index++, pbInfoFlags, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
			if (Status == SCARD_S_SUCCESS)
			{
				Size = 8;
				if (!pTr->m_RcBuf.GetBulk(Index, pbUID, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
				Index += 8;
			}
			if (Status == SCARD_S_SUCCESS && (*pbInfoFlags & 0x01))
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(Index++, pbDSFID, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
			if (Status == SCARD_S_SUCCESS && (*pbInfoFlags & 0x02))
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(Index++, pbAFI, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
			if (Status == SCARD_S_SUCCESS && (*pbInfoFlags & 0x04))
			{
				Size = 2;
				if (!pTr->m_RcBuf.GetBulk(Index, (LPBYTE)pbMemorySize, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
				Index += 2;
			}
			if (Status == SCARD_S_SUCCESS && (*pbInfoFlags & 0x08))
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(Index, pbICReference, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_GetMultipleBSS_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bGetMultipleBSS,
IN     LPBYTE pbUID,
IN     BYTE   bFirstBlockNumber,
IN OUT LPBYTE pbNumberOfBlocks,
OUT LPBYTE pbFlags,
OUT LPBYTE pbBlockSecurityStatus,
OUT LPBYTE pbErrorCode)

{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG  Status = SCARD_S_SUCCESS;
	int   BlockCount = *pbNumberOfBlocks + 1;
	ZeroMemory(pbBlockSecurityStatus, BlockCount);
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bGetMultipleBSS != 0x2c)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD PieceCount = 0;
	DWORD SendLength = (fAddress) ? 12 : 4;
	int   Max_Blocks = ICODE_PACKET_SIZE;
	int   PieceFirst = bFirstBlockNumber;
	DWORD IndexIn = 0;
	DWORD IndexBSS = 0;
	*pbNumberOfBlocks = 0;

	do
	{
		if (Status == SCARD_S_SUCCESS)
		{
			PieceCount = (BlockCount > Max_Blocks) ? Max_Blocks : BlockCount;
			pTr->m_TrBuf.Clear();
			pTr->m_TrBuf.AddByte(0x48);
			pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
			pTr->m_TrBuf.AddByte(bFlags);
			pTr->m_TrBuf.AddByte(bGetMultipleBSS);
			if (fAddress)
				pTr->m_TrBuf.AddBulk(pbUID, 8);
			pTr->m_TrBuf.AddByte((BYTE)PieceFirst);
			pTr->m_TrBuf.AddByte((BYTE)(PieceCount - 1));
			pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

			Status = pTr->SingleTransceive();
		}

		DWORD Size = 0;

		if (Status == SCARD_S_SUCCESS)
		{
			Size = pTr->m_RcBuf.GetLength();
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}

		//        DWORD Expected;

		if (Status == SCARD_S_SUCCESS)
		{
			*pbFlags = pTr->m_RcBuf[3];
			if (*pbFlags & 0x01)
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);

				break;
			}
			else
			{
				//                Expected = 4 + PieceCount * BlockStatusSize;
				//                if( Size < Expected )
				//                    Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
				if (Status == SCARD_S_SUCCESS)
				{
					IndexIn = 4;
					do
					{
						pbBlockSecurityStatus[IndexBSS++] = pTr->m_RcBuf[IndexIn++];
						*pbNumberOfBlocks += 1;
					} while (IndexIn < pTr->m_RcBuf.GetLength());
				}
			}
		}

		if (Status == SCARD_S_SUCCESS)
		{
			PieceFirst += PieceCount;
			BlockCount -= PieceCount;
		}
		else
			break;

	} while (BlockCount > 0);

	if (Status == SCARD_S_SUCCESS)
		*pbNumberOfBlocks -= 1;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SetEAS_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bSetEAS,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bSetEAS != 0xa2)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bSetEAS);
		pTr->m_TrBuf.AddByte(bICMfgCode);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ResetEAS_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bResetEAS,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bResetEAS != 0xa3)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bResetEAS);
		pTr->m_TrBuf.AddByte(bICMfgCode);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_LockEAS_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bLockEAS,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 2500;         // 23,6 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bLockEAS != 0xa4)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bLockEAS);
		pTr->m_TrBuf.AddByte(bICMfgCode);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 4)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			if (Size < 5)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			else
			{
				*pbErrorCode = pTr->m_RcBuf[4];
				Status = TranslateError(*pbErrorCode);
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_EASAlarm_15693(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bEASAlarm,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbEASData,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		if ((bEASAlarm != 0xa5)
			|| (fAddress && (pbUID == NULL)))
			Status = SCARD_E_INVALID_PARAMETER;
	}

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	if (Status == SCARD_S_SUCCESS)
	{
		SendLength = (fAddress) ? 11 : 3;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
		pTr->m_TrBuf.AddByte(bFlags);
		pTr->m_TrBuf.AddByte(bEASAlarm);
		pTr->m_TrBuf.AddByte(bICMfgCode);
		if (fAddress)
			pTr->m_TrBuf.AddBulk(pbUID, 8);
		pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

		Status = pTr->SingleTransceive();
	}

	DWORD Size = 0;

	if (Status == SCARD_S_SUCCESS)
	{
		Size = pTr->m_RcBuf.GetLength();
		if (Size < 5)
			Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pbFlags = pTr->m_RcBuf[3];
		if (*pbFlags & 0x01)
		{
			*pbErrorCode = pTr->m_RcBuf[4];
			Status = TranslateError(*pbErrorCode);

			ZeroMemory(pbEASData, ICODE_EAS_DATA_SIZE);
		}
		else
		{
			DWORD Expected = 4 + ICODE_EAS_DATA_SIZE;
			if (Size < Expected)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
			if (Status == SCARD_S_SUCCESS)
			{
				DWORD Index = 4;
				Size = ICODE_EAS_DATA_SIZE;
				if (!pTr->m_RcBuf.GetBulk(Index, pbEASData, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_GetRandomNumber(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPDWORD pdwRandomNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 11 : 3;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB2);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		DWORD Expected = 4 +ICODE_GET_RANDOM_NUMBER_SIZE;
		if (Size < Expected)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		DWORD Index = 4;
		Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		*pdwRandomNumber = 0;
		if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
			return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_SetPassword(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     BYTE	  ucPasswordIdentifier,
IN     DWORD  dwPassword,
IN     DWORD  dwRandomNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	DWORD dwPasswordXOR = dwPassword ^ ((dwRandomNumber & 0x0000FFFF) << 16) ^ (dwRandomNumber & 0x0000FFFF);

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 16 : 8;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB3);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddByte(ucPasswordIdentifier);
	pTr->m_TrBuf.AddBulk((LPBYTE)&dwPasswordXOR, 4);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_WritePassword(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     BYTE	  ucPasswordIdentifier,
IN     DWORD  dwPassword,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 16 : 8;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB4);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddByte(ucPasswordIdentifier);
	pTr->m_TrBuf.AddBulk((LPBYTE)&dwPassword, 4);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_LockPassword(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     BYTE	  ucPasswordIdentifier,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 12 : 4;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB5);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddByte(ucPasswordIdentifier);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_PasswordProtectEAS_AFI(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 11 : 3;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xA6);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_64BitProtection(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 11 : 3;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xBB);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_ProtectPage(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     BYTE   ucProtectionPointerAddress,
IN     BYTE   ucExtendedProtectionStatus,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 13 : 5;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB6);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddByte(ucProtectionPointerAddress);
	pTr->m_TrBuf.AddByte(ucExtendedProtectionStatus);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_LockPageProtectionCondition(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     BYTE   ucProtectionPointerAddress,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 12 : 4;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB7);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddByte(ucProtectionPointerAddress);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_Destroy(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     DWORD  dwPassword,
IN     DWORD  dwRandomNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	DWORD dwPasswordXOR = dwPassword ^ ((dwRandomNumber & 0x0000FFFF) << 16) ^ (dwRandomNumber & 0x0000FFFF);

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 15 : 7;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xB9);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddBulk((LPBYTE)&dwPasswordXOR, 4);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_15693_ICode_EnablePrivacy(IN     LPVOID pReader,
IN     BYTE   bFlags,
IN     BYTE   bICMfgCode,
IN     LPBYTE pbUID,
IN     DWORD  dwPassword,
IN     DWORD  dwRandomNumber,
OUT LPBYTE pbFlags,
OUT LPBYTE pbErrorCode)
{
	const DWORD Timeout = 1200;         // 11,3 ms
	LONG Status = SCARD_S_SUCCESS;
	*pbErrorCode = 0;
	*pbFlags = 0x01;
	BOOL fAddress = bFlags & 0x20;

	if (pReader == NULL)
		return SCARD_E_READER_UNAVAILABLE;

	if (fAddress && (pbUID == NULL))
		return SCARD_E_INVALID_PARAMETER;

	DWORD dwPasswordXOR = dwPassword ^ ((dwRandomNumber & 0x0000FFFF) << 16) ^ (dwRandomNumber & 0x0000FFFF);

	CTransfer* pTr = (CTransfer*)pReader;
	DWORD SendLength;

	SendLength = (fAddress) ? 15 : 7;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x48);
	pTr->m_TrBuf.AddBulk((LPBYTE)&SendLength, 2);
	pTr->m_TrBuf.AddByte(bFlags);
	pTr->m_TrBuf.AddByte(0xBA);
	pTr->m_TrBuf.AddByte(bICMfgCode);
	if (fAddress)
		pTr->m_TrBuf.AddBulk(pbUID, 8);
	pTr->m_TrBuf.AddBulk((LPBYTE)&dwPasswordXOR, 4);
	pTr->m_TrBuf.AddBulk((LPBYTE)&Timeout, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 0;

	Size = pTr->m_RcBuf.GetLength();
	if (Size < 4)
		return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

	*pbFlags = pTr->m_RcBuf[3];
	if (*pbFlags & 0x01)
	{
		if (Size < 5)
			return SCARD_E_COMM_DATA_LOST;    // 0x8010002F

		*pbErrorCode = pTr->m_RcBuf[4];
		Status = TranslateError(*pbErrorCode);

		//ZeroMemory(pbEASData, ICODE_GET_RANDOM_NUMBER_SIZE);
		//*pdwRandomNumber = (DWORD)-1;
	}
	else
	{
		//DWORD Expected = 4 + ICODE_GET_RANDOM_NUMBER_SIZE;
		//if (Size < Expected)
		//	return SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		//DWORD Index = 4;
		//Size = ICODE_GET_RANDOM_NUMBER_SIZE;
		//*pdwRandomNumber = 0;
		//if (!pTr->m_RcBuf.GetBulk(Index, (BYTE*)pdwRandomNumber, &Size))
		//	return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

