#include "StdAfx.h"

#include <codecvt>

#include "Clscrfl.h"

#include "./NFC/NdefMessage.h"
#include "./NFC/NFCLinkLayer.h"
#include "./NFC/SNEP.h"
#include "./NFC/Tlv.h"

#include <iconv.h>

extern std::map<PVOID, READER_SETTINGS> m_ReaderSettings;
extern CLSCRF_CARD*  m_pCurCard;

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_Create( 
IN LPBYTE  lpMsgDataBytes, 
IN DWORD  dwMsgDataLength,
OUT LPHANDLE phNDEFMessage)
{
	if (phNDEFMessage == NULL)
		return SCARD_E_INVALID_PARAMETER;

	NdefMessage* msg = NULL;

	if ((dwMsgDataLength == 0) || (lpMsgDataBytes == NULL))
		msg = new NdefMessage();
	else
		msg = new NdefMessage(lpMsgDataBytes, dwMsgDataLength);

	*phNDEFMessage = msg;

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_GetRecordsCount(
IN HANDLE hNDEFMessage,
OUT LPDWORD  pdwRecordsCount)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwRecordsCount == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwRecordsCount = ((NdefMessage*)hNDEFMessage)->getRecordCount();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_GetDataSize(
IN HANDLE hNDEFMessage,
OUT LPDWORD  pdwDataSize)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwDataSize == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwDataSize = ((NdefMessage*)hNDEFMessage)->getEncodedSize();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_GetData(
IN HANDLE hNDEFMessage,
OUT LPBYTE  lpData)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	((NdefMessage*)hNDEFMessage)->encode(lpData);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecord(
IN HANDLE hNDEFMessage,
IN HANDLE hNDEFRecord)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (!((NdefMessage*)hNDEFMessage)->addRecord((NdefRecord*)hNDEFRecord))
		return SCARD_E_UNEXPECTED;

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecordEmpty(
IN HANDLE hNDEFMessage)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NdefMessage*)hNDEFMessage)->addEmptyRecord();

	return SCARD_S_SUCCESS;
}

//CLSCRF_API LONG CLSCRF_DECL
//CLSCRF_NFC_NDEF_Message_AddRecordText(
//IN HANDLE hNDEFMessage,
//IN LPWSTR lpwstrText,
//IN LPWSTR lpwstrEncoding)
//{
//	if (hNDEFMessage == NULL)
//		return SCARD_E_INVALID_PARAMETER;
//
//	if (lpwstrText == NULL)
//		return SCARD_E_INVALID_PARAMETER;
//
//	if (lpwstrEncoding == NULL)
//		lpwstrEncoding = L"en";
//
//	if (CString(lpwstrEncoding) == CString(""))
//		lpwstrEncoding = L"en";
//
//	((NdefMessage*)hNDEFMessage)->addTextRecord(CString(lpwstrText), CString(lpwstrEncoding));
//
//	return SCARD_S_SUCCESS;
//}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecordText(
IN HANDLE hNDEFMessage,
IN LPWSTR lpwstrText,
IN LPWSTR lpwstrLanguage)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrText == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrLanguage == NULL)
		lpwstrLanguage = (LPWSTR)L"en";

	//if (CString(lpwstrLanguage) == CString(""))
	if (std::wstring(lpwstrLanguage) == std::wstring(L""))
		lpwstrLanguage = (LPWSTR)L"en";

	// your "Unicode" string
	wchar_t const * utf16_string = lpwstrText;// L"Hello, World!";
	wchar_t const * utf16_encoding = lpwstrLanguage;// L"Hello, World!";

	//std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> convert;

	//std::string utf8_string = convert.to_bytes(utf16_string);

	size_t srclen = wcslen(utf16_string);
	size_t enclen = wcslen(utf16_encoding);
	char *utf8_string = new char[srclen];
	char *utf8_encoding = new char[enclen];
	size_t dstlen = srclen;
	size_t dstenclen = enclen;
	iconv_t conv = iconv_open("UTF-8", "UTF-16");
	iconv(conv, (char**)&utf16_string, &srclen, &utf8_string, &dstlen);
	iconv(conv, (char**)&utf16_encoding, &enclen, &utf8_encoding, &dstenclen);
	iconv_close(conv);

	//((NdefMessage*)hNDEFMessage)->addTextRecord(utf8_string.c_str(), CString(lpwstrLanguage));
	((NdefMessage*)hNDEFMessage)->addTextRecord(utf8_string, utf8_encoding);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecordUri(
IN HANDLE hNDEFMessage,
IN LPWSTR lpwstrUri)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrUri == NULL)
		return SCARD_E_INVALID_PARAMETER;

	// your "Unicode" string
	wchar_t const * utf16_string = lpwstrUri;

	//std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> convert;

	//std::string utf8_string = convert.to_bytes(utf16_string);

	size_t srclen = wcslen(utf16_string);
	char *utf8_string = new char[srclen];
	size_t dstlen = srclen;
	iconv_t conv = iconv_open("UTF-8", "UTF-16");
	iconv(conv, (char**)&utf16_string, &srclen, &utf8_string, &dstlen);
	iconv_close(conv);

	//((NdefMessage*)hNDEFMessage)->addUriRecord(CString(utf8_string.c_str()));
	((NdefMessage*)hNDEFMessage)->addUriRecord(std::string(utf8_string));

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecordMimeMedia(
IN HANDLE hNDEFMessage,
IN LPWSTR lpwstrMimeType,
IN LPBYTE lpPayload,
IN DWORD dwPayloadLength)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrMimeType == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (lpPayload == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwPayloadLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	wchar_t const * utf16_string = lpwstrMimeType;
	size_t srclen = wcslen(utf16_string);
	char *utf8_string = new char[srclen];
	size_t dstlen = srclen;
	iconv_t conv = iconv_open("UTF-8", "UTF-16");
	iconv(conv, (char**)&utf16_string, &srclen, &utf8_string, &dstlen);
	iconv_close(conv);
	
	//((NdefMessage*)hNDEFMessage)->addMimeMediaRecord(CString(lpwstrMimeType), lpPayload, dwPayloadLength);
	((NdefMessage*)hNDEFMessage)->addMimeMediaRecord(std::string(utf8_string), lpPayload, dwPayloadLength);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_AddRecordMimeMediaText(
IN HANDLE hNDEFMessage,
IN LPWSTR lpwstrMimeType,
IN LPWSTR lpwstrText)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrMimeType == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrText == NULL)
		return SCARD_E_INVALID_PARAMETER;

	// your "Unicode" string
	wchar_t const * utf16_string = lpwstrText;// L"Hello, World!";
	wchar_t const * utf16_mime = lpwstrMimeType;// L"Hello, World!";
	
	//std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> convert;

	//std::string utf8_string = convert.to_bytes(utf16_string);

	size_t srclen = wcslen(utf16_string);
	size_t mimelen = wcslen(utf16_mime);
	char *utf8_string = new char[srclen];
	char *utf8_mime = new char[mimelen];
	size_t dstlen = srclen;
	size_t dstmimelen = mimelen;
	iconv_t conv = iconv_open("UTF-8", "UTF-16");
	iconv(conv, (char**)&utf16_string, &srclen, &utf8_string, &dstlen);
	iconv(conv, (char**)&utf16_mime, &mimelen, &utf8_mime, &dstmimelen);
	iconv_close(conv);
	
	//((NdefMessage*)hNDEFMessage)->addMimeMediaRecord(CString(lpwstrMimeType), utf8_string.c_str());
	((NdefMessage*)hNDEFMessage)->addMimeMediaRecord(std::string(utf8_mime), utf8_string);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_GetRecord(
IN HANDLE hNDEFMessage,
IN DWORD  dwRecordIndex,
OUT LPHANDLE phNDEFRecord)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (phNDEFRecord == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwRecordIndex >= ((NdefMessage*)hNDEFMessage)->getRecordCount())
		return SCARD_E_INVALID_PARAMETER;

	*phNDEFRecord = ((NdefMessage*)hNDEFMessage)->getRecord(dwRecordIndex);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Message_Destroy(
IN HANDLE hNDEFMessage)
{
	if (hNDEFMessage == 0)
		return SCARD_E_INVALID_PARAMETER;

	delete (NdefMessage*)hNDEFMessage;

	return SCARD_S_SUCCESS;
}




CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_Create(
OUT LPHANDLE phNDEFRecord)
{
	if (phNDEFRecord == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*phNDEFRecord = new NdefRecord();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetDataSize(
IN HANDLE hNDEFRecord,
OUT LPDWORD  pdwDataSize)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwDataSize == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwDataSize = ((NdefRecord*)hNDEFRecord)->getEncodedSize();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetTnf(
IN HANDLE hNDEFRecord,
OUT LPBYTE  pucTnf)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pucTnf == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pucTnf = ((NdefRecord*)hNDEFRecord)->getTnf();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_SetTnf(
IN HANDLE hNDEFRecord,
IN BYTE  ucTnf)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->setTnf(ucTnf);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetTypeLength(
IN HANDLE hNDEFRecord,
OUT LPDWORD  pdwTypeLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwTypeLength == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwTypeLength = ((NdefRecord*)hNDEFRecord)->getTypeLength();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetType(
IN HANDLE hNDEFRecord,
OUT LPBYTE  lpTypeData)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpTypeData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->getType(lpTypeData);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetTypeStr(
IN HANDLE hNDEFRecord,
OUT LPWSTR  lpwstrType)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrType == NULL)
		return SCARD_E_INVALID_PARAMETER;

	//CString strType = ((NdefRecord*)hNDEFRecord)->getType();
	std::string strType = ((NdefRecord*)hNDEFRecord)->getType();
	//CStringW strwType = strType;
	//std::wstring strwType = strType;
	std::wstring strwType(strType.begin(), strType.end());

	//LPWSTR lpwstrType = strwType.GetBuffer();
	//memcpy(lpwstrType, lpwstrType, wcslen(lpwstrType));
	//memcpy(lpwstrType, strwType.GetBuffer(), strwType.GetLength() + 1);
	//memcpy(lpwstrType, strwType.GetBuffer(), strwType.GetLength());
	//strwType.ReleaseBuffer();
	wcscpy(lpwstrType, strwType.c_str());

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_SetType(
IN HANDLE hNDEFRecord,
IN LPBYTE  lpTypeData,
IN DWORD dwTypeLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpTypeData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwTypeLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->setType(lpTypeData, dwTypeLength);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_SetTypeStr(
IN HANDLE hNDEFRecord,
IN LPWSTR  lpwstrType)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrType == NULL)
		return SCARD_E_INVALID_PARAMETER;

	//CStringW strwType = lpwstrType;
	std::wstring strwType = lpwstrType;
	//CString strType = strwType;
	//std::string strType = strwType;
	//LPSTR lpstrType = strType.GetBuffer();

	wchar_t const * utf16_string = lpwstrType;
	size_t srclen = wcslen(utf16_string);
	char *utf8_string = new char[srclen];
	size_t dstlen = srclen;
	iconv_t conv = iconv_open("UTF-8", "UTF-16");
	iconv(conv, (char**)&utf16_string, &srclen, &utf8_string, &dstlen);
	iconv_close(conv);
	
	//((NdefRecord*)hNDEFRecord)->setType((LPBYTE)lpstrType, strlen(lpstrType));
	((NdefRecord*)hNDEFRecord)->setType((LPBYTE)utf8_string, strlen(utf8_string));

	//strType.ReleaseBuffer();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetPayloadLength(
IN HANDLE hNDEFRecord,
OUT LPDWORD  pdwPayloadLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwPayloadLength == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwPayloadLength = ((NdefRecord*)hNDEFRecord)->getPayloadLength();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetPayload(
IN HANDLE hNDEFRecord,
OUT LPBYTE  lpPayloadData)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpPayloadData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->getPayload(lpPayloadData);

	return SCARD_S_SUCCESS;
}

//CLSCRF_API LONG CLSCRF_DECL
//CLSCRF_NFC_NDEF_Record_GetPayloadText(
//IN HANDLE hNDEFRecord,
//OUT LPWSTR  lpwstrPayload)
//{
//	if (hNDEFRecord == NULL)
//		return SCARD_E_INVALID_PARAMETER;
//
//	if (lpwstrPayload == NULL)
//		return SCARD_E_INVALID_PARAMETER;
//
//	//CString strPayload;;
//	TCHAR lpStrPayload[1024] = { 0 };
//	((NdefRecord*)hNDEFRecord)->getPayload((uint8_t*)lpStrPayload);
//	CStringW strwPayload = lpStrPayload;
//
//	//LPWSTR lpwstrType = strwType.GetBuffer();
//	//memcpy(lpwstrType, lpwstrType, wcslen(lpwstrType));
//	memcpy(lpwstrPayload, strwPayload.GetBuffer(), strwPayload.GetLength() + 1);
//	strwPayload.ReleaseBuffer();
//
//	return SCARD_S_SUCCESS;
//}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetPayloadText(
IN HANDLE hNDEFRecord,
OUT LPWSTR  lpwstrText)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrText == NULL)
		return SCARD_E_INVALID_PARAMETER;

	char lpStrPayload[1024] = { 0 };
	((NdefRecord*)hNDEFRecord)->getPayload((uint8_t*)lpStrPayload);

	// your "Unicode" string
	//wchar_t const * utf16_string = L"Hello, World!";
	std::string utf8_string(lpStrPayload);

	//std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> convert;

	//std::wstring utf16_string = convert.from_bytes(utf8_string);
	
	std::wstring utf16_string(utf8_string.begin(), utf8_string.end());

	wcscpy(lpwstrText, utf16_string.c_str());

	return SCARD_S_SUCCESS;
}

char *uri_identifier_code(int b)
{
	/*
	* Section 3.2.2 "URI Identifier Code" of "URI Record Type Definition
	* Technical Specification"
	*/
	switch (b) {
	case 0x00:
		return (char*)"";
	case 0x01:
		return (char*)"http://www.";
	case 0x02:
		return (char*)"https://www.";
	case 0x03:
		return (char*)"http://";
	case 0x04:
		return (char*)"https://";
	case 0x05:
		return (char*)"tel:";
	case 0x06:
		return (char*)"mailto:";
	case 0x07:
		return (char*)"ftp://anonymous:anonymous@";
	case 0x08:
		return (char*)"ftp://ftp.";
	case 0x09:
		return (char*)"ftps://";
	case 0x0A:
		return (char*)"sftp://";
	case 0x0B:
		return (char*)"smb://";
	case 0x0C:
		return (char*)"nfs://";
	case 0x0D:
		return (char*)"ftp://";
	case 0x0E:
		return (char*)"dav://";
	case 0x0F:
		return (char*)"news:";
	case 0x10:
		return (char*)"telnet://";
	case 0x11:
		return (char*)"imap:";
	case 0x12:
		return (char*)"rtsp://";
	case 0x13:
		return (char*)"urn:";
	case 0x14:
		return (char*)"pop:";
	case 0x15:
		return (char*)"sip:";
	case 0x16:
		return (char*)"sips:";
	case 0x17:
		return (char*)"tftp:";
	case 0x18:
		return (char*)"btspp://";
	case 0x19:
		return (char*)"btl2cap://";
	case 0x1A:
		return (char*)"btgoep://";
	case 0x1B:
		return (char*)"tcpobex://";
	case 0x1C:
		return (char*)"irdaobex://";
	case 0x1D:
		return (char*)"file://";
	case 0x1E:
		return (char*)"urn:epc:id:";
	case 0x1F:
		return (char*)"urn:epc:tag:";
	case 0x20:
		return (char*)"urn:epc:pat:";
	case 0x21:
		return (char*)"urn:epc:raw:";
	case 0x22:
		return (char*)"urn:epc:";
	case 0x23:
		return (char*)"urn:nfc:";
	default:
		return (char*)"RFU";
	}
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetPayloadUri(
IN HANDLE hNDEFRecord,
OUT LPWSTR  lpwstrUri)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrUri == NULL)
		return SCARD_E_INVALID_PARAMETER;

	char lpStrPayload[1024] = { 0 };
	((NdefRecord*)hNDEFRecord)->getPayload((uint8_t*)lpStrPayload);

	// your "Unicode" string
	//wchar_t const * utf16_string = L"Hello, World!";
	std::string utf8_string(lpStrPayload + 1);
	std::string utf8_string0 = uri_identifier_code(lpStrPayload[0]);
	utf8_string = utf8_string0 + utf8_string;
	
	//std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> convert;

	//std::wstring utf16_string = convert.from_bytes(utf8_string);

	std::wstring utf16_string(utf8_string.begin(), utf8_string.end());

	wcscpy(lpwstrUri, utf16_string.c_str());

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_SetPayload(
IN HANDLE hNDEFRecord,
IN LPBYTE  lpPayloadData,
IN DWORD dwPayloadLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpPayloadData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwPayloadLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->setPayload(lpPayloadData, dwPayloadLength);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetIdLength(
IN HANDLE hNDEFRecord,
OUT LPDWORD  pdwIdLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwIdLength == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pdwIdLength = ((NdefRecord*)hNDEFRecord)->getIdLength();

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetId(
IN HANDLE hNDEFRecord,
OUT LPBYTE  lpIdData)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpIdData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->getId(lpIdData);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_GetIdStr(
IN HANDLE hNDEFRecord,
OUT LPWSTR lpwstrId)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpwstrId == NULL)
		return SCARD_E_INVALID_PARAMETER;

	//CString strId = ((NdefRecord*)hNDEFRecord)->getId();
	std::string strId = ((NdefRecord*)hNDEFRecord)->getId();
	//CStringW strwId = strId;
	//std::wstring strwId = strId;
	std::wstring strwId(strId.begin(), strId.end());

	//LPWSTR lpwstrId = strwId.GetBuffer();
	//memcpy(plpwstrId, lpwstrId, wcslen(lpwstrId));
	//strwId.ReleaseBuffer();

	wcscpy(lpwstrId, strwId.c_str());

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_SetId(
IN HANDLE hNDEFRecord,
IN LPBYTE  lpIdData,
IN DWORD dwIdLength)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpIdData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwIdLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NdefRecord*)hNDEFRecord)->setId(lpIdData, dwIdLength);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_NDEF_Record_Destroy(
IN HANDLE hNDEFRecord)
{
	if (hNDEFRecord == 0)
		return SCARD_E_INVALID_PARAMETER;

	delete (NdefRecord*)hNDEFRecord;

	return SCARD_S_SUCCESS;
}




CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_Create(IN LPVOID pReader,
OUT LPHANDLE phLLCPLink)
{
	if (phLLCPLink == NULL)
		return SCARD_E_INVALID_PARAMETER;

	NFCReader *pNfcReader = new NFCReader;
	pNfcReader->m_pReader = pReader;
	*phLLCPLink = new NFCLinkLayer(pNfcReader);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_SetTimeout(
IN HANDLE hLLCPLink,
IN DWORD dwTimeoutMs)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	((NFCLinkLayer*)hLLCPLink)->_dwTimeoutMs = dwTimeoutMs;

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_SetGeneralInformation(
IN HANDLE hLLCPLink,
IN PBYTE lpGI,
IN DWORD dwGILength)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpGI == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (sizeof(((NFCLinkLayer*)hLLCPLink)->_nfcReader->m_Gi) < dwGILength)
		return SCARD_E_INSUFFICIENT_BUFFER;

	memcpy(((NFCLinkLayer*)hLLCPLink)->_nfcReader->m_Gi, lpGI, dwGILength);

	((NFCLinkLayer*)hLLCPLink)->_nfcReader->m_dwGiLength = dwGILength;

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_SetNFCID3Information(
IN HANDLE hLLCPLink,
IN PBYTE lpNFCID3i)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpNFCID3i == NULL)
		return SCARD_E_INVALID_PARAMETER;

	memcpy(((NFCLinkLayer*)hLLCPLink)->_nfcReader->m_Nfcid3i, lpNFCID3i, 10);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_SetBaudrate(
IN HANDLE hLLCPLink,
IN BYTE ucDsiDri)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (ucDsiDri > 2)
		return SCARD_E_INVALID_PARAMETER;

	((NFCLinkLayer*)hLLCPLink)->_nfcReader->m_ucDsiDri = ucDsiDri;

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ServerOpen(
IN HANDLE hLLCPLink)
{
	if (hLLCPLink == NULL)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->openSNEPServerLink();
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ServerReceive(
IN HANDLE hLLCPLink,
OUT LPBYTE  lpData,
OUT LPDWORD lpdwDataLength)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->serverLinkRxData(lpData, (uint32_t*)lpdwDataLength);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ServerClose(
IN HANDLE hLLCPLink)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->closeSNEPServerLink();
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ClientOpen(
IN HANDLE hLLCPLink)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->openSNEPClientLink();
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ClientTransmit(
IN HANDLE hLLCPLink,
IN LPBYTE  lpData,
IN DWORD dwDataLength)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwDataLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->clientLinkTxData(lpData, dwDataLength);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_ClientClose(
IN HANDLE hLLCPLink)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	return ((NFCLinkLayer*)hLLCPLink)->closeSNEPClientLink();
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_LLCP_Destroy(
IN HANDLE hLLCPLink)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	delete ((NFCLinkLayer*)hLLCPLink)->_nfcReader;
	delete (NFCLinkLayer*)hLLCPLink;

	return SCARD_S_SUCCESS;
}




CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_SNEP_Create(
IN HANDLE hLLCPLink,
OUT LPHANDLE phSNEP)
{
	if (hLLCPLink == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (phSNEP == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*phSNEP = new SNEP((NFCLinkLayer*)hLLCPLink);

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_SNEP_Receive(
IN HANDLE hSNEP,
OUT LPBYTE  lpData,
OUT LPDWORD pdwDataLength)
{
	if (hSNEP == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	return ((SNEP*)hSNEP)->rxNDEFPayload(lpData, (uint32_t*)pdwDataLength);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_SNEP_Transmit(
IN HANDLE hSNEP,
IN LPBYTE  lpData,
IN DWORD dwDataLength)
{
	if (hSNEP == 0)
		return SCARD_E_INVALID_PARAMETER;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (dwDataLength == 0)
		return SCARD_E_INVALID_PARAMETER;

	return ((SNEP*)hSNEP)->pushPayload(lpData, dwDataLength);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_SNEP_Destroy(
IN HANDLE hSNEP)
{
	if (hSNEP == 0)
		return SCARD_E_INVALID_PARAMETER;

	delete (SNEP*)hSNEP;

	return SCARD_S_SUCCESS;
}



CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_ForumTags_SupposeTypeISO14443A(IN     LPVOID pReader,
IN LPBYTE  ATQ,
IN BYTE  SAK,
OUT LPBYTE pucTagType)
{
	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	if (ATQ == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pucTagType == NULL)
		return SCARD_E_INVALID_PARAMETER;

	*pucTagType = 0xFF;

	if ((SAK & 0x44) == 0x40)
		*pucTagType = 0;

	if (SAK == 0x00 && ATQ[0] == 0x44 && ATQ[1] == 0x00)
		*pucTagType = 2;

	if (SAK == 0x20 && ATQ[0] == 0x44 && ATQ[1] == 0x03)
		*pucTagType = 4;

	return SCARD_S_SUCCESS;
}

LONG Transmit_TCL_APDU(IN PVOID pReader,
	IN LPCBYTE pbSendBuffer,
	IN DWORD cbSendLength,
	OUT LPBYTE pbRecvBuffer,
	IN OUT LPDWORD pcbRecvLength)
{
	CLSCRF_CARD* pCard = m_pCurCard;

	if (pCard == NULL)
	{
		*pcbRecvLength = 0;
		return SCARD_E_NO_SMARTCARD;
	}

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE] = { 0 };
	BYTE RecvBuffer[BUF_SIZE];
	DWORD RecvLength = sizeof(RecvBuffer);
	LONG Status = SCARD_S_SUCCESS;

	if (cbSendLength < 1)
	{
		*pcbRecvLength = 0;
		return SCARD_E_INVALID_PARAMETER;
	}

	SendBuffer[0] = 0x0a | pCard->BlockNumberBit;
	SendBuffer[1] = pCard->CID;
	memcpy(&SendBuffer[2], pbSendBuffer, cbSendLength);

	*pcbRecvLength = 0;

	Status = CLSCRF_DirectIO_Card(pReader,
		SendBuffer, cbSendLength + 2,
		pCard->Timeout,
		RecvBuffer, &RecvLength);
	if (Status)
	{
	}
	else if (RecvLength)
	{
		pCard->BlockNumberBit ^= 0x01;

		if (RecvLength > 2)
		{
			*pcbRecvLength = RecvLength - 2;
			memcpy(pbRecvBuffer, &RecvBuffer[2], *pcbRecvLength);
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_ForumTags_BeginType4(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	BYTE CID = 0;
	BYTE pbATS[37] = { 0 };

	//   T=CL
	Status = CLSCRF_ISO14443A_4_RATS(pReader, CID, pbATS);

	if (Status != SCARD_S_SUCCESS)
	{
		return Status;
	}

	//   
	Status = CLSCRF_ISO14443A_4_PPS(pReader, CID, 0);

	if (Status != SCARD_S_SUCCESS)
	{
		return Status;
	}

	//    T=CL
	Status = CLSCRF_Mfrc_Set_Rf_Mode(pReader, 0);

	if (Status != SCARD_S_SUCCESS)
	{
		return Status;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_ForumTags_Format(IN     LPVOID pReader,
IN BYTE ucTagType)
{
	LONG Status = SCARD_S_SUCCESS;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	switch (ucTagType)
	{
	case 2:
	{
		BYTE pRecBuff[1024] = { 0 };

		Status = CLSCRF_MifareUltralight_Read(pReader, 0x00, pRecBuff);

		if (Status != SCARD_S_SUCCESS)
		{
			return Status;
		}

		if (pRecBuff[12] == 0xE1)
		{
			return SCARD_E_CARD_UNSUPPORTED;
		}

		if (pRecBuff[12] != 0x00)
		{
			return SCARD_E_CARD_UNSUPPORTED;
		}

		int iTagDataPages = 0;
		for (byte nPage = 0; nPage < 60; nPage++)
		{
			Status = CLSCRF_MifareUltralight_Read(pReader, (BYTE)(4 * nPage + 4), pRecBuff);

			if (Status != 0)
				break;

			iTagDataPages += 4;
		}

		BYTE pSendBuff[] = { 0xE1, 0x10, (BYTE)floor((double)(iTagDataPages - 4) * 4 / 8), 0x00 };
		Status = CLSCRF_MifareUltralight_Write(pReader, 3, pSendBuff);

		if (Status != SCARD_S_SUCCESS)
		{
			return Status;
		}

		//CByteArray ucaData;
		std::vector<BYTE> ucaData;

		Tlv* pTlv = Tlv::createLockControlTlv();// new Tlv((BYTE)Tlv::TlvType::LockControl, &ucaLockControl);
		//CByteArray* pcba = pTlv->toByteArray();
		std::vector<BYTE>* pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba;
		delete pTlv;
		pTlv = Tlv::createNDEFMessageTlv({});// new Tlv((BYTE)Tlv::TlvType::NDEF, {});
		pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba; 
		delete pTlv;
		pTlv = Tlv::createTerminatorTlv();// new Tlv((BYTE)Tlv::TlvType::Terminator, {});
		pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba; 
		delete pTlv;

		BYTE* pBuff = ucaData.data();

		for (BYTE nPage = 0; nPage < (BYTE)floor((double)ucaData.size() / 4); nPage++)
		{
			Status = CLSCRF_MifareUltralight_Write(pReader, (BYTE)(4 + nPage), pBuff + 4 * nPage);
			if (Status != SCARD_S_SUCCESS)
			{
				//ucaData.FreeExtra();

				return Status;
			}
		}

		//ucaData.FreeExtra();
	}
		break;
	case 4:
	{
		BYTE SW1, SW2;

		BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x01, 0x00 };
		BYTE SelectAppRAPDU[1024] = { 0 };

		DWORD dwSendLength = sizeof(SelectAppCAPDU);
		DWORD dwRecvLength = sizeof(SelectAppRAPDU);
		LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			return Status;
		}
		if (dwRecvLength < 2)
			return SCARD_E_UNSUPPORTED_FEATURE;
		SW1 = SelectAppRAPDU[dwRecvLength - 2];
		SW2 = SelectAppRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x00, 0x00 };
			BYTE SelectAppRAPDU[1024] = { 0 };

			DWORD dwSendLength = sizeof(SelectAppCAPDU);
			DWORD dwRecvLength = sizeof(SelectAppRAPDU);
			LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
			if (Status != 0)
			{
				return Status;
			}
			if (dwRecvLength < 2)
				return SCARD_E_UNSUPPORTED_FEATURE;
			SW1 = SelectAppRAPDU[dwRecvLength - 2];
			SW2 = SelectAppRAPDU[dwRecvLength - 1];
			if (SW1 != 0x90 || SW2 != 0x00)
			{
				//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
				return SCARD_E_UNEXPECTED;
			}
		}

		BYTE SelectCCCAPDU[] = { 0x00, 0xA4, 0x00, 0x0C, 0x02, 0xE1, 0x03 };
		BYTE SelectCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectCCCAPDU);
		dwRecvLength = sizeof(SelectCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectCCCAPDU, dwSendLength, SelectCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectCCRAPDU[dwRecvLength - 2];
		SW2 = SelectCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE ReadCCCAPDU[] = { 0x00, 0xB0, 0x00, 0x00, 0x00 };
		BYTE ReadCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(ReadCCCAPDU);
		dwRecvLength = sizeof(ReadCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, ReadCCCAPDU, dwSendLength, ReadCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = ReadCCRAPDU[dwRecvLength - 2];
		SW2 = ReadCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		if (ReadCCRAPDU[13] != 0x00)
		{
			//toolStripStatusLabel2.Text = ":  NDEF     .";
			return SCARD_E_NO_ACCESS;
		}

		if (ReadCCRAPDU[14] != 0x00)
		{
			//toolStripStatusLabel2.Text = ":  NDEF     .";
			return SCARD_E_NO_ACCESS;
		}

		int iMaxLe = ReadCCRAPDU[3] * 256 + ReadCCRAPDU[4] - 2;
		int iMaxLc = ReadCCRAPDU[5] * 256 + ReadCCRAPDU[6];
		int iMaxNDEFSize = ReadCCRAPDU[11] * 256 + ReadCCRAPDU[12];

		BYTE SelectNDEFCAPDU[] = { 0x00, 0xA4, 0x00, 0x00, 0x02, ReadCCRAPDU[9], ReadCCRAPDU[10] };
		BYTE SelectNDEFRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectNDEFCAPDU);
		dwRecvLength = sizeof(SelectNDEFRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectNDEFCAPDU, dwSendLength, SelectNDEFRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectNDEFRAPDU[dwRecvLength - 2];
		SW2 = SelectNDEFRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE WriteNLENNDEFCAPDU[] = { 0x00, 0xD6, 0x00, 0x00, 0x05, 0x00, 0x03, 0xD0, 0x00, 0x00 };
		BYTE WriteNLENNDEFRAPDU[1024] = { 0 };
		dwSendLength = sizeof(WriteNLENNDEFCAPDU);
		dwRecvLength = sizeof(WriteNLENNDEFRAPDU);

		Status = Transmit_TCL_APDU(pReader, WriteNLENNDEFCAPDU, dwSendLength, WriteNLENNDEFRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = WriteNLENNDEFRAPDU[dwRecvLength - 2];
		SW2 = WriteNLENNDEFRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}
	}
		break;
	default:
		return SCARD_E_UNSUPPORTED_FEATURE;
	}

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_ForumTags_Read(IN     LPVOID pReader,
IN BYTE ucTagType,
OUT LPBYTE  lpData,
IN OUT LPDWORD lpdwDataSize,
OUT LPBYTE lpucPermissions)
{
	LONG Status = SCARD_S_SUCCESS;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (lpdwDataSize == NULL)
		return SCARD_E_INVALID_PARAMETER;


	switch (ucTagType)
	{
	case 2:
	{
		BYTE pRecBuff[1024] = { 0 };
		BYTE pDataBuff[1024] = { 0 };

		Status = CLSCRF_MifareUltralight_Read(pReader, 0x00, pRecBuff);

		if (Status != SCARD_S_SUCCESS)
		{
			return Status;
		}

		if (pRecBuff[12] != 0xE1)
		{
			return SCARD_E_CARD_UNSUPPORTED;
		}

		int iTagDataPages = pRecBuff[14] * 8 / 4;

		bool fReadOnly = (pRecBuff[15] == 0x0F);

		if (lpucPermissions != NULL)
			*lpucPermissions = fReadOnly ? 0x01 : 0x00;

		//CByteArray cbaData;
		std::vector<BYTE> cbaData((BYTE)ceil((double)iTagDataPages / 4) * 16);
		//cbaData.SetSize((BYTE)ceil((double)iTagDataPages / 4) * 16);
		for (BYTE nPage = 0; nPage < (BYTE)ceil((double)iTagDataPages / 4); nPage++)
		{
			Status = CLSCRF_MifareUltralight_Read(pReader, (BYTE)(4 * nPage + 4), pRecBuff);

			if (Status != SCARD_S_SUCCESS)
			{
				return Status;
			}

			memcpy(&pDataBuff[16 * nPage], pRecBuff, 16);

			for (int i = 0; i < 16; i++)
				cbaData[nPage * 16 + i] = pRecBuff[i];
			
			if (Tlv::fromByteArray(&cbaData, 0)->back()->type() == Tlv::TlvType::Terminator)
				break;
		}

		TlvList* ltlv = Tlv::fromByteArray(&cbaData, 0);

		//CByteArray* pcbaNDEF = NULL;
		std::vector<BYTE>* pcbaNDEF = NULL;

		//for each(Tlv* tlv in *ltlv)
		for (int i = 0; i < (*ltlv).size(); i++)
		{
			Tlv* tlv = (*ltlv)[i];
			if (tlv->type() != (BYTE)Tlv::TlvType::NDEF)
				continue;

			pcbaNDEF = tlv->value();

			if (pcbaNDEF->size() == 0)
				continue;
		}

		delete ltlv;

		if (!pcbaNDEF)
			return SCARD_E_CARD_UNSUPPORTED;

		if (pcbaNDEF->size() > *lpdwDataSize)
			return SCARD_E_INSUFFICIENT_BUFFER;

		if (pcbaNDEF->size() == 0)
		{
			*lpdwDataSize = 0;
		}
		else
		{
			memcpy(lpData, pcbaNDEF->data(), pcbaNDEF->size());
			//pcbaNDEF->FreeExtra();
			*lpdwDataSize = pcbaNDEF->size();
		}
	}
		break;
	case 4:
	{
		BYTE SW1, SW2;

		BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x01, 0x00 };
		BYTE SelectAppRAPDU[1024] = { 0 };

		DWORD dwSendLength = sizeof(SelectAppCAPDU);
		DWORD dwRecvLength = sizeof(SelectAppRAPDU);
		LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			return Status;
		}
		if (dwRecvLength < 2)
			return SCARD_E_UNSUPPORTED_FEATURE;
		SW1 = SelectAppRAPDU[dwRecvLength - 2];
		SW2 = SelectAppRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x00, 0x00 };
			BYTE SelectAppRAPDU[1024] = { 0 };

			DWORD dwSendLength = sizeof(SelectAppCAPDU);
			DWORD dwRecvLength = sizeof(SelectAppRAPDU);
			LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
			if (Status != 0)
			{
				return Status;
			}
			if (dwRecvLength < 2)
				return SCARD_E_UNSUPPORTED_FEATURE;
			SW1 = SelectAppRAPDU[dwRecvLength - 2];
			SW2 = SelectAppRAPDU[dwRecvLength - 1];
			if (SW1 != 0x90 || SW2 != 0x00)
			{
				//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
				return SCARD_E_UNEXPECTED;
			}
		}

		BYTE SelectCCCAPDU[] = { 0x00, 0xA4, 0x00, 0x0C, 0x02, 0xE1, 0x03 };
		BYTE SelectCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectCCCAPDU);
		dwRecvLength = sizeof(SelectCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectCCCAPDU, dwSendLength, SelectCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectCCRAPDU[dwRecvLength - 2];
		SW2 = SelectCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE ReadCCCAPDU[] = { 0x00, 0xB0, 0x00, 0x00, 0x00 };
		BYTE ReadCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(ReadCCCAPDU);
		dwRecvLength = sizeof(ReadCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, ReadCCCAPDU, dwSendLength, ReadCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = ReadCCRAPDU[dwRecvLength - 2];
		SW2 = ReadCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		if (ReadCCRAPDU[13] != 0x00)
		{
			//toolStripStatusLabel2.Text = " NDEF     .";
			return SCARD_E_NO_ACCESS;
		}

		if (lpucPermissions != NULL)
			*lpucPermissions = (ReadCCRAPDU[14] != 0x00) ? 0x01 : 0x00; // Read only?

		int iMaxLe = ReadCCRAPDU[3] * 256 + ReadCCRAPDU[4] - 2;
		int iMaxLc = ReadCCRAPDU[5] * 256 + ReadCCRAPDU[6];

		BYTE SelectNDEFCAPDU[] = { 0x00, 0xA4, 0x00, 0x00, 0x02, ReadCCRAPDU[9], ReadCCRAPDU[10] };
		BYTE SelectNDEFRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectNDEFCAPDU);
		dwRecvLength = sizeof(SelectNDEFRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectNDEFCAPDU, dwSendLength, SelectNDEFRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectNDEFRAPDU[dwRecvLength - 2];
		SW2 = SelectNDEFRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE ReadNLENCAPDU[] = { 0x00, 0xB0, 0x00, 0x00, 0x02 };
		BYTE ReadNLENRAPDU[1024] = { 0 };
		dwSendLength = sizeof(ReadNLENCAPDU);
		dwRecvLength = sizeof(ReadNLENRAPDU);

		Status = Transmit_TCL_APDU(pReader, ReadNLENCAPDU, dwSendLength, ReadNLENRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = ReadNLENRAPDU[dwRecvLength - 2];
		SW2 = ReadNLENRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		DWORD LenNDEF = (DWORD)ReadNLENRAPDU[0] * 256 + ReadNLENRAPDU[1];

		//_rawMsg = new byte[0];
		//CByteArray _rawMsg;
		std::vector<BYTE> _rawMsg;

		DWORD uiReadPos = 2;
		BYTE ucReadSize = 0;
		while (uiReadPos - 2 < LenNDEF)
		{
			if (LenNDEF - uiReadPos + 2 > iMaxLe)
				ucReadSize = iMaxLe;
			else
				ucReadSize = (BYTE)(LenNDEF - uiReadPos + 2);

			BYTE ReadNDEFCAPDU[] = { 0x00, 0xB0, (BYTE)((uiReadPos - (uiReadPos % 0x100)) / 0x100), (byte)(uiReadPos % 0x100), ucReadSize };
			BYTE ReadNDEFRAPDU[1024] = { 0 };
			dwSendLength = sizeof(ReadNDEFCAPDU);
			dwRecvLength = sizeof(ReadNDEFRAPDU);

			Status = Transmit_TCL_APDU(pReader, ReadNDEFCAPDU, dwSendLength, ReadNDEFRAPDU, &dwRecvLength);
			if (Status != 0)
			{
				//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
				return Status;
			}
			SW1 = ReadNDEFRAPDU[dwRecvLength - 2];
			SW2 = ReadNDEFRAPDU[dwRecvLength - 1];
			if (SW1 != 0x90 || SW2 != 0x00)
			{
				//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
				return SCARD_E_UNEXPECTED;
			}

			//CByteArray tmpArr;
			std::vector<BYTE> tmpArr(dwRecvLength - 2);
			//tmpArr.SetSize(dwRecvLength - 2);
			for (int i = 0; i < dwRecvLength - 2; i++)
				tmpArr[i] = ReadNDEFRAPDU[i];

			//_rawMsg.Append(tmpArr);
			_rawMsg.insert(_rawMsg.end(), tmpArr.begin(), tmpArr.end());

			uiReadPos += ucReadSize;
		}

		if (_rawMsg.size() > *lpdwDataSize)
			return SCARD_E_INSUFFICIENT_BUFFER;

		for (int i = 0; i < _rawMsg.size(); i++)
			lpData[i] = _rawMsg[i];

		*lpdwDataSize = _rawMsg.size();
	}
		break;
	default:
		return SCARD_E_UNSUPPORTED_FEATURE;
	}

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC_ForumTags_Write(IN     LPVOID pReader,
IN BYTE ucTagType,
IN LPBYTE  lpData,
IN DWORD dwDataSize)
{
	LONG Status = SCARD_S_SUCCESS;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	if (lpData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	//if (dwDataSize == 0)
	//	return SCARD_E_INVALID_PARAMETER;

	switch (ucTagType)
	{
	case 2:
	{
		//CByteArray ucaData;
		std::vector<BYTE> ucaData;
		//CByteArray ucaLockControl;
		std::vector<BYTE> ucaLockControl;
		ucaLockControl.push_back(0xA0);
		ucaLockControl.push_back(0x10);
		ucaLockControl.push_back(0x44);
		Tlv* pTlv = new Tlv((BYTE)Tlv::TlvType::LockControl, &ucaLockControl);
		//CByteArray* pcba = pTlv->toByteArray();
		std::vector<BYTE>* pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba;
		delete pTlv;
		//CByteArray ucaNDEF;
		std::vector<BYTE> ucaNDEF(dwDataSize);
		//ucaNDEF.SetSize(dwDataSize);
		for (DWORD i = 0; i < dwDataSize; i++)
			ucaNDEF[i] = lpData[i];
		pTlv = new Tlv((BYTE)Tlv::TlvType::NDEF, &ucaNDEF);
		pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba;
		delete pTlv;
		pTlv = new Tlv((BYTE)Tlv::TlvType::Terminator, {});
		pcba = pTlv->toByteArray();
		//ucaData.Append(*pcba);
		ucaData.insert(ucaData.end(), (*pcba).begin(), (*pcba).end());
		delete pcba;
		delete pTlv;

		BYTE* pBuff = ucaData.data();

		for (BYTE nPage = 0; nPage < (BYTE)ceil((double)ucaData.size() / 4); nPage++)
		{
			Status = CLSCRF_MifareUltralight_Write(pReader, (BYTE)(4 + nPage), pBuff + 4 * nPage);
			if (Status != SCARD_S_SUCCESS)
			{
				//ucaData.FreeExtra();

				return Status;
			}
		}

		//ucaData.FreeExtra();
	}
		break;
	case 4:
	{
		BYTE SW1, SW2;

		BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x01, 0x00 };
		BYTE SelectAppRAPDU[1024] = { 0 };

		DWORD dwSendLength = sizeof(SelectAppCAPDU);
		DWORD dwRecvLength = sizeof(SelectAppRAPDU);
		LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			return Status;
		}
		if (dwRecvLength < 2)
			return SCARD_E_UNSUPPORTED_FEATURE;
		SW1 = SelectAppRAPDU[dwRecvLength - 2];
		SW2 = SelectAppRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			BYTE SelectAppCAPDU[] = { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x00, 0x00 };
			BYTE SelectAppRAPDU[1024] = { 0 };

			DWORD dwSendLength = sizeof(SelectAppCAPDU);
			DWORD dwRecvLength = sizeof(SelectAppRAPDU);
			LONG Status = Transmit_TCL_APDU(pReader, SelectAppCAPDU, dwSendLength, SelectAppRAPDU, &dwRecvLength);
			if (Status != 0)
			{
				return Status;
			}
			if (dwRecvLength < 2)
				return SCARD_E_UNSUPPORTED_FEATURE;
			SW1 = SelectAppRAPDU[dwRecvLength - 2];
			SW2 = SelectAppRAPDU[dwRecvLength - 1];
			if (SW1 != 0x90 || SW2 != 0x00)
			{
				//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
				return SCARD_E_UNEXPECTED;
			}
		}

		BYTE SelectCCCAPDU[] = { 0x00, 0xA4, 0x00, 0x0C, 0x02, 0xE1, 0x03 };
		BYTE SelectCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectCCCAPDU);
		dwRecvLength = sizeof(SelectCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectCCCAPDU, dwSendLength, SelectCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectCCRAPDU[dwRecvLength - 2];
		SW2 = SelectCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE ReadCCCAPDU[] = { 0x00, 0xB0, 0x00, 0x00, 0x00 };
		BYTE ReadCCRAPDU[1024] = { 0 };
		dwSendLength = sizeof(ReadCCCAPDU);
		dwRecvLength = sizeof(ReadCCRAPDU);

		Status = Transmit_TCL_APDU(pReader, ReadCCCAPDU, dwSendLength, ReadCCRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = ReadCCRAPDU[dwRecvLength - 2];
		SW2 = ReadCCRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		if (ReadCCRAPDU[13] != 0x00)
		{
			//toolStripStatusLabel2.Text = ":  NDEF     .";
			return SCARD_E_NO_ACCESS;
		}

		if (ReadCCRAPDU[14] != 0x00)
		{
			//toolStripStatusLabel2.Text = ":  NDEF     .";
			return SCARD_E_NO_ACCESS;
		}

		int iMaxLe = ReadCCRAPDU[3] * 256 + ReadCCRAPDU[4] - 2;
		int iMaxLc = ReadCCRAPDU[5] * 256 + ReadCCRAPDU[6];
		int iMaxNDEFSize = ReadCCRAPDU[11] * 256 + ReadCCRAPDU[12];

		if (iMaxNDEFSize < dwDataSize)
		{
			//toolStripStatusLabel2.Text = ":    NDEF.";
			return SCARD_E_NO_MEMORY;
		}

		BYTE SelectNDEFCAPDU[] = { 0x00, 0xA4, 0x00, 0x00, 0x02, ReadCCRAPDU[9], ReadCCRAPDU[10] };
		BYTE SelectNDEFRAPDU[1024] = { 0 };
		dwSendLength = sizeof(SelectNDEFCAPDU);
		dwRecvLength = sizeof(SelectNDEFRAPDU);

		Status = Transmit_TCL_APDU(pReader, SelectNDEFCAPDU, dwSendLength, SelectNDEFRAPDU, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = SelectNDEFRAPDU[dwRecvLength - 2];
		SW2 = SelectNDEFRAPDU[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		BYTE WriteNLENCAPDU1[] = { 0x00, 0xD6, 0x00, 0x00, 0x02, 0x00, 0x00 };
		BYTE WriteNLENRAPDU1[1024] = { 0 };
		dwSendLength = sizeof(WriteNLENCAPDU1);
		dwRecvLength = sizeof(WriteNLENRAPDU1);

		Status = Transmit_TCL_APDU(pReader, WriteNLENCAPDU1, dwSendLength, WriteNLENRAPDU1, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = WriteNLENRAPDU1[dwRecvLength - 2];
		SW2 = WriteNLENRAPDU1[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}

		DWORD uiWritePos = 2;
		BYTE ucWriteSize = 0;
		while (uiWritePos - 2 < dwDataSize)
		{
			if (dwDataSize - uiWritePos + 2 > iMaxLc)
				ucWriteSize = iMaxLc;
			else
				ucWriteSize = (BYTE)(dwDataSize - uiWritePos + 2);

			BYTE WriteNDEFCAPDU[] = { 0x00, 0xD6, (BYTE)((uiWritePos - (uiWritePos % 0x100)) / 0x100), (byte)(uiWritePos % 0x100), ucWriteSize };
			BYTE* lpWriteNDEFCAPDU = new BYTE[sizeof(WriteNDEFCAPDU) + ucWriteSize];
			memcpy(lpWriteNDEFCAPDU, WriteNDEFCAPDU, sizeof(WriteNDEFCAPDU));
			memcpy(&lpWriteNDEFCAPDU[sizeof(WriteNDEFCAPDU)], &lpData[uiWritePos - 2], ucWriteSize);
			BYTE WriteNDEFRAPDU[1024] = { 0 };
			dwSendLength = sizeof(WriteNDEFCAPDU) + ucWriteSize;
			dwRecvLength = sizeof(WriteNDEFRAPDU);

			Status = Transmit_TCL_APDU(pReader, lpWriteNDEFCAPDU, dwSendLength, WriteNDEFRAPDU, &dwRecvLength);
			delete lpWriteNDEFCAPDU;
			if (Status != 0)
			{
				//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
				return Status;
			}
			SW1 = WriteNDEFRAPDU[dwRecvLength - 2];
			SW2 = WriteNDEFRAPDU[dwRecvLength - 1];
			if (SW1 != 0x90 || SW2 != 0x00)
			{
				//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
				return SCARD_E_UNEXPECTED;
			}

			uiWritePos += ucWriteSize;
		}

		BYTE WriteNLENCAPDU2[] = { 0x00, 0xD6, 0x00, 0x00, 0x02, (BYTE)floor((double)dwDataSize / 256), (BYTE)(dwDataSize % 256) };
		BYTE WriteNLENRAPDU2[1024] = { 0 };
		dwSendLength = sizeof(WriteNLENCAPDU2);
		dwRecvLength = sizeof(WriteNLENRAPDU2);

		Status = Transmit_TCL_APDU(pReader, WriteNLENCAPDU2, dwSendLength, WriteNLENRAPDU2, &dwRecvLength);
		if (Status != 0)
		{
			//toolStripStatusLabel2.Text = ": 0x" + Status.ToString("X");
			return Status;
		}
		SW1 = WriteNLENRAPDU2[dwRecvLength - 2];
		SW2 = WriteNLENRAPDU2[dwRecvLength - 1];
		if (SW1 != 0x90 || SW2 != 0x00)
		{
			//toolStripStatusLabel2.Text = " SW1: 0x" + SW1.ToString("X") + " SW2: 0x" + SW2.ToString("X");
			return SCARD_E_UNEXPECTED;
		}
	}
		break;
	default:
		return SCARD_E_UNSUPPORTED_FEATURE;
	}

	return SCARD_S_SUCCESS;
}
