#include "StdAfx.h"

#include "Clscrfl.h"

extern std::map<PVOID, READER_SETTINGS> m_ReaderSettings;

extern CLSCRF_CARD m_DefaultCard;

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_GetLastInternalError(IN     LPVOID pReader,
OUT LPBYTE pbError)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		DWORD Size = 1;

		if (!pTr->m_RcBuf.GetBulk(0, pbError, &Size))
			Status = SCARD_E_COMM_DATA_LOST;
	}
	else
	{
		*pbError = 0xff;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_GetState(IN     LPVOID  pReader,
OUT LPDWORD pdwState)
{
	LONG Status = SCARD_S_SUCCESS;

	*pdwState = 0;
	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x0e);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwState, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Mfrc_On(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x10);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Mfrc_Off(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x04);
		pTr->m_TrBuf.AddByte(0x80);
		pTr->m_TrBuf.AddByte(0x01);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Get_Mfrc_Version(IN     LPVOID pReader,
OUT LPBYTE pbMfrcVersion)
{
	LONG Status = SCARD_S_SUCCESS;
// 	0x1F13C5731877
// 	0x0123456789AB
// 
// 	0, 1 - поддерживаемые карты : 7 бит - рфу, 6 бит - 18000, 5 бит - epc / uid, 4 бит - 15693, 3 бит - 14443B, 2 бит - Ultrlight C, 1 бит - Mifare Plus, 0 бит - в целом ISO14443A(признак использования микросхемы считывателя)
// 	2 - high version
// 	3 - low version
// 	4 - 7 бит - наличие SAM AV2(&0x80)
// 	5 - 0 бит - USB, 1 бит - 485, 2 бит - 232. Если(SAM AV2 == 1), то 1 бит = 0, 2 бит - 232 либо 485
// 	6 - наличие звука(4 бит), 2 реле, подключающих питание(6 бит) и землю(5 бит), светодиод обычный 0, отдельно smd 1 (7 бит)
// 	7 - тип микросхемы : 3 - CLRC663, 2 - CLRC632, 1 - MFRC531, 0 - MFRC530
// 	8, 9 - версия микросхемы считывателя
	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x64);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = pTr->m_RcBuf.GetLength() - 1;
			//            Size = 5;
			if (!pTr->m_RcBuf.GetBulk(1, pbMfrcVersion, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
			else if (Size < 6)
				pbMfrcVersion[5] = 0;   // Firmware Version 0
		}
	}
	else
	{
		ZeroMemory(pbMfrcVersion, 5);
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Get_Mfrc_Serial_Number(IN     LPVOID pReader,
OUT LPBYTE pbMfrcSN)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x22);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 4;
			if (!pTr->m_RcBuf.GetBulk(1, pbMfrcSN, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		ZeroMemory(pbMfrcSN, 4);
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Get_ReaderID(IN     LPVOID pReader,
OUT LPBYTE pbReaderID)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x58);
		pTr->m_TrBuf.AddByte(0x58);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 16;
			if (!pTr->m_RcBuf.GetBulk(1, pbReaderID, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		ZeroMemory(pbReaderID, 16);
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Reader_Restart(IN     LPVOID pReader,
IN LPBYTE pbReaderID)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x58);
		pTr->m_TrBuf.AddByte(0x77);
		pTr->m_TrBuf.AddBulk(pbReaderID, 16);

		DWORD Size;
		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Mfrc_Rf_Off_On(IN LPVOID pReader,
IN USHORT usDelay)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x20);
		pTr->m_TrBuf.AddBulk((BYTE*)&usDelay, 2);

		DWORD OldTimeout = pTr->GetTimeout();
		DWORD NewTimeout = OldTimeout + usDelay;
		pTr->SetTimeout(NewTimeout);
		Status = pTr->SingleTransceive();
		pTr->SetTimeout(OldTimeout);
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Mfrc_Set_Rf_Mode(IN  LPVOID pReader,
IN  BYTE   bRfMode)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x51);
		pTr->m_TrBuf.AddByte(bRfMode);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443A_ActivateEx(IN     LPVOID  pReader,
IN BYTE ucCID,
IN BYTE ucBaudrate,
IN BYTE ucRfOffTime,
IN BYTE ucRfOnPause,
IN BYTE ucDisableTCLSwitch,

OUT LPBYTE  pbATQ,
OUT LPBYTE  pucSAK,
OUT LPBYTE  pbATS,
OUT LPBYTE  pbUID,
IN OUT LPBYTE pucUIDLength
)
{
	return CLSCRF_ISO14443_ActivateEx(
		pReader, 
		0x00, 
		ucCID, 
		ucBaudrate, 
		ucRfOffTime, 
		ucRfOnPause, 
		ucDisableTCLSwitch, 
		0x00, 
		0x00, 
		0x00, 
		pbATQ, 
		pucSAK, 
		pbATS, 
		pbUID, 
		pucUIDLength, 
		NULL, 
		NULL, 
		NULL);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443B_ActivateEx(IN     LPVOID  pReader,
IN BYTE ucCID,
IN BYTE ucBaudrate,
IN BYTE ucRfOffTime,
IN BYTE ucRfOnPause,
IN BYTE ucDisableTCLSwitch,

IN BYTE ucAFI,
IN BYTE ucExtATQ,
IN BYTE ucSni,

OUT LPBYTE pucPUPI,
OUT LPBYTE pucAppData,
OUT LPBYTE pucProtInfo,
OUT LPBYTE pucProtInfoLength,

OUT LPBYTE  pucMBLI)
{
	BYTE ATQB[256] = { 0 };
	BYTE ucATQB_Len = 255;//256;

	ULONG res = CLSCRF_ISO14443_ActivateEx(
		pReader,
		0x01,
		ucCID,
		ucBaudrate,
		ucRfOffTime,
		ucRfOnPause,
		ucDisableTCLSwitch,
		ucAFI,
		ucExtATQ,
		ucSni,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL,
		pucMBLI,
		ATQB,
		&ucATQB_Len);

	if (res != SCARD_S_SUCCESS)
		return res;

	memcpy(pucPUPI, &ATQB[1], 4);
	memcpy(pucAppData, &ATQB[1+4], 4);
	if (*pucProtInfoLength < ucATQB_Len - 9)
		return SCARD_E_INSUFFICIENT_BUFFER;
	*pucProtInfoLength = ucATQB_Len - 9;
	memcpy(pucProtInfo, &ATQB[1+4+4], *pucProtInfoLength);

	return res;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443_ActivateEx(IN     LPVOID  pReader,
IN BYTE ucCardStandard,
IN BYTE ucCID,
IN BYTE ucBaudrate,
IN BYTE ucRfOffTime,
IN BYTE ucRfOnPause,
IN BYTE ucDisableTCLSwitch,

IN BYTE ucAFI,
IN BYTE ucExtATQ,
IN BYTE ucSni,

OUT LPBYTE  pbATQ,
OUT LPBYTE  pucSAK,
OUT LPBYTE  pbATS,
OUT LPBYTE  pbUID,
IN OUT LPBYTE pucUIDLength,

OUT LPBYTE  pucMBLI,
OUT LPBYTE  pbATQB,
IN OUT LPBYTE pucATQBLength
)
{
	LONG Status = SCARD_E_INVALID_PARAMETER;

	if (ucCardStandard > 1)
		return SCARD_E_INVALID_PARAMETER;

	if (ucCardStandard == 0)        // ISO 14443-A
	{
		if ((pbATQ == NULL) || (pucSAK == NULL) || (pbUID == NULL) || (pucUIDLength == NULL))
			return SCARD_E_INVALID_PARAMETER;

		if (ucDisableTCLSwitch == 0)    // ISO 14443-4
		{
			if (pbATS == NULL)
				return SCARD_E_INVALID_PARAMETER;
		}
	}
	else                            // ISO 14443-B
	{
		if ((pucMBLI == NULL) || (pbATQB == NULL) || (pucATQBLength == NULL))
			return SCARD_E_INVALID_PARAMETER;
	}

	if (pbATQ != NULL)
		ZeroMemory(pbATQ, 2);
	if (pucSAK != NULL)
		*pucSAK = 0;
	if (pbATS != NULL)
		pbATS[0] = 0;
	if (pucUIDLength != NULL)
		*pucUIDLength = 0;
	if (pucMBLI != NULL)
		*pucMBLI = 0;
	if (pucATQBLength != NULL)
		*pucATQBLength = 0;

	if (pReader != NULL)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x75);
		pTr->m_TrBuf.AddByte((ucCardStandard << 4) | (ucBaudrate & 0x0F));
		pTr->m_TrBuf.AddByte((ucRfOffTime << 4) | (ucRfOnPause & 0x0F));
		pTr->m_TrBuf.AddByte((ucDisableTCLSwitch << 7) | (ucCID & 0x0F));
		if (ucCardStandard == 0x01)
		{
			pTr->m_TrBuf.AddByte(ucAFI);
			pTr->m_TrBuf.AddByte(((ucExtATQ << 4) & 0x10) | (ucSni & 0x07));
		}

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status != SCARD_S_SUCCESS)
			return Status;

		if (ucCardStandard == 0x00)
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, pbATQ, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(3, pucSAK, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(4, (LPBYTE)pucUIDLength, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = *pucUIDLength;
			if (!pTr->m_RcBuf.GetBulk(5, pbUID, &Size))
				return SCARD_E_COMM_DATA_LOST;

			if (ucDisableTCLSwitch == 0x00)
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(5 + *pucUIDLength, pbATS, &Size))
					return SCARD_E_COMM_DATA_LOST;

				Size = pbATS[0];
				if (!pTr->m_RcBuf.GetBulk(5 + *pucUIDLength, pbATS, &Size))
					return SCARD_E_COMM_DATA_LOST;
			}
			m_DefaultCard.Timeout = 6800;     // Timeout = 64 ms
			m_DefaultCard.CardType = CLSCRF_ICC_TYPE_ISO14443A;
			m_DefaultCard.UID_Len = *pucUIDLength;
			memcpy(&m_DefaultCard.UID, pbUID, *pucUIDLength);
			memcpy(&m_DefaultCard.ATQ, pbATQ, 2);
			m_DefaultCard.ATQ.SAK = *pucSAK;
			m_DefaultCard.Active = 1;
			m_DefaultCard.RfMode = 0x00;//RF_14443_A;
			m_DefaultCard.BlockNumberBit = 0;
		}
		else if (ucCardStandard == 0x01)
		{
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(1, pucMBLI, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(2, (LPBYTE)pucATQBLength, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = *pucATQBLength;
			if (!pTr->m_RcBuf.GetBulk(3, pbATQB, &Size))
				return SCARD_E_COMM_DATA_LOST;

			m_DefaultCard.Timeout = 6800;     // Timeout = 64 ms
			m_DefaultCard.CardType = CLSCRF_ICC_TYPE_ISO14443B;
			m_DefaultCard.UID_Len = 4;
			if (*pucATQBLength > 8)
			{
				memcpy(&m_DefaultCard.UID, pbATQB, 4);
				memcpy(&m_DefaultCard.ATQ.AppData, &pbATQB[4], 4);
				memcpy(&m_DefaultCard.ATQ.ProtInf, &pbATQB[8], *pucATQBLength - 8);
			}
			m_DefaultCard.Active = 1;
			m_DefaultCard.RfMode = 0x10;//RF_14443_B;
			m_DefaultCard.BlockNumberBit = 0;
		}
	}
	else
		return SCARD_E_READER_UNAVAILABLE;

	return Status;
}

/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443_ActivateEx(IN     LPVOID  pReader,
IN BYTE ucCardStandard,
IN BYTE ucCID,
IN BYTE ucBaudrate,
IN BYTE ucRfOffTime,
IN BYTE ucRfOnPause,
IN BYTE ucDisableTCLSwitch,

IN BYTE ucAFI,
IN BYTE ucExtATQ,
IN BYTE ucSni,

OUT LPBYTE  pbATQ,
OUT LPBYTE  pucSAK,
OUT LPBYTE  pbATS,
OUT LPBYTE  pbUID,
IN OUT LPBYTE pucUIDLength,

OUT LPBYTE  pucMBLI,
OUT LPBYTE  pbATQB,
IN OUT LPBYTE pucATQBLength
)
{
	LONG Status = SCARD_E_INVALID_PARAMETER;

	if (ucDisableTCLSwitch == 0)
	{
		if (pbATS == NULL)
			return WSAEFAULT;
	}

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x75);
		pTr->m_TrBuf.AddByte((ucCardStandard << 4) | (ucBaudrate & 0x0F));
		pTr->m_TrBuf.AddByte((ucRfOffTime << 4) | (ucRfOnPause & 0x0F));
		pTr->m_TrBuf.AddByte((ucDisableTCLSwitch << 7) | (ucCID & 0x0F));
		if (ucCardStandard == 0x01)
		{
			pTr->m_TrBuf.AddByte(ucAFI);
			pTr->m_TrBuf.AddByte(((ucExtATQ << 4) & 0x10) | (ucSni & 0x07));
		}

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status != SCARD_S_SUCCESS)
			return Status;

		if (ucCardStandard == 0x00)
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, pbATQ, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(3, pucSAK, &Size))
				return SCARD_E_COMM_DATA_LOST;

			*pucUIDLength = 0; //kbhSCARD_E_INSUFFICIENT_BUFFER
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(4, (LPBYTE)pucUIDLength, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = *pucUIDLength;
			if (!pTr->m_RcBuf.GetBulk(5, pbUID, &Size))
				return SCARD_E_COMM_DATA_LOST;

			pbATS[0] = 0;
			if (ucDisableTCLSwitch == 0x00)
			{
				Size = 1;
				if (!pTr->m_RcBuf.GetBulk(5 + *pucUIDLength, pbATS, &Size))
					return SCARD_E_COMM_DATA_LOST;

				Size = pbATS[0];
				if (!pTr->m_RcBuf.GetBulk(5 + *pucUIDLength, pbATS, &Size))
					return SCARD_E_COMM_DATA_LOST;
			}
			m_DefaultCard.Timeout = 6800;     // Timeout = 64 ms
			m_DefaultCard.CardType = CLSCRF_ICC_TYPE_ISO14443A;
			m_DefaultCard.UID_Len = *pucUIDLength;
			memcpy(&m_DefaultCard.UID, pbUID, *pucUIDLength);
			memcpy(&m_DefaultCard.ATQ, pbATQ, 2);
			m_DefaultCard.ATQ.SAK = *pucSAK;
			m_DefaultCard.Active = 1;
			m_DefaultCard.RfMode = 0x00;//RF_14443_A;
			m_DefaultCard.BlockNumberBit = 0;
		}

		if (ucCardStandard == 0x01)
		{
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(1, pucMBLI, &Size))
				return SCARD_E_COMM_DATA_LOST;

			*pucATQBLength = 0; //kbhSCARD_E_INSUFFICIENT_BUFFER
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(2, (LPBYTE)pucATQBLength, &Size))
				return SCARD_E_COMM_DATA_LOST;

			Size = *pucATQBLength;
			if (!pTr->m_RcBuf.GetBulk(3, pbATQB, &Size))
				return SCARD_E_COMM_DATA_LOST;

			m_DefaultCard.Timeout = 6800;     // Timeout = 64 ms
			m_DefaultCard.CardType = CLSCRF_ICC_TYPE_ISO14443B;
			m_DefaultCard.UID_Len = 4;
			if (*pucATQBLength > 8)
			{
				memcpy(&m_DefaultCard.UID, pbATQB, 4);
				memcpy(&m_DefaultCard.ATQ.AppData, &pbATQB[4], 4);
				memcpy(&m_DefaultCard.ATQ.ProtInf, &pbATQB[8], *pucATQBLength - 8);
			}
			m_DefaultCard.Active = 1;
			m_DefaultCard.RfMode = 0x10;//RF_14443_B;
			m_DefaultCard.BlockNumberBit = 0;
		}
	}
	else
	{
		ZeroMemory(pbATQ, 2);
		*pucSAK = 0;
		*pucMBLI = 0;
		ZeroMemory(pbUID, 10);

		return SCARD_E_READER_UNAVAILABLE;
	}


	return Status;
}
*/
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ISO14443_4_Exchange(LPVOID  pReader,
IN DWORD   dwOption,
IN LPCBYTE pbSendBuffer,
IN DWORD   dwSendLength,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (dwSendLength == 0)
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_INVALID_PARAMETER;
	}
	else if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x47);
		pTr->m_TrBuf.AddBulk((LPBYTE)&dwOption, 2);
		pTr->m_TrBuf.AddBulk((LPBYTE)&dwSendLength, 2);
		pTr->m_TrBuf.AddBulk(pbSendBuffer, dwSendLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Sound(IN  LPVOID pReader,
IN  BYTE   nBeepCount)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x05);
		pTr->m_TrBuf.AddByte(nBeepCount);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}
/*
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//---------------------------------------------------------------------------
//  Зажечь/погасить светодиод
//---------------------------------------------------------------------------
CLSCRF_API LONG CLSCRF_DECL CLSCRF_Led( IN  LPVOID pReader,
IN  BYTE   bState )
{
LONG Status = SCARD_S_SUCCESS;

if( pReader )
{
CTransfer* pTr = (CTransfer*)pReader;
pTr->m_TrBuf.Clear();
pTr->m_TrBuf.AddByte( 0x06 );
pTr->m_TrBuf.AddByte( bState );

Status = pTr->SingleTransceive();
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}
*/
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Led(IN LPVOID pReader,
IN BYTE   bBlinkColor,
IN BYTE   bBlinkCount,
IN BYTE   bPostColor)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x07);
		pTr->m_TrBuf.AddByte(bBlinkColor);
		pTr->m_TrBuf.AddByte(bBlinkCount);
		pTr->m_TrBuf.AddByte(bPostColor);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Switches(IN LPVOID pReader,
IN BYTE   ucControl)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x37);
		pTr->m_TrBuf.AddByte(ucControl);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_UART_Baudrate(IN  LPVOID pReader,
IN  DWORD  dwBaudrate)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		Status = SCARD_E_READER_UNAVAILABLE;

	if (Status == SCARD_S_SUCCESS)
		switch (dwBaudrate)
	{
		case 9600:
		case 14400:
		case 19200:
		case 38400:
		case 57600:
		case 115200:
		case 250000:
		case 750000:
		case 1500000:
			break;

		default:
			Status = SCARD_E_INVALID_PARAMETER;
	}

	if (Status == SCARD_S_SUCCESS)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x0f);
		BYTE br = (BYTE)(F_OSC / (16 * dwBaudrate));
		pTr->m_TrBuf.AddByte(br);

		Status = pTr->SingleTransceive();
	}

	return Status;
}

//#if ADDRESS_RS_485
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_GetIOAddress(IN     LPVOID pReader,
OUT LPBYTE pbIOAddr)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		*pbIOAddr = pTr->GetIOAddress();
	}
	else
	{
		*pbIOAddr = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SetIOAddress(IN LPVOID pReader,
IN BYTE   bIOAddr)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->SetIOAddress(bIOAddr);
		m_ReaderSettings[pReader].SetIOAdderss(bIOAddr);
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_ReadDeviceAddress(IN     LPVOID pReader,
OUT LPBYTE pbDevAddr)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x7a);

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			Size = 1;
			if (!pTr->m_RcBuf.GetBulk(1, pbDevAddr, &Size))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		*pbDevAddr = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_WriteDeviceAddress(IN LPVOID pReader,
IN BYTE   bDevAddr)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x77);
		pTr->m_TrBuf.AddByte(bDevAddr);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SetIndicatorMessage(IN LPVOID pReader,
IN LPCSTR lpMessage,
IN BYTE  ucMessageLength,
IN DWORD dwInfiniteLoop,
IN DWORD dwLineIndex,
IN DWORD dwMode,
IN BYTE ucRepeats)
{
	LONG Status = SCARD_S_SUCCESS;

	if ((lpMessage == NULL) || (ucMessageLength == 0) || (ucMessageLength > 200))
		return SCARD_E_INVALID_PARAMETER;

	if ((dwMode == 0x00) || (dwMode == 0x03))
		return SCARD_E_INVALID_PARAMETER;

	CHAR szCodedMessage[301] = { 0 };

	CHAR cCodingMap[256] = { 0 };
	for (int i = 0; i < 256; i++)
		cCodingMap[i] = 0x20;
	for (int i = 0x20; i < 0x80; i++)
		cCodingMap[i] = i;
	cCodingMap[0x5C] = 0x2F;//"\" -> "/"
	cCodingMap[0x7B] = 0x28;//"{" -> "("
	cCodingMap[0x7C] = 0xD1;//"|" -> "|мал"
	cCodingMap[0x7D] = 0x29;//"}" -> ")"
	cCodingMap[0x7E] = 0x20;//"⌂" -> " "

	/*
	//cp?
	cCodingMap['Б'] = 0xA0; cCodingMap['Ю'] = 0xB0; cCodingMap['ч'] = 0xC0; cCodingMap['Д'] = 0xE0;
	cCodingMap['Г'] = 0xA1; cCodingMap['Я'] = 0xB1; cCodingMap['ш'] = 0xC1; cCodingMap['Ц'] = 0xE1;
	cCodingMap['Ё'] = 0xA2; cCodingMap['б'] = 0xB2; cCodingMap['ъ'] = 0xC2; cCodingMap['Щ'] = 0xE2;
	cCodingMap['Ж'] = 0xA3; cCodingMap['в'] = 0xB3; cCodingMap['ы'] = 0xC3; cCodingMap['д'] = 0xE3;
	cCodingMap['З'] = 0xA4; cCodingMap['г'] = 0xB4; cCodingMap['ь'] = 0xC4; cCodingMap['ф'] = 0xE4;
	cCodingMap['И'] = 0xA5; cCodingMap['ё'] = 0xB5; cCodingMap['э'] = 0xC5; cCodingMap['ц'] = 0xE5;
	cCodingMap['Й'] = 0xA6; cCodingMap['ж'] = 0xB6; cCodingMap['ю'] = 0xC6; cCodingMap['щ'] = 0xE6;
	cCodingMap['Л'] = 0xA7; cCodingMap['з'] = 0xB7; cCodingMap['я'] = 0xC7; cCodingMap['\''] = 0xE7;
	cCodingMap['П'] = 0xA8; cCodingMap['и'] = 0xB8;
	cCodingMap['У'] = 0xA9; cCodingMap['й'] = 0xB9;
	cCodingMap['Ф'] = 0xAA; cCodingMap['к'] = 0xBA;
	cCodingMap['Ч'] = 0xAB; cCodingMap['л'] = 0xBB;
	cCodingMap['Ш'] = 0xAC; cCodingMap['м'] = 0xBC; cCodingMap['№'] = 0xCC;
	cCodingMap['Ъ'] = 0xAD; cCodingMap['н'] = 0xBD;
	cCodingMap['Ы'] = 0xAE; cCodingMap['п'] = 0xBE;
	cCodingMap['Э'] = 0xAF; cCodingMap['т'] = 0xBF;
	*/

	//cp1251
	cCodingMap[0xC1] = 0xA0; cCodingMap[0xDE] = 0xB0; cCodingMap[0xF7] = 0xC0; cCodingMap[0xC4] = 0xE0;
	cCodingMap[0xC3] = 0xA1; cCodingMap[0xDF] = 0xB1; cCodingMap[0xF8] = 0xC1; cCodingMap[0xD6] = 0xE1;
	cCodingMap[0xA8] = 0xA2; cCodingMap[0xE1] = 0xB2; cCodingMap[0xFA] = 0xC2; cCodingMap[0xD9] = 0xE2;
	cCodingMap[0xC6] = 0xA3; cCodingMap[0xE2] = 0xB3; cCodingMap[0xFB] = 0xC3; cCodingMap[0xE4] = 0xE3;
	cCodingMap[0xC7] = 0xA4; cCodingMap[0xE3] = 0xB4; cCodingMap[0xFC] = 0xC4; cCodingMap[0xF4] = 0xE4;
	cCodingMap[0xC8] = 0xA5; cCodingMap[0xB8] = 0xB5; cCodingMap[0xFD] = 0xC5; cCodingMap[0xF6] = 0xE5;
	cCodingMap[0xC9] = 0xA6; cCodingMap[0xE6] = 0xB6; cCodingMap[0xFE] = 0xC6; cCodingMap[0xF9] = 0xE6;
	cCodingMap[0xCB] = 0xA7; cCodingMap[0xE7] = 0xB7; cCodingMap[0xFF] = 0xC7; cCodingMap['\''] = 0xE7;
	cCodingMap[0xCF] = 0xA8; cCodingMap[0xE8] = 0xB8;
	cCodingMap[0xD3] = 0xA9; cCodingMap[0xE9] = 0xB9;
	cCodingMap[0xD4] = 0xAA; cCodingMap[0xEA] = 0xBA;
	cCodingMap[0xD7] = 0xAB; cCodingMap[0xEB] = 0xBB;
	cCodingMap[0xD8] = 0xAC; cCodingMap[0xEC] = 0xBC; cCodingMap['№'] = 0xCC;
	cCodingMap[0xDA] = 0xAD; cCodingMap[0xED] = 0xBD;
	cCodingMap[0xDB] = 0xAE; cCodingMap[0xEF] = 0xBE;
	cCodingMap[0xDD] = 0xAF; cCodingMap[0xF2] = 0xBF;

	cCodingMap[0xC0] = 0x41;
	cCodingMap[0xC2] = 0x42;
	cCodingMap[0xC5] = 0x45;
	cCodingMap[0xCA] = 0x4B;
	cCodingMap[0xCC] = 0x4D;
	cCodingMap[0xCD] = 0x48;
	cCodingMap[0xCE] = 0x4F;
	cCodingMap[0xD0] = 0x50;
	cCodingMap[0xD1] = 0x43;
	cCodingMap[0xD2] = 0x54;
	cCodingMap[0xD5] = 0x58;
	cCodingMap[0xDC] = 0x62;
	cCodingMap[0xE0] = 0x61;
	cCodingMap[0xE5] = 0x65;
	cCodingMap[0xEE] = 0x6F;
	cCodingMap[0xF0] = 0x70;
	cCodingMap[0xF1] = 0x63;
	cCodingMap[0xF3] = 0x79;
	cCodingMap[0xF5] = 0x78;
	//АВЕКМНОРСТХЬ
	//аеорсух
	if (pReader)
	{
		for (int i = 0; i < ucMessageLength; i++)
			szCodedMessage[i] = cCodingMap[(BYTE)lpMessage[i]];

		BYTE ucMode = 0x00;
		ucMode += ((dwInfiniteLoop > 0) ? 0x80 : 0x00);
		ucMode += ((dwLineIndex > 0) ? 0x40 : 0x00);
		ucMode += ((dwMode << 4) & 0x30);
		ucMode += (ucRepeats & 0x0F);

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x45);
		pTr->m_TrBuf.AddByte(0x01);
		pTr->m_TrBuf.AddByte(ucMode);
		pTr->m_TrBuf.AddByte(ucMessageLength);
		pTr->m_TrBuf.AddBulk((PUCHAR)szCodedMessage, ucMessageLength);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_DirectIO_Reader(IN     LPVOID  pReader,
IN     LPCBYTE pbSendBuffer,
IN     DWORD   dwSendLength,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;

	if (dwSendLength == 0)
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_INVALID_PARAMETER;
	}
	else if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddBulk(pbSendBuffer, dwSendLength);

		Status = pTr->SingleTransceive();

		//if( Status == SCARD_S_SUCCESS )
		//{
		DWORD Size = pTr->m_RcBuf.GetLength();
		if (Size < *pdwRecvLength)
			*pdwRecvLength = Size;

		if (!pTr->m_RcBuf.GetBulk(0, pbRecvBuffer, pdwRecvLength))
			Status = SCARD_E_COMM_DATA_LOST;
		//}
		//else
		//    *pdwRecvLength = 0;
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_DirectIO_Card(IN     LPVOID  pReader,
IN     LPCBYTE pbSendBuffer,
IN     DWORD   dwSendLength,
IN     DWORD   dwTimeout,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (dwSendLength == 0)
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_INVALID_PARAMETER;
	}
	else if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x48);
		pTr->m_TrBuf.AddBulk((LPBYTE)&dwSendLength, 2);
		pTr->m_TrBuf.AddBulk(pbSendBuffer, dwSendLength);
		pTr->m_TrBuf.AddBulk((LPBYTE)&dwTimeout, 4);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}
