#include "StdAfx.h"

#include "Clscrfl.h"

#include "Clscrfl_SAM.h"

#include "SAM_Set.h"

#include <cstdio>

extern std::map<PVOID, READER_SETTINGS> m_ReaderSettings;

//std::map<PVOID, DF_SETTINGS_GLOBAL> m_DF_Settings;

//std::map<PVOID, CRYPTO_SETTINGS_GLOBAL> m_CryptoSettings;

/*
struct SAMReaderParams
{
BYTE ucLogicalChannel;
BYTE ucSW1;
BYTE ucSW2;
};

std::map<LPVOID,SAMReaderParams> SAM_pmLogicalChannels;// = new std::map<LPVOID,SAMReaderParams>();
*/

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_GetStatusDescription(IN BYTE ucSW1,
IN BYTE ucSW2,
OUT LPSTR pError)
{
	//strcpy(pError, "Unknown error");
	sprintf(pError, "Command specific error (SW1=0x%02X SW2=0x%02X)", ucSW1, ucSW2);

	if ((ucSW1 == 0x64) && (ucSW2 == 0x00))
		strcpy(pError, "EEPROM busy, access collision, memory range cannot be mapped, execution error");
	if ((ucSW1 == 0x64) && (ucSW2 == 0x01))
		strcpy(pError, "RC222 interface error");
	if ((ucSW1 == 0x65) && (ucSW2 == 0x02))
		strcpy(pError, "SAM key reference number invalid");
	if ((ucSW1 == 0x65) && (ucSW2 == 0x03))
		strcpy(pError, "SAM key usage counter number invalid");
	if ((ucSW1 == 0x65) && (ucSW2 == 0x81))
		strcpy(pError, "EEPROM high voltage failure; SAM permanently disabled");
	if ((ucSW1 == 0x67) && (ucSW2 == 0x00))
		strcpy(pError, "Wrong length of the APDU or wrong Lc byte");
	if ((ucSW1 == 0x68) && (ucSW2 == 0x83))
		strcpy(pError, "final chained command expected; running command aborted, new command ignored");
	if ((ucSW1 == 0x69) && (ucSW2 == 0x84))
		strcpy(pError, "CRC mismatch; wrong key handling buffer size");
	if ((ucSW1 == 0x6A) && (ucSW2 == 0x86))
		strcpy(pError, "Incorrect value for parameter P1 and/or P2");
	if ((ucSW1 == 0x6C) && (ucSW2 == 0x00))
		strcpy(pError, "Incorrect Le value");
	if ((ucSW1 == 0x6D) && (ucSW2 == 0x00))
		strcpy(pError, "Instruction code not supported or invalid or not available in the current state of the SAM");
	if ((ucSW1 == 0x6E) && (ucSW2 == 0x00))
		strcpy(pError, "Class not supported");
	if ((ucSW1 == 0x6F) && (ucSW2 == 0x00))
		strcpy(pError, "No precize diagnosis (physical address invalid, EEPROM error, no correct key loaded for cryptographic operation, improper length for cryptographic operation, internal limit exceeded...)");
	if ((ucSW1 == 0x90) && (ucSW2 == 0x00))
		strcpy(pError, "OK");

	if ((ucSW1 == 0x90) && (ucSW2 == 0xBE))
		strcpy(pError, "MIFARE PICC sent incorrect amount of data, wrong length of CMAC");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xBF))
		strcpy(pError, "MIFARE PICC returned error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE0))
		strcpy(pError, "no response from card (no card in field)");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE1))
		strcpy(pError, "ISO 14442 protocol error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE2))
		strcpy(pError, "Parity error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE3))
		strcpy(pError, "Buffer overflow");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE4))
		strcpy(pError, "(N)ACK expected, CRC mismatch");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE5))
		strcpy(pError, "RF field inactive");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE6))
		strcpy(pError, "Temperature error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE7))
		strcpy(pError, "FIFO write error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE8))
		strcpy(pError, "Collision error");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xE9))
		strcpy(pError, "internal transaction buffer overflow (more than 256 bytes received from the card)");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xEA))
		strcpy(pError, "UID check byte (BCC) incorrect");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xEB))
		strcpy(pError, "incorrect block number");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xEC))
		strcpy(pError, "ATS length byte does not match actual length of received ATS or PPS response invalid (incorrect length or PPSS byte); block length or PCB invalid; incorrect (no R-block or I-block) answer received");
	if ((ucSW1 == 0x90) && (ucSW2 == 0xED))
		strcpy(pError, "User buffer length below ISO14443-4 minimum; invalid UID bit count (too many bits received); CID, NAD either not expected, missing or incorrect");
	if ((ucSW1 == 0x90) && ((ucSW2 & 0xF0) == 0xF0))
		strcpy(pError, "MIFARE PICC sent NACK");

	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SetCurrentHolder(IN LPVOID pReader,
IN BYTE ucNumSAM)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xB2);
		pTr->m_TrBuf.AddByte(ucNumSAM);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SetCurrentHolderGetATR(IN  LPVOID pReader,
IN  BYTE ucNumSAM,
IN  DWORD dwATRBufSize,
OUT PBYTE  pATR,
OUT PDWORD pdwATRSize
)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xB2);
		pTr->m_TrBuf.AddByte(0x10 | ucNumSAM);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwATRSize = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwATRSize = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwATRSize, &Size))
			{
				*pdwATRSize = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwATRSize > dwATRBufSize)
			{
				*pdwATRSize = dwATRBufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwATRSize) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pATR, pdwATRSize))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}

	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SetCurrentHolderEx(IN  LPVOID pReader,
IN  BYTE ucNumHolder,
IN  BYTE ucWT10,
IN  BYTE ucPPS1,
IN  BYTE ucDeactivate,
IN  DWORD dwATRBufSize,
OUT PBYTE  pATR,
OUT PDWORD pdwATRSize
)
{
	LONG Status = SCARD_S_SUCCESS;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	ucNumHolder &= 0x0F;

	if (ucNumHolder)
	{
		if (ucWT10 != 0xFF)
			ucNumHolder |= 0x80;

		if (ucPPS1 != 0xFF)
			ucNumHolder |= 0x40;

		if (ucDeactivate)
			ucNumHolder |= 0x20;

		if (dwATRBufSize > 0)
		{
			if (!pATR || !pdwATRSize)
				return SCARD_E_INVALID_PARAMETER;

			ucNumHolder |= 0x10;
		}
	}

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xB2);
	pTr->m_TrBuf.AddByte(ucNumHolder);
	if (ucNumHolder)
	{
		if (ucWT10 != 0xFF)
			pTr->m_TrBuf.AddByte(ucWT10);
		if (ucPPS1 != 0xFF)
			pTr->m_TrBuf.AddByte(ucPPS1);
	}

	Status = pTr->SingleTransceive();

	DWORD Size = pTr->m_RcBuf.GetLength();

	if (ucNumHolder && (dwATRBufSize > 0))
	{
		if (Size < 2)  // Length is absent
		{
			*pdwATRSize = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwATRSize = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwATRSize, &Size))
			{
				*pdwATRSize = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwATRSize > dwATRBufSize)
			{
				*pdwATRSize = dwATRBufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwATRSize) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pATR, pdwATRSize))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_GetCurrentHolder(IN LPVOID pReader,
OUT LPBYTE pucNumSAM)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xB4);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			if (!pTr->m_RcBuf.GetBulk(1, pucNumSAM, &dwLen))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_APDU(IN     LPVOID  pReader,
IN     LPCBYTE pbSendBuffer,
IN     DWORD   dwSendLength,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (dwSendLength == 0)
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_INVALID_PARAMETER;
	}
	else if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0xb6);
		pTr->m_TrBuf.AddBulk((LPBYTE)&dwSendLength, 2);
		pTr->m_TrBuf.AddBulk(pbSendBuffer, dwSendLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}
/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SetLogicalChannel(	IN LPVOID pReader,
IN BYTE ucLogicalChannel )
{
LONG Status = SCARD_S_SUCCESS;

if( pReader )
{
if (ucLogicalChannel <= 3)
{
if( SAM_pmLogicalChannels.find(pReader) == SAM_pmLogicalChannels.end() )
{
SAM_pmLogicalChannels[pReader].ucSW1 = 0;
SAM_pmLogicalChannels[pReader].ucSW2 = 0;
}
SAM_pmLogicalChannels[pReader].ucLogicalChannel = ucLogicalChannel;
}
else
Status = SCARD_E_INVALID_PARAMETER;
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}

BYTE SAM_GetLogicalChannel(	IN LPVOID pReader )
{
if( pReader )
if( SAM_pmLogicalChannels.find(pReader) != SAM_pmLogicalChannels.end() )
return SAM_pmLogicalChannels[pReader].ucLogicalChannel;

return 0;
}

void SAM_SetLastError(	IN LPVOID pReader,
IN BYTE ucSW1,
IN BYTE ucSW2 )
{
if( pReader )
{
if( SAM_pmLogicalChannels.find(pReader) == SAM_pmLogicalChannels.end() )
SAM_pmLogicalChannels[pReader].ucLogicalChannel = 0;
SAM_pmLogicalChannels[pReader].ucSW1 = ucSW1;
SAM_pmLogicalChannels[pReader].ucSW2 = ucSW2;
}
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_GetLastError(	IN LPVOID pReader,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2 )
{
LONG Status = SCARD_S_SUCCESS;

if ((pucSW1 == NULL) || (pucSW2 == NULL))
return SCARD_E_INVALID_PARAMETER;

if( pReader )
{
if( SAM_pmLogicalChannels.find(pReader) != SAM_pmLogicalChannels.end() )
{
*pucSW1 = SAM_pmLogicalChannels[pReader].ucSW1;
*pucSW2 = SAM_pmLogicalChannels[pReader].ucSW2;
}
else
{
SAM_SetLastError(pReader, 0, 0);
*pucSW1 = 0;
*pucSW1 = 0;
}
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}
*/

/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_AuthenticateHost(IN LPVOID pReader,
IN BYTE ucLC,
IN SAM_AUTH_MODE AuthMode,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN BYTE ucDivLength,
IN LPBYTE pbDivInp,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2 )
{
LONG Status = SCARD_S_SUCCESS;
DWORD dwLen = 1;

if( pReader )
{
CTransfer* pTr = (CTransfer*)pReader;
pTr->m_TrBuf.Clear();
pTr->m_TrBuf.AddByte( 0xB8 );
pTr->m_TrBuf.AddByte( ucLC );
pTr->m_TrBuf.AddByte( AuthMode.ucBitMask );
pTr->m_TrBuf.AddByte( ucKeyNo );
pTr->m_TrBuf.AddByte( ucKeyVer );
pTr->m_TrBuf.AddByte( ucDivLength );
pTr->m_TrBuf.AddBulk( pbDivInp, ucDivLength );

Status = pTr->SingleTransceive();

if( Status == SCARD_S_SUCCESS )
{
if(
( ! pTr->m_RcBuf.GetBulk( 1, pucSW1, &dwLen ) )
|| ( ! pTr->m_RcBuf.GetBulk( 2, pucSW2, &dwLen ) )
)
Status = SCARD_E_COMM_DATA_LOST;
}
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}
*/


CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_GetVersion(IN LPVOID pReader,
IN BYTE ucLC,
OUT CLSCRF_SAM_VERSION* pVersion,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		BYTE SendBuffer[256] = { 0 };
		BYTE ReceiveBuffer[256] = { 0 };

		int iPos = 0;

		SendBuffer[iPos++] = 0x80 | ucLC;
		SendBuffer[iPos++] = 0x60;
		SendBuffer[iPos++] = 0x00;
		SendBuffer[iPos++] = 0x00;
		SendBuffer[iPos++] = 0x00;
		DWORD dwReceiveLength = sizeof(ReceiveBuffer);//12+(((ucKeyKind == 3) || (ucKeyKind == 5))?0:1);

		Status = CLSCRF_SAM_APDU(
			pReader,
			SendBuffer,
			iPos,
			ReceiveBuffer,
			&dwReceiveLength
			);

		iPos = 0;
		if (Status == SCARD_S_SUCCESS)
		{
			if (dwReceiveLength - 1 >= sizeof(CLSCRF_SAM_VERSION))
				memcpy(pVersion, ReceiveBuffer, sizeof(CLSCRF_SAM_VERSION));
			else
				Status = SCARD_F_INTERNAL_ERROR;
			//	MessageBox(NULL,"Wrong response!", "", MB_OK);
			*pucSW1 = ReceiveBuffer[dwReceiveLength - 2];
			*pucSW2 = ReceiveBuffer[dwReceiveLength - 1];
		}
		//else if (( Status == SCARD_F_INTERNAL_ERROR )&&(GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		//{
		//	*pucSW1 = ReceiveBuffer[dwReceiveLength-2];
		//	*pucSW2 = ReceiveBuffer[dwReceiveLength-1];
		//      }
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

// ucDivLength[1] - äëèíà âõîäíûõ äàííûõ (8-áàéòîâûé êëþ÷ DES, 
//                  16-áàéòîâûé êëþ÷ AES èëè 1-31-áàéòîâûé âåêòîð äèâåðñèôèêàöèè);
// pbDivInp[1-31] - âõîäíûå äàííûå (êëþ÷è DES, AES èëè âåêòîð äèâåðñèôèêàöèè);

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_AuthenticateHost(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucHostMode,
IN BYTE ucKeyKind,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN PBYTE pKey,
//IN BYTE ucDivLength, 
//IN LPBYTE pbDivInp, 
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	BYTE SW[2] = { 0 };

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;

		BYTE bRes = SAM_AuthenticateHost(
			pReader,
			ucLC,
			ucKeyKind,
			pKey,
			ucKeyNo,
			ucKeyVer,
			ucHostMode,
			SW);

		if (bRes != 0)
		{
			Status = SCARD_F_INTERNAL_ERROR;
			pTr->m_RcBuf.Clear();
			pTr->m_RcBuf.AddByte(bRes);
		}

		*pucSW1 = SW[0];
		*pucSW2 = SW[1];

	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_ActivateOfflineKey(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN BYTE ucDivLength,
IN LPBYTE pbDivInp,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	//BYTE SW[2] = {0};

	if (pReader)
	{
		BYTE SendBuffer[256] = { 0 };
		BYTE ReceiveBuffer[256] = { 0 };

		int iPos = 0;

		SendBuffer[iPos++] = 0x80 | ucLC;
		SendBuffer[iPos++] = 0x01;
		SendBuffer[iPos++] = 0x00;
		SendBuffer[iPos++] = (pbDivInp == NULL) ? 0x00 : 0x01;
		SendBuffer[iPos++] = 0x02 + (pbDivInp == NULL) ? 0 : ucDivLength;
		SendBuffer[iPos++] = ucKeyNo;
		SendBuffer[iPos++] = ucKeyVer;
		if (pbDivInp != NULL)
		{
			memcpy(&SendBuffer[iPos], pbDivInp, ucDivLength); iPos += ucDivLength;
		}

		DWORD dwReceiveLength = sizeof(ReceiveBuffer);//12+(((ucKeyKind == 3) || (ucKeyKind == 5))?0:1);

		Status = CLSCRF_SAM_APDU(
			pReader,
			SendBuffer,
			iPos,
			ReceiveBuffer,
			&dwReceiveLength
			);

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			*pucSW1 = ReceiveBuffer[0];
			*pucSW2 = ReceiveBuffer[1];
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_Lock(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyKind,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN PBYTE pUnlockKey,
IN BYTE ucUnlockKeyNo,
IN BYTE ucUnlockKeyVer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	BYTE SW[2] = { 0 };

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;

		BYTE bRes = SAM_LockUnlock(
			pReader,
			ucLC,
			(pUnlockKey == NULL) ? 0x01 : 0x02,
			ucKeyNo,
			ucKeyVer,
			pUnlockKey,
			ucUnlockKeyNo,
			ucUnlockKeyVer,
			0,
			SW);

		if (bRes != 0)
		{
			Status = SCARD_F_INTERNAL_ERROR;
			pTr->m_RcBuf.Clear();
			pTr->m_RcBuf.AddByte(bRes);
		}

		*pucSW1 = SW[0];
		*pucSW2 = SW[1];
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_Unlock(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyKind,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN PBYTE pUnlockKey,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	BYTE SW[2] = { 0 };

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;

		BYTE bRes = SAM_LockUnlock(
			pReader,
			ucLC,
			0x00,
			ucKeyNo,
			ucKeyVer,
			pUnlockKey,
			0,
			0,
			0,
			SW);

		if (bRes != 0)
		{
			Status = SCARD_F_INTERNAL_ERROR;
			pTr->m_RcBuf.Clear();
			pTr->m_RcBuf.AddByte(bRes);
		}

		*pucSW1 = SW[0];
		*pucSW2 = SW[1];
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SwitchToAV2Mode(	IN LPVOID pReader,
IN BYTE ucLC,
IN LPBYTE pbNewPosAkey,
IN BYTE ucNewPosAkeyVer,
IN LPBYTE pbNewPosBkey,
IN BYTE ucNewPosBkeyVer,
IN LPBYTE pbNewPosCkey,
IN BYTE ucNewPosCkeyVer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2 )
{
LONG Status = SCARD_S_SUCCESS;
DWORD dwLen = 1;

if( pReader )
{
CTransfer* pTr = (CTransfer*)pReader;
pTr->m_TrBuf.Clear();
pTr->m_TrBuf.AddByte( 0xB7 );
pTr->m_TrBuf.AddByte( ucLC );
pTr->m_TrBuf.AddBulk( pbNewPosAkey, 16 );
pTr->m_TrBuf.AddByte( ucNewPosAkeyVer );
pTr->m_TrBuf.AddBulk( pbNewPosBkey, 16 );
pTr->m_TrBuf.AddByte( ucNewPosBkeyVer );
pTr->m_TrBuf.AddBulk( pbNewPosCkey, 16 );
pTr->m_TrBuf.AddByte( ucNewPosCkeyVer );

Status = pTr->SingleTransceive();

if( Status == SCARD_S_SUCCESS )
{
if(
( ! pTr->m_RcBuf.GetBulk( 1, pucSW1, &dwLen ) )
|| ( ! pTr->m_RcBuf.GetBulk( 2, pucSW2, &dwLen ) )
)
Status = SCARD_E_COMM_DATA_LOST;
}
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}
*/
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_SwitchToAV2Mode(IN LPVOID pReader,
IN BYTE ucLC,
IN LPBYTE pbNewPosAkey,
IN BYTE ucNewPosAkeyVer,
IN LPBYTE pbNewPosBkey,
IN BYTE ucNewPosBkeyVer,
IN LPBYTE pbNewPosCkey,
IN BYTE ucNewPosCkeyVer,
IN DWORD dwMaxChainBlocks,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	BYTE SW[2] = { 0 };

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		BYTE bRes = SAM_SwitchToAV2Mode(
			pReader,
			ucLC,
			pbNewPosAkey,
			ucNewPosAkeyVer,
			pbNewPosBkey,
			ucNewPosBkeyVer,
			pbNewPosCkey,
			ucNewPosCkeyVer,
			dwMaxChainBlocks,
			SW);

		if (bRes != 0)
		{
			Status = SCARD_F_INTERNAL_ERROR;
			pTr->m_RcBuf.Clear();
			pTr->m_RcBuf.AddByte(bRes);
		}

		*pucSW1 = SW[0];
		*pucSW2 = SW[1];
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_Reset(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;
	//DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xba);
		pTr->m_TrBuf.AddByte(0x21);
		pTr->m_TrBuf.AddByte(0x43);
		pTr->m_TrBuf.AddByte(0x65);
		pTr->m_TrBuf.AddByte(0x87);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_KillAuth(IN LPVOID pReader, 
IN BYTE ucType)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xba);
		pTr->m_TrBuf.AddByte(ucType);
		pTr->m_TrBuf.AddByte(0x43);
		pTr->m_TrBuf.AddByte(0x65);
		pTr->m_TrBuf.AddByte(0x87);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_Halt_A(IN LPVOID pReader,
IN BYTE ucLC)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x58);
		pTr->m_TrBuf.AddByte(0x50);
		pTr->m_TrBuf.AddByte(ucLC);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

extern 	BYTE SessionKeyEnc[16];
extern	BYTE SessionKeyMAC[16];
extern	BYTE KeyX[24];

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_ChangeKeyEntry(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyNo,
IN BYTE ucKeyKind,
IN LPBYTE pbKeyA,
IN LPBYTE pbKeyB,
IN LPBYTE pbKeyC,
IN DWORD dwDF_AID,
IN BYTE ucDF_KeyNr,
IN BYTE ucCEKNo,
IN BYTE ucCEKV,
IN BYTE ucKUC,
IN CLSCRF_SAM_SETTINGS* pSettings,
IN BYTE ucKeyVerA,
IN BYTE ucKeyVerB,
IN BYTE ucKeyVerC,
IN CLSCRF_SAM_EXT_SETTINGS* pExtSettings,
IN BOOL fPlainKeyUpdate,
IN BOOL fOfflineKeyChange,
IN LPBYTE pbSAM_UID,
IN CLSCRF_SAM_NV_PROGRAMMING_MASK* pNVProgrammingMask,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	//BYTE pMyKeyA[16] = {0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0x00, 0x00, 0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0x00, 0x00};
	//BYTE pMyKeyB[16] = {0xF0, 0xE1, 0xD2, 0xC3, 0xB4, 0xA5, 0x00, 0x00, 0x89, 0x7A, 0x6B, 0x5C, 0x4D, 0x3E, 0x00, 0x00};
	//BYTE pMyKeyC[16] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00};
	//BYTE pMyKeyA[16] = {0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16};
	//BYTE pMyKeyB[16] = {0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16};
	//BYTE pMyKeyC[16] = {0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16};
	//BYTE pMyKeyB[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF};
	//BYTE pMyKeyC[16] = {0xAB, 0xCD, 0xEF, 0x01, 0x23, 0x45, 0x67, 0x89, 0x90, 0x81, 0x72, 0x63, 0x54, 0x5E, 0x74, 0x0F};
	//SAM_SETTINGS MySET = {0};
	//MySET.wSettings = 0x0010;
	//MySET.wSettings = 0x0020;
	//MySET.wSettings = 0x0120;
	//MySET.wSettings = 0x0024;
	//BYTE ucMyKeyVerA = 0x00;
	//BYTE ucMyKeyVerB = 0x01;
	//BYTE ucMyKeyVerC = 0x02;
	//BYTE ucMyKeyVerA = 0x01;
	//BYTE ucMyKeyVerB = 0x02;
	//BYTE ucMyKeyVerC = 0x03;
	//BYTE ucMyCEKNo = 0x00;
	//BYTE ucMyCEKV = 0x01;
	//BYTE ucMyCEKV = 0x00;
	//BYTE ucMyKUC = 0x02;
	//BYTE ucMyKUC = 0xFF;
	//SAM_NV_PROGRAMMING_MASK MyNVProgrammingMask = {0};
	//MyNVProgrammingMask.ucNVProgramingMask = 0xFF;
	//BYTE ucMyKeyNo = 0x01;
	//BYTE ucMyKeyNo = 0x02;
	//BYTE ucMyKeyNo = 0x17;
	//BYTE ucMyKeyNo = 0x07;
	//SAM_EXT_SETTINGS MyExtSettings = {0};
	//MyExtSettings.ucExtSettings = 0x01;
	//MyExtSettings.ucExtSettings = 0x00;
	//MyExtSettings.ucExtSettings = 0x11;

	if (pReader)
	{
		BYTE SendBuffer[256] = { 0 };
		BYTE ReceiveBuffer[256] = { 0 };

		BYTE KeyEntry[128] = { 0 };
		int iPos = 0;

		int iLenEKeyEntry = 0;

		DWORD dwCounter = 0;
		BYTE MAC_Data[256] = { 0 };

		memcpy(KeyEntry + iPos, pbKeyA, 16); iPos += 16;
		memcpy(KeyEntry + iPos, pbKeyB, 16); iPos += 16;
		memcpy(KeyEntry + iPos, pbKeyC, 16); iPos += 16;
		//memcpy(KeyEntry + iPos, pMyKeyA, 16); iPos += 16;
		//memcpy(KeyEntry + iPos, pMyKeyB, 16); iPos += 16;
		//memcpy(KeyEntry + iPos, pMyKeyC, 16); iPos += 16;
		memcpy(KeyEntry + iPos, &dwDF_AID, 3); iPos += 3;
		KeyEntry[iPos++] = ucDF_KeyNr;
		KeyEntry[iPos++] = ucCEKNo;
		//KeyEntry[iPos++] = ucMyCEKNo;
		KeyEntry[iPos++] = ucCEKV;
		//KeyEntry[iPos++] = ucMyCEKV;
		KeyEntry[iPos++] = ucKUC;
		//KeyEntry[iPos++] = ucMyKUC;
		memcpy(KeyEntry + iPos, &(pSettings->wSettings), 2); iPos += 2;
		//memcpy(KeyEntry + iPos, &MySET, 2); iPos += 2;
		KeyEntry[iPos++] = ucKeyVerA;
		KeyEntry[iPos++] = ucKeyVerB;
		KeyEntry[iPos++] = ucKeyVerC;
		//KeyEntry[iPos++] = ucMyKeyVerA;
		//KeyEntry[iPos++] = ucMyKeyVerB;
		//KeyEntry[iPos++] = ucMyKeyVerC;
		memcpy(KeyEntry + iPos, &(pExtSettings->ucExtSettings), 1); iPos += 1;
		//memcpy(KeyEntry + iPos, &MyExtSettings, 1); iPos += 1;
		if (fPlainKeyUpdate == TRUE)
		{
			//memcpy(KeyEntry + iPos, pbSAM_UID, 7);
			//iPos += 7;

			SendBuffer[0] = 0x80 | ucLC;
			SendBuffer[1] = 0xC1;
			//SendBuffer[2] = ucMyKeyNo;
			SendBuffer[2] = ucKeyNo;
			//SendBuffer[3] = MyNVProgrammingMask.ucNVProgramingMask;
			SendBuffer[3] = pNVProgrammingMask->ucNVProgramingMask;
			SendBuffer[4] = (BYTE)iPos;
			memcpy(&SendBuffer[5], KeyEntry, iPos);
			iPos += 5;
		}
		else
		{
			//BYTE MyKc[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF};
			//BYTE ucMyKeyKind = 4;

			BYTE Kc[24] = { 0 };
			BYTE Ke[16] = { 0 };
			BYTE Km[16] = { 0 };
			//BYTE MyKe[16] = {0x09, 0x2D, 0x5F, 0x2A, 0xA7, 0x8F, 0x5A, 0x22, 0xB5, 0xF5, 0xA0, 0x1F, 0x93, 0x1A, 0x83, 0xFB};
			//BYTE MyKm[16] = {0x2C, 0xA7, 0xAD, 0xBD, 0x49, 0x69, 0xDD, 0x3F, 0x22, 0xBE, 0xC6, 0xB5, 0xC3, 0x99, 0x52, 0xCA};
			//BYTE MyKe[16] = {0x9A, 0xF4, 0x2A, 0x7D, 0xAC, 0xC8, 0x15, 0x5B, 0x77, 0xC4, 0x2A, 0xFE, 0xB2, 0x4D, 0x7A, 0x8A};
			//BYTE MyKm[16] = {0xD9, 0x9B, 0x93, 0x3C, 0xD5, 0x87, 0xC2, 0xFC, 0x01, 0x26, 0x60, 0x49, 0x27, 0xF9, 0x20, 0xFC};
			BYTE Kce[24] = { 0 };
			BYTE Kcm[24] = { 0 };
			BYTE IVLoad[16] = { 0x01, 0x01, 0x01, 0x01, 0x00 };
			BYTE IV[24] = { 0 };
			BYTE Output[256] = { 0 };
			//BYTE DataField[256] = {0};
			WORD wChange_Ctr = 100;

			//AES-128
			//if (ucKeyKind == 4)
			//if (ucMyKeyKind == 4)
			//{
			BYTE SV1[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71 };
			BYTE SV2[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72 };

			memcpy(Kc, KeyX, 16);

			aes_128_encrypt_cbc(SV1, 16, Kc, IV, Kce);
			aes_128_encrypt_cbc(SV2, 16, Kc, IV, Kcm);

			//memcpy(Kce, SessionKeyEnc, 16);
			//memcpy(Kcm, SessionKeyMAC, 16);

			//aes_128_encrypt_cbc( SV1, 16, MyKc, IV, Kce );
			//aes_128_encrypt_cbc( SV2, 16, MyKc, IV, Kcm );
			//}
			/*			//AES-192
			if (ucKeyKind == 5)
			{
			BYTE SV1a[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73};
			BYTE SV1b[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74};
			BYTE SV2a[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75};
			BYTE SV2b[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76};

			BYTE Kcea[24] = {0};
			aes_192_encrypt_cbc( SV1a, 24, Kc, IV, Kcea );

			BYTE Kceb[24] = {0};
			aes_192_encrypt_cbc( SV1b, 24, Kc, IV, Kceb );

			memcpy(Kce, &Kcea[8], 8);
			memcpy(&Kce[16], Kceb, 8);
			xor_size(Kcea, &Kceb[8], &Kce[8], 8);

			BYTE Kcma[24] = {0};
			aes_192_encrypt_cbc( SV1a, 24, Kc, IV, Kcma );

			BYTE Kcmb[24] = {0};
			aes_192_encrypt_cbc( SV1b, 24, Kc, IV, Kcmb );

			memcpy(Kcm, &Kcma[8], 8);
			memcpy(&Kcm[16], Kcmb, 8);
			xor_size(Kcma, &Kcmb[8], &Kcm[8], 8);
			}
			*/
			if (fOfflineKeyChange == TRUE)
			{
				KeyEntry[iPos++] = 0x80;

				iLenEKeyEntry = 16 * (int)ceil((float)iPos / 16);

				aes_128_encrypt_cbc(
					KeyEntry,
					iLenEKeyEntry,
					Kce,
					IV,
					Output);

				iPos = 0;
				SendBuffer[iPos++] = 0x80 | ucLC;
				SendBuffer[iPos++] = 0xC1;
				//SendBuffer[iPos++] = ucMyKeyNo;
				SendBuffer[iPos++] = ucKeyNo;
				//SendBuffer[iPos++] = MyNVProgrammingMask.ucNVProgramingMask;
				SendBuffer[iPos++] = pNVProgrammingMask->ucNVProgramingMask;
				SendBuffer[iPos++] = (BYTE)(2 + iLenEKeyEntry + 8);
				memcpy(SendBuffer + iPos, &wChange_Ctr, 2); iPos += 2;
				memcpy(SendBuffer + iPos, Output, iLenEKeyEntry); iPos += iLenEKeyEntry;

				memcpy(m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey, Kcm, 16);
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = CLSCRF_DF_AUTH_TYPE_AES;
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize = 16;

				DFSubkeys(pReader);

				DFEncrypt(
					pReader,
					SendBuffer,
					&SendBuffer[iPos],
					iPos,
					NULL,
					true,
					true);

				for (int i = 0; i < 8; i++)
					SendBuffer[iPos + i] = SendBuffer[iPos + i * 2 + 1];
				iPos += 8;

				/*
				MAC_Data[iPos++] = 0x80;

				iLenEKeyEntry = 16*ceil((float)iPos / 16);

				aes_128_encrypt_cbc(
				MAC_Data,
				iLenEKeyEntry,
				Kcm,
				IV,
				Output,
				IV);
				*/
			}
			else
			{
				//	memcpy(Ke, Kce, 16);
				//	memcpy(Km, Kcm, 16);

				memcpy(Ke, SessionKeyEnc, 16);
				memcpy(Km, SessionKeyMAC, 16);

				/*
				SendBuffer[0] = 0x80;
				SendBuffer[1] = 0xC1;
				SendBuffer[2] = ucMyKeyNo;
				SendBuffer[3] = MyNVProgrammingMask.ucNVProgramingMask;
				SendBuffer[4] = iPos;
				memcpy(&SendBuffer[5], KeyEntry, iPos);
				iPos += 5;
				*/

				memcpy(IVLoad + 4, &dwCounter, 4);
				memcpy(IVLoad + 8, &dwCounter, 4);
				memcpy(IVLoad + 12, &dwCounter, 4);

				aes_128_encrypt_cbc(
					IVLoad,
					16,
					//MyKe,
					Ke,
					IV,
					IV);

				KeyEntry[iPos++] = 0x80;

				iLenEKeyEntry = 16 * (int)ceil((float)iPos / 16);

				aes_128_encrypt_cbc(
					KeyEntry,
					iLenEKeyEntry,
					//MyKe,
					Ke,
					IV,
					Output);

				iPos = 0;
				MAC_Data[iPos++] = 0x80 | ucLC;
				MAC_Data[iPos++] = 0xC1;
				memcpy(MAC_Data + iPos, &dwCounter, 4); iPos += 4;
				//MAC_Data[iPos++] = ucMyKeyNo;
				MAC_Data[iPos++] = ucKeyNo;
				//MAC_Data[iPos++] = MyNVProgrammingMask.ucNVProgramingMask;
				MAC_Data[iPos++] = pNVProgrammingMask->ucNVProgramingMask;
				MAC_Data[iPos++] = (BYTE)(2 + iLenEKeyEntry + 8);
				memcpy(MAC_Data + iPos, &wChange_Ctr, 2); iPos += 2;
				memcpy(MAC_Data + iPos, Output, iLenEKeyEntry); iPos += iLenEKeyEntry;

				//memcpy(m_DF_Settings[pReader].SessionKey, MyKm, 16);
				memcpy(m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey, Km, 16);
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = CLSCRF_DF_AUTH_TYPE_AES;
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize = 16;

				DFSubkeys(pReader);

				DFEncrypt(
					pReader,
					MAC_Data,
					MAC_Data,
					iPos,
					NULL,
					true,
					true);

				for (int i = 0; i < 8; i++)
					MAC_Data[i] = MAC_Data[i * 2 + 1];

				iPos = 0;
				SendBuffer[iPos++] = 0x80 | ucLC;
				SendBuffer[iPos++] = 0xC1;
				//SendBuffer[iPos++] = ucMyKeyNo;
				SendBuffer[iPos++] = ucKeyNo;
				//SendBuffer[iPos++] = MyNVProgrammingMask.ucNVProgramingMask;
				SendBuffer[iPos++] = pNVProgrammingMask->ucNVProgramingMask;
				SendBuffer[iPos++] = (BYTE)(2 + iLenEKeyEntry + 8);
				memcpy(SendBuffer + iPos, &wChange_Ctr, 2); iPos += 2;
				memcpy(SendBuffer + iPos, Output, iLenEKeyEntry); iPos += iLenEKeyEntry;
				memcpy(SendBuffer + iPos, MAC_Data, 8); iPos += 8;

				//memcpy(DataField, &wChange_Ctr, 2);

				/*
				ULONG crc32 = 0;
				CRC32 ccc32;
				ccc32.Hash(KeyEntry, iPos);
				crc32 = ccc32.Get();
				memcpy(KeyEntry + iPos, &crc32, 4); iPos += 4;
				*/
			}
		}

		DWORD dwReceiveLength = sizeof(ReceiveBuffer);//12+(((ucKeyKind == 3) || (ucKeyKind == 5))?0:1);

		Status = CLSCRF_SAM_APDU(
			pReader,
			SendBuffer,
			iPos,
			ReceiveBuffer,
			&dwReceiveLength
			);

		iPos = 0;
		if (Status == SCARD_S_SUCCESS)
		{
			if (dwReceiveLength == 2)
			{
				*pucSW1 = ReceiveBuffer[iPos++];
				*pucSW2 = ReceiveBuffer[iPos++];
			}
			else if (dwReceiveLength == 10)
			{
				MAC_Data[iPos++] = ReceiveBuffer[8];
				MAC_Data[iPos++] = ReceiveBuffer[9];
				dwCounter++;
				memcpy(MAC_Data + iPos, &dwCounter, 4); iPos += 4;

				DFEncrypt(
					pReader,
					MAC_Data,
					MAC_Data,
					iPos,
					NULL,
					true,
					true);

				for (int i = 0; i < 8; i++)
					MAC_Data[i] = MAC_Data[i * 2 + 1];

				if (memcmp(ReceiveBuffer, MAC_Data, 8) > 0)
				{
					*pucSW1 = ReceiveBuffer[8];
					*pucSW2 = ReceiveBuffer[9];
				}
				else
				{
					Status = SCARD_E_INVALID_VALUE;
				}
			}
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_DisableKeyEntry(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyNo,
IN BYTE ucKeyKind,
IN LPBYTE pbSAM_UID,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		BYTE SendBuffer[256] = { 0 };
		BYTE ReceiveBuffer[256] = { 0 };

		//BYTE KeyEntry[128] = {0};
		int iPos = 0;

		//int iLenEKeyEntry = 0;

		//DWORD dwCounter = 0;
		//BYTE MAC_Data[256] = {0};

		BYTE Kc[24] = { 0 };
		//BYTE Ke[16] = {0};
		//BYTE Km[16] = {0};
		BYTE Kce[24] = { 0 };
		BYTE Kcm[24] = { 0 };
		//BYTE IVLoad[16] = {0x01, 0x01, 0x01, 0x01, 0x00};
		BYTE IV[24] = { 0 };
		//BYTE Output[256] = {0};
		//BYTE DataField[256] = {0};
		WORD wChange_Ctr = 0;

		//AES-128
		if (ucKeyKind == 4)
		{
			BYTE SV1[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71, 0x71 };
			BYTE SV2[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72, 0x72 };

			aes_128_encrypt_cbc(SV1, 16, Kc, IV, Kce);
			aes_128_encrypt_cbc(SV2, 16, Kc, IV, Kcm);

			memcpy(m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey, Kcm, 16);
			m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = CLSCRF_DF_AUTH_TYPE_AES;
			m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize = 16;
		}
		//AES-192
		if (ucKeyKind == 5)
		{
			BYTE SV1a[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73, 0x73 };
			BYTE SV1b[16] = { (BYTE) (wChange_Ctr & 0xFF), (BYTE) ((wChange_Ctr >> 8) & 0xFF), 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74, 0x74 };
			//BYTE SV2a[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75, 0x75};
			//BYTE SV2b[16] = {wChange_Ctr & 0xFF, (wChange_Ctr >> 8) & 0xFF, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76, 0x76};

			BYTE Kcea[24] = { 0 };
			aes_192_encrypt_cbc(SV1a, 24, Kc, IV, Kcea);

			BYTE Kceb[24] = { 0 };
			aes_192_encrypt_cbc(SV1b, 24, Kc, IV, Kceb);

			memcpy(Kce, &Kcea[8], 8);
			memcpy(&Kce[16], Kceb, 8);
			xor_size(Kcea, &Kceb[8], &Kce[8], 8);

			BYTE Kcma[24] = { 0 };
			aes_192_encrypt_cbc(SV1a, 24, Kc, IV, Kcma);

			BYTE Kcmb[24] = { 0 };
			aes_192_encrypt_cbc(SV1b, 24, Kc, IV, Kcmb);

			memcpy(Kcm, &Kcma[8], 8);
			memcpy(&Kcm[16], Kcmb, 8);
			xor_size(Kcma, &Kcmb[8], &Kcm[8], 8);

			memcpy(m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey, Kcm, 24);
			m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = CLSCRF_DF_AUTH_TYPE_AES;
			m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize = 24;
		}

		DFSubkeys(pReader);

		if (pbSAM_UID == NULL)
		{
			SendBuffer[0] = 0x80 | ucLC;
			SendBuffer[1] = 0xD8;
			SendBuffer[2] = ucKeyNo;
			SendBuffer[3] = 0x00;
			SendBuffer[4] = 0x0A;
			memcpy(&SendBuffer[5], &wChange_Ctr, 2);

			DFEncrypt(
				pReader,
				SendBuffer,
				&SendBuffer[7],
				7,
				&iPos,
				true,
				true);

			for (int i = 0; i < 8; i++)
				SendBuffer[7 + i] = SendBuffer[7 + i * 2 + 1];

			iPos += 7;
		}
		else
		{

			SendBuffer[0] = 0x80 | ucLC;
			SendBuffer[1] = 0xD8;
			SendBuffer[2] = ucKeyNo;
			SendBuffer[3] = 0x00;
			SendBuffer[4] = 0x1A;
			memcpy(&SendBuffer[5], &wChange_Ctr, 2);
			aes_128_encrypt_cbc(
				pbSAM_UID,
				7,
				Kce,
				IV,
				&SendBuffer[7]);

			DFEncrypt(
				pReader,
				SendBuffer,
				&SendBuffer[23],
				23,
				&iPos,
				true,
				true);

			for (int i = 0; i < 8; i++)
				SendBuffer[23 + i] = SendBuffer[23 + i * 2 + 1];

			iPos += 23;
		}

		DWORD dwReceiveLength = sizeof(ReceiveBuffer);//12+(((ucKeyKind == 3) || (ucKeyKind == 5))?0:1);

		Status = CLSCRF_SAM_APDU(
			pReader,
			SendBuffer,
			iPos,
			ReceiveBuffer,
			&dwReceiveLength
			);

		iPos = 0;
		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = ReceiveBuffer[iPos++];
			*pucSW2 = ReceiveBuffer[iPos++];
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			*pucSW1 = ReceiveBuffer[iPos++];
			*pucSW2 = ReceiveBuffer[iPos++];
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_GetKeyEntry(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucKeyKind,
IN BYTE ucKeyNo,
OUT LPBYTE pucKeyAVersion,
OUT LPBYTE pucKeyBVersion,
OUT LPBYTE pucKeyCVersion,
OUT LPDWORD pdwDESFireAID,
OUT LPBYTE pucDESFireKeyNo,
OUT LPBYTE pucKeyNoCEK,
OUT LPBYTE pucKeyVCEK,
OUT LPBYTE pucRefNoKUC,
OUT CLSCRF_SAM_SETTINGS* pSettings,
OUT CLSCRF_SAM_EXT_SETTINGS* pExtSettings,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		BYTE SendBuffer[256] = { 0 };
		BYTE ReceiveBuffer[256] = { 0 };

		int iPos = 0;

		SendBuffer[iPos++] = 0x80 | ucLC;
		SendBuffer[iPos++] = 0x64;
		SendBuffer[iPos++] = ucKeyNo;
		SendBuffer[iPos++] = 0x00;
		SendBuffer[iPos++] = 0x00;
		DWORD dwReceiveLength = sizeof(ReceiveBuffer);//12+(((ucKeyKind == 3) || (ucKeyKind == 5))?0:1);

		Status = CLSCRF_SAM_APDU(
			pReader,
			SendBuffer,
			iPos,
			ReceiveBuffer,
			&dwReceiveLength
			);

		iPos = 0;
		if ((Status == SCARD_S_SUCCESS) && (dwReceiveLength >= 0x0E))
		{
			*pucKeyAVersion = ReceiveBuffer[iPos++];
			*pucKeyBVersion = ReceiveBuffer[iPos++];
			if ((ucKeyKind != 3) && (ucKeyKind != 5))
				*pucKeyCVersion = ReceiveBuffer[iPos++];
			memcpy(pdwDESFireAID, &ReceiveBuffer[iPos], 3); iPos += 3;
			*pucDESFireKeyNo = ReceiveBuffer[iPos++];
			*pucKeyNoCEK = ReceiveBuffer[iPos++];
			*pucKeyVCEK = ReceiveBuffer[iPos++];
			*pucRefNoKUC = ReceiveBuffer[iPos++];
			memcpy(&pSettings->wSettings, &ReceiveBuffer[iPos], 2); iPos += 2;
			pExtSettings->ucExtSettings = ReceiveBuffer[iPos++];
		}
		else if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			*pucSW1 = ReceiveBuffer[dwReceiveLength - 2];
			*pucSW2 = ReceiveBuffer[dwReceiveLength - 1];
		}
		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = ReceiveBuffer[dwReceiveLength - 2];
			*pucSW2 = ReceiveBuffer[dwReceiveLength - 1];
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}


CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareAuthenticate(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucAuthType,
IN LPBYTE pbUID,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
IN BYTE ucKeyType,
IN DWORD  dwSector,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xab);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucAuthType);
		pTr->m_TrBuf.AddBulk(pbUID, 4);
		pTr->m_TrBuf.AddByte(ucKeyNo);
		pTr->m_TrBuf.AddByte(ucKeyVer);
		pTr->m_TrBuf.AddByte(ucKeyType);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, 0));

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareRead(IN LPVOID pReader,
IN BYTE ucLC,
IN DWORD  dwSector,
IN DWORD  dwBlock,
OUT LPBYTE pbRecvBuffer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xac);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, dwBlock));

		DWORD Size;
		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			if (pTr->m_RcBuf.GetLength() != 19)
				Status = SCARD_E_COMM_DATA_LOST;
			else
			{
				Size = 16;
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, &Size))
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}

		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen);
			pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen);
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}
	if (Status)
		ZeroMemory(pbRecvBuffer, 16);

	if (Status == SCARD_S_SUCCESS)
	{
		*pucSW1 = 0x90;
		*pucSW2 = 0x00;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareWrite(IN LPVOID pReader,
IN BYTE ucLC,
IN DWORD  dwSector,
IN DWORD  dwBlock,
IN LPBYTE pbSendBuffer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xad);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, dwBlock));
		pTr->m_TrBuf.AddBulk(pbSendBuffer, 16);

		Status = pTr->SingleTransceive();

		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	if (Status == SCARD_S_SUCCESS)
	{
		*pucSW1 = 0x90;
		*pucSW2 = 0x00;
	}

	return Status;
}

LONG SAM_MifareValue(IN LPVOID pReader,
	IN BYTE ucLC,
	IN BYTE ucCmd,
	IN DWORD  dwSector,
	IN DWORD  dwSourceBlock,
	IN DWORD  dwValue,
	IN DWORD  dwTargetBlock,
	OUT LPBYTE pucSW1,
	OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xae);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucCmd);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, dwSourceBlock));
		pTr->m_TrBuf.AddBulk((unsigned char*)&dwValue, 4);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, dwTargetBlock));

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareDecrement(IN LPVOID pReader,
IN BYTE ucLC,
IN DWORD  dwSector,
IN DWORD  dwSourceBlock,
IN DWORD  dwValue,
IN DWORD  dwTargetBlock,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	return SAM_MifareValue(pReader, ucLC, PICC_DECREMENT,
		dwSector, dwSourceBlock, dwValue, dwTargetBlock, pucSW1, pucSW2);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareIncrement(IN LPVOID pReader,
IN BYTE ucLC,
IN DWORD  dwSector,
IN DWORD  dwSourceBlock,
IN DWORD  dwValue,
IN DWORD  dwTargetBlock,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	return SAM_MifareValue(pReader, ucLC, PICC_INCREMENT,
		dwSector, dwSourceBlock, dwValue, dwTargetBlock, pucSW1, pucSW2);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareRestore(IN LPVOID pReader,
IN BYTE ucLC,
IN DWORD  dwSector,
IN DWORD  dwSourceBlock,
IN DWORD  dwTargetBlock,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	return SAM_MifareValue(pReader, ucLC, PICC_RESTORE,
		dwSector, dwSourceBlock, 0, dwTargetBlock, pucSW1, pucSW2);
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareChangeKey(IN LPVOID pReader,
IN BYTE ucLC,
IN CLSCRF_SAM_CK_PARAM_MASK ParamMask,
IN BYTE ucKeyNo,
IN BYTE ucKeyVerA,
IN BYTE ucKeyVerB,
IN DWORD  dwSector,
IN LPBYTE pbAccess,
IN LPBYTE pbUID,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xaf);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ParamMask.ucBits);
		pTr->m_TrBuf.AddByte(ucKeyNo);
		pTr->m_TrBuf.AddByte(ucKeyVerA);
		pTr->m_TrBuf.AddByte(ucKeyVerB);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress(dwSector, 0));
		pTr->m_TrBuf.AddBulk(pbAccess, 4);
		pTr->m_TrBuf.AddBulk(pbUID, 4);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

/*
LONG CLSCRF_SAM_Mifare_Complex(
	IN LPVOID pReader,
	IN BYTE ucSAMLC,
	IN BYTE ucSAMHolder,
	IN BYTE ucSAMKeyNo,
	IN BYTE ucSAMKeyVer,
	IN BYTE ucSAMResetType,
	IN BYTE ucRfOffTime,
	IN BYTE ucRfOnPauseTime,
	IN BYTE ucMifareKeyType,
	IN BYTE ucMifareOperation,
	IN BYTE ucMifareSector,
	IN BYTE ucMifareBlock,
	IN BYTE ucMifareBlocksCount,
	OUT LPBYTE pucATQ,
	OUT LPBYTE pucSAK,
	OUT LPBYTE pucUID,
	IN OUT LPBYTE pucUIDLength,
	IN LPBYTE pucWriteData,
	OUT LPBYTE pucReadData,
	OUT LPBYTE pucLastTask,
	OUT LPBYTE pucResBlkCnt,
	OUT LPBYTE pucSW1,
	OUT LPBYTE pucSW2)
{
	LONG CmdStatus, Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;
	BYTE ucUIDBufSize = 0x00;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	if (!pucLastTask || !pucResBlkCnt || !pucSW1 || !pucSW2)
		return SCARD_E_INVALID_VALUE;

	*pucLastTask = 0x00;
	*pucResBlkCnt = 0x00;
	*pucSW1 = 0x00;
	*pucSW2 = 0x00;

	bool fExpectUID = !((ucRfOnPauseTime == 0x00) && (ucRfOffTime == 0x00));

	if (fExpectUID && (!pucATQ || !pucSAK || !pucUID || !pucUIDLength))
		return SCARD_E_INVALID_VALUE;

	if (fExpectUID)
	{
		pucATQ[0] = 0x00;
		pucATQ[1] = 0x00;
		*pucSAK = 0x00;
		ZeroMemory(pucUID, *pucUIDLength);
		ucUIDBufSize = *pucUIDLength;
		*pucUIDLength = 0x00;
	}

	bool fReadData = (ucMifareBlocksCount > 0) && (ucMifareOperation == 0x01 || ucMifareOperation == 0x03);
	bool fWriteData = (ucMifareBlocksCount > 0) && (ucMifareOperation == 0x02 || ucMifareOperation == 0x03);

	if (fReadData && !pucReadData)
		return SCARD_E_INVALID_VALUE;

	if (fWriteData && !pucWriteData)
		return SCARD_E_INVALID_VALUE;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x58);
	pTr->m_TrBuf.AddByte(0xAB);
	pTr->m_TrBuf.AddByte((ucSAMLC & 0x03) | ((ucSAMResetType & 0x03) << 2) | ((ucSAMHolder & 0x07) << 4));
	pTr->m_TrBuf.AddByte((ucRfOnPauseTime & 0x0F) | ((ucRfOffTime & 0x0F) << 4));
	pTr->m_TrBuf.AddByte((ucMifareKeyType & 0x0F) | ((ucMifareOperation & 0x0F) << 4));
	pTr->m_TrBuf.AddByte(ucSAMKeyNo);
	pTr->m_TrBuf.AddByte(ucSAMKeyVer);
	pTr->m_TrBuf.AddByte(ucMifareSector);
	pTr->m_TrBuf.AddByte(ucMifareBlock);
	pTr->m_TrBuf.AddByte(ucMifareBlocksCount);
	if (fWriteData)
		pTr->m_TrBuf.AddBulk(pucWriteData, 16 * ucMifareBlocksCount);

	CmdStatus = pTr->SingleTransceive();

	DWORD dwDataPos = 1;

	dwLen = 1;
	if (pTr->m_RcBuf.GetBulk(dwDataPos, pucSW1, &dwLen))
	{
		dwDataPos += 1;
		dwLen = 1;
		if (pTr->m_RcBuf.GetBulk(dwDataPos, pucSW2, &dwLen))
		{
			dwDataPos += 1;
			dwLen = 1;
			BYTE ucLastResult = 0x00;
			if (pTr->m_RcBuf.GetBulk(dwDataPos, &ucLastResult, &dwLen))
			{
				dwDataPos += 1;
				*pucLastTask = ucLastResult >> 4;
				*pucResBlkCnt = ucLastResult & 0x0F;
			}
			else
				Status = SCARD_E_COMM_DATA_LOST;
		}
		else
			Status = SCARD_E_COMM_DATA_LOST;
	}
	else
		Status = SCARD_E_COMM_DATA_LOST;

	if ((Status == 0) && (*pucLastTask > 2))// After ActivateIdleA
	{
		if (fExpectUID)
		{
			dwLen = 2;
			if (pTr->m_RcBuf.GetBulk(dwDataPos, pucATQ, &dwLen))
			{
				dwDataPos += 2;
				dwLen = 1;
				if (pTr->m_RcBuf.GetBulk(dwDataPos, pucSAK, &dwLen))
				{
					dwDataPos += 1;
					dwLen = 1;
					if (pTr->m_RcBuf.GetBulk(dwDataPos, pucUIDLength, &dwLen))
					{
						dwDataPos += 1;
						if (*pucUIDLength > 0)
						{
							if (*pucUIDLength <= ucUIDBufSize)
							{
								DWORD dwUIDLength = *pucUIDLength;
								if (pTr->m_RcBuf.GetBulk(dwDataPos, pucUID, &dwUIDLength))
								{
									dwDataPos += dwUIDLength;
									*pucUIDLength = (BYTE)dwUIDLength;
								}
								else
								{
									dwDataPos += *pucUIDLength;
									Status = SCARD_E_COMM_DATA_LOST;
								}
							}
							else
								Status = SCARD_E_INSUFFICIENT_BUFFER;
						}
					}
					else
						Status = SCARD_E_COMM_DATA_LOST;
				}
				else
					Status = SCARD_E_COMM_DATA_LOST;
			}
			else
				Status = SCARD_E_COMM_DATA_LOST;
		}

		if ((fReadData) && (Status != SCARD_E_COMM_DATA_LOST) && (*pucResBlkCnt))
		{
			dwLen = *pucResBlkCnt * 16;
			if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucReadData, &dwLen))
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}

	if (CmdStatus == SCARD_S_SUCCESS)
		return Status;
	else
		return CmdStatus;
}
*/

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_Mifare_Complex(IN LPVOID pReader,
IN BYTE ucSAMLC,
IN BYTE ucSAMHolder,
IN BYTE ucSAMKeyNo,
IN BYTE ucSAMKeyVer,
IN BYTE ucSAMResetType,
IN BYTE ucRfOffTime,
IN BYTE ucRfOnPauseTime,
IN BYTE ucMifareKeyType,
IN BYTE ucMifareOperation,
IN BYTE ucMifareSector,
IN BYTE ucMifareBlock,
IN BYTE ucMifareBlocksCount,
OUT LPBYTE pucATQ,
OUT LPBYTE pucSAK,
OUT LPBYTE pucUID,
IN OUT LPBYTE pucUIDLength,
IN LPBYTE pucWriteData,
OUT LPBYTE pucReadData, 
OUT LPBYTE pucLastTask,
OUT LPBYTE pucResBlkCnt,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;
	BYTE ucUIDBufSize = 0x00;

	if (!pReader)
		return SCARD_E_READER_UNAVAILABLE;

	if (!pucSW1 || !pucSW2)
		return SCARD_E_INVALID_VALUE;

	*pucSW1 = 0x00;
	*pucSW2 = 0x00;

	bool fExpectUID = !((ucRfOnPauseTime == 0x00) && (ucRfOffTime == 0x00));

	if (fExpectUID && (!pucATQ || !pucSAK || !pucUID || !pucUIDLength))
		return SCARD_E_INVALID_VALUE;

	if (pucATQ)
		*(PUSHORT)pucATQ = 0x0000;
	if (pucSAK)
		*pucSAK = 0x00;
	if (pucUID && pucUIDLength)
	{
		ZeroMemory(pucUID, *pucUIDLength);
		ucUIDBufSize = *pucUIDLength;
		*pucUIDLength = 0x00;
	}

	if (pucLastTask)
		*pucLastTask = 0x00;
	if (pucResBlkCnt)
		*pucResBlkCnt = 0x00;

	bool fReadData = (ucMifareBlocksCount > 0) && (ucMifareOperation == 0x01 || ucMifareOperation == 0x03);
	bool fWriteData = (ucMifareBlocksCount > 0) && (ucMifareOperation == 0x02 || ucMifareOperation == 0x03);

	if (fReadData && !pucReadData)
		return SCARD_E_INVALID_VALUE;

	if (fWriteData && !pucWriteData)
		return SCARD_E_INVALID_VALUE;


	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0x58);
	pTr->m_TrBuf.AddByte(0xAB);
	pTr->m_TrBuf.AddByte((ucSAMLC & 0x03) | ((ucSAMResetType & 0x03) << 2) | ((ucSAMHolder & 0x07) << 4));
	pTr->m_TrBuf.AddByte((ucRfOnPauseTime & 0x0F) | ((ucRfOffTime & 0x0F) << 4));
	pTr->m_TrBuf.AddByte((ucMifareKeyType & 0x0F) | ((ucMifareOperation & 0x0F) << 4));
	pTr->m_TrBuf.AddByte(ucSAMKeyNo);
	pTr->m_TrBuf.AddByte(ucSAMKeyVer);
	pTr->m_TrBuf.AddByte(ucMifareSector);
	pTr->m_TrBuf.AddByte(ucMifareBlock);
	pTr->m_TrBuf.AddByte(ucMifareBlocksCount);
	if (fWriteData)
		pTr->m_TrBuf.AddBulk(pucWriteData, 16 * ucMifareBlocksCount);

	Status = pTr->SingleTransceive();

	DWORD dwDataPos = 1;

	BYTE ucExpectBytes = 4;

	if (pTr->m_RcBuf.GetLength() < ucExpectBytes)
		return SCARD_E_COMM_DATA_LOST;

	dwLen = 1;
	if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucSW1, &dwLen))
		return SCARD_E_COMM_DATA_LOST;
	dwDataPos += 1;
	dwLen = 1;
	if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucSW2, &dwLen))
		return SCARD_E_COMM_DATA_LOST;
	dwDataPos += 1;

	dwLen = 1;
	BYTE ucLastResult = 0x00;
	if (!pTr->m_RcBuf.GetBulk(dwDataPos, &ucLastResult, &dwLen))
		return SCARD_E_COMM_DATA_LOST;
	dwDataPos += 1;

	if (pucLastTask)
		*pucLastTask = ucLastResult >> 4;
	if (pucResBlkCnt)
		*pucResBlkCnt = ucLastResult & 0x0F;

	if (Status != SCARD_S_SUCCESS)
		return Status;

	//*pucSW1 = 0x90;
	//*pucSW2 = 0x00;

	ucExpectBytes += (fExpectUID ? 4 : 0) + (fReadData ? 16 * ucMifareBlocksCount : 0);
			
	if (pTr->m_RcBuf.GetLength() < ucExpectBytes)
		return SCARD_E_COMM_DATA_LOST;

	if (fExpectUID) 
	{
		dwLen = 2;
		if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucATQ, &dwLen))
			return SCARD_E_COMM_DATA_LOST;
		dwDataPos += 2;

		dwLen = 1;
		if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucSAK, &dwLen))
			return SCARD_E_COMM_DATA_LOST;
		dwDataPos += 1;

		dwLen = 1;
		if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucUIDLength, &dwLen))
			return SCARD_E_COMM_DATA_LOST;
		dwDataPos += 1;

		if (*pucUIDLength > 0)
		{
			if (*pucUIDLength > ucUIDBufSize)
				return SCARD_E_INSUFFICIENT_BUFFER;

			ucExpectBytes += *pucUIDLength;

			if (pTr->m_RcBuf.GetLength() < ucExpectBytes)
				return SCARD_E_COMM_DATA_LOST;

			DWORD dwUIDLength = *pucUIDLength;
			if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucUID, &dwUIDLength))
				return SCARD_E_COMM_DATA_LOST;
			dwDataPos += dwUIDLength;
			*pucUIDLength = dwUIDLength;
		}
	}

	if (fReadData && ((ucLastResult & 0x0F) > 0))
	{
		dwLen = 16 * (ucLastResult & 0x0F);
		if (!pTr->m_RcBuf.GetBulk(dwDataPos, pucReadData, &dwLen))
			return SCARD_E_COMM_DATA_LOST;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareUltralightC_Authenticate(IN LPVOID pReader,
IN BYTE ucLC,
IN LPBYTE pbUID,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x2e);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0x00); // RFU
		pTr->m_TrBuf.AddByte(ucKeyNo);
		pTr->m_TrBuf.AddByte(ucKeyVer);
		pTr->m_TrBuf.AddBulk(pbUID, 7);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifareUltralightC_WriteKey(IN LPVOID pReader,
IN BYTE ucLC,
IN LPBYTE pbUID,
IN BYTE ucKeyNo,
IN BYTE ucKeyVer,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0x2f);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0x00); // RFU
		pTr->m_TrBuf.AddByte(ucKeyNo);
		pTr->m_TrBuf.AddByte(ucKeyVer);
		pTr->m_TrBuf.AddBulk(pbUID, 7);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_WritePerso(IN LPVOID pReader,
IN BYTE  ucLC,
IN BYTE  ucValueType,
IN BYTE  ucSectorNumber,
IN BYTE  ucBlockNumber,
IN BYTE  ucKeyNumber,
IN BYTE  ucKeyVersion,
IN BYTE  ucLenDiv,
IN LPBYTE pbDiv,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF8);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucValueType);
		pTr->m_TrBuf.AddByte(ucSectorNumber);
		pTr->m_TrBuf.AddByte(ucBlockNumber);
		pTr->m_TrBuf.AddByte(ucKeyNumber);
		pTr->m_TrBuf.AddByte(ucKeyVersion);
		pTr->m_TrBuf.AddByte(ucLenDiv);
		if (ucLenDiv > 0)
			pTr->m_TrBuf.AddBulk(pbDiv, ucLenDiv);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_CommitPerso(IN LPVOID pReader,
IN BYTE  ucLC,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	DWORD dwPass = 0x87654321;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xFA);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwPass, 4);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_Authenticate(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE  ucAuthType,
IN BYTE  ucSecLevel,
IN BYTE  ucKeyType,
IN BYTE  ucSectorNumber,
IN BYTE  ucKeyNumber,
IN BYTE  ucKeyVersion,
IN BYTE  ucLenCap,
IN LPBYTE pbPcdCap,
IN BYTE  ucLenDiv,
IN LPBYTE pbDiv,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF0);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(
			((ucAuthType > 0x01) ? 0x02 : 0x00)
			| ((ucSecLevel << 2) & 0x0C)
			);
		pTr->m_TrBuf.AddByte(ucKeyType);
		pTr->m_TrBuf.AddByte(ucSectorNumber);
		pTr->m_TrBuf.AddByte((ucAuthType == 0x00) ? 0x00 : ucKeyNumber);// ucAuthType == 0x00 - ñáðîñ àóòåíòèôèêàöèè
		pTr->m_TrBuf.AddByte(ucKeyVersion);
		pTr->m_TrBuf.AddByte(ucLenCap);
		if (ucLenCap > 0)
			if (pbPcdCap)
				pTr->m_TrBuf.AddBulk(pbPcdCap, ucLenCap);
		pTr->m_TrBuf.AddByte(ucLenDiv);
		if (ucLenDiv > 0)
			if (pbDiv)
				pTr->m_TrBuf.AddBulk(pbDiv, ucLenDiv);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_MultiBlockRead(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucSectorNumber,
IN BYTE ucBlockNumber,
IN BYTE ucBlocksCount,
OUT LPBYTE pbData,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		if (ucBlocksCount < 1)
			ucBlocksCount = 1;

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF6);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress((DWORD)ucSectorNumber, (DWORD)ucBlockNumber));
		pTr->m_TrBuf.AddByte(ucBlocksCount);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;

			DWORD Size = 1;

			BYTE bStatusCode = 0x00;

			if (pTr->m_RcBuf.GetBulk(0, &bStatusCode, &Size))
			{
				if (bStatusCode == 0x00)
				{
					Size = 2;

					DWORD dwLen = 0;

					if (pTr->m_RcBuf.GetBulk(3, (PBYTE)&dwLen, &Size))
					{
						if (dwLen > 0)
						{
							pTr->m_RcBuf.GetBulk(5, pbData, &dwLen);
						}
					}
				}
			}

		}
		else if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			DWORD Size = 1;

			pTr->m_RcBuf.GetBulk(1, pucSW1, &Size);
			pTr->m_RcBuf.GetBulk(2, pucSW2, &Size);
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_MultiBlockWrite(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucSectorNumber,
IN BYTE ucBlockNumber,
IN BYTE ucBlocksCount,
IN LPBYTE pbData,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		if (ucBlocksCount < 1)
			ucBlocksCount = 1;

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF7);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(GetActualBlockAddress((DWORD)ucSectorNumber, (DWORD)ucBlockNumber));
		pTr->m_TrBuf.AddByte(ucBlocksCount);
		pTr->m_TrBuf.AddBulk(pbData, ucBlocksCount * 16);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			DWORD Size = 1;

			pTr->m_RcBuf.GetBulk(1, pucSW1, &Size);
			pTr->m_RcBuf.GetBulk(2, pucSW2, &Size);
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_ReadData(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucReadMode,
IN BYTE ucValueType,
IN BYTE ucSectorNumber,
IN BYTE ucBlockNumber,
IN BYTE ucBlocksCount,
OUT LPBYTE pbData,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		if (ucBlocksCount < 1)
			ucBlocksCount = 1;

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF4);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucReadMode);
		pTr->m_TrBuf.AddByte(ucValueType);
		pTr->m_TrBuf.AddByte(ucSectorNumber);
		pTr->m_TrBuf.AddByte(ucBlockNumber);
		pTr->m_TrBuf.AddByte(ucBlocksCount);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;

			DWORD Size = 1;

			BYTE bStatusCode = 0x00;

			if (pTr->m_RcBuf.GetBulk(0, &bStatusCode, &Size))
			{
				if (bStatusCode == 0x00)
				{
					Size = 2;

					DWORD dwLen = 0;

					if (pTr->m_RcBuf.GetBulk(3, (PBYTE)&dwLen, &Size))
					{
						if (dwLen > 0)
						{
							pTr->m_RcBuf.GetBulk(5, pbData, &dwLen);
						}
					}
				}
			}

		}
		else if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			DWORD Size = 1;

			pTr->m_RcBuf.GetBulk(1, pucSW1, &Size);
			pTr->m_RcBuf.GetBulk(2, pucSW2, &Size);
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_WriteData(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucWriteMode,
IN BYTE ucValueType,
IN BYTE ucSectorNumber,
IN BYTE ucBlockNumber,
IN BYTE ucBlocksCount,
IN LPBYTE pbData,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		if (ucBlocksCount < 1)
			ucBlocksCount = 1;

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF5);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucWriteMode);
		pTr->m_TrBuf.AddByte(ucValueType);
		pTr->m_TrBuf.AddByte(ucSectorNumber);
		pTr->m_TrBuf.AddByte(ucBlockNumber);
		pTr->m_TrBuf.AddByte(ucBlocksCount);
		pTr->m_TrBuf.AddBulk(pbData, (DWORD)ucBlocksCount * 16);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_ChangeKey(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucKeyType,
IN BYTE ucSectorNumber,
IN BYTE ucKeyNumber,
IN BYTE ucKeyVersion,
IN BYTE ucLenDiv,
IN LPBYTE pbDiv,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xFD);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucKeyType);
		pTr->m_TrBuf.AddByte(ucSectorNumber);
		pTr->m_TrBuf.AddByte(ucKeyNumber);
		pTr->m_TrBuf.AddByte(ucKeyVersion);
		pTr->m_TrBuf.AddByte(ucLenDiv);
		if (ucLenDiv > 0)
			if (pbDiv)
				pTr->m_TrBuf.AddBulk(pbDiv, ucLenDiv);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_Increment(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucSourceSectorNumber,
IN BYTE ucSourceBlockNumber,
IN DWORD dwValue,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xB0);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucSourceSectorNumber);
		pTr->m_TrBuf.AddByte(ucSourceBlockNumber);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwValue, 4);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_Decrement(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucSourceSectorNumber,
IN BYTE ucSourceBlockNumber,
IN DWORD dwValue,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xB2);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucSourceSectorNumber);
		pTr->m_TrBuf.AddByte(ucSourceBlockNumber);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwValue, 4);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_Transfer(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucDestinationSectorNumber,
IN BYTE ucDestinationBlockNumber,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xB4);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucDestinationSectorNumber);
		pTr->m_TrBuf.AddByte(ucDestinationBlockNumber);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_IncrementTransfer(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucSourceSectorNumber,
IN BYTE ucSourceBlockNumber,
IN BYTE ucDestinationSectorNumber,
IN BYTE ucDestinationBlockNumber,
IN DWORD dwValue,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xB6);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucSourceSectorNumber);
		pTr->m_TrBuf.AddByte(ucSourceBlockNumber);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwValue, 4);
		pTr->m_TrBuf.AddByte(ucDestinationSectorNumber);
		pTr->m_TrBuf.AddByte(ucDestinationBlockNumber);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_DecrementTransfer(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucSourceSectorNumber,
IN BYTE ucSourceBlockNumber,
IN BYTE ucDestinationSectorNumber,
IN BYTE ucDestinationBlockNumber,
IN DWORD dwValue,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xB8);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucSourceSectorNumber);
		pTr->m_TrBuf.AddByte(ucSourceBlockNumber);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwValue, 4);
		pTr->m_TrBuf.AddByte(ucDestinationSectorNumber);
		pTr->m_TrBuf.AddByte(ucDestinationBlockNumber);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_Restore(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucEncryptionMode,
IN BYTE ucSourceSectorNumber,
IN BYTE ucSourceBlockNumber,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF1);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0xC2);
		pTr->m_TrBuf.AddByte(ucEncryptionMode);
		pTr->m_TrBuf.AddByte(ucSourceSectorNumber);
		pTr->m_TrBuf.AddByte(ucSourceBlockNumber);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_ProximityCheck(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE  ucKeyNumber,
IN BYTE  ucKeyVersion,
IN BYTE  ucLenDiv,
IN LPBYTE pbDiv,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF9);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucKeyNumber);
		pTr->m_TrBuf.AddByte(ucKeyVersion);
		pTr->m_TrBuf.AddByte(ucLenDiv);
		if (ucLenDiv > 0)
			if (pbDiv)
				pTr->m_TrBuf.AddBulk(pbDiv, ucLenDiv);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_VirtualCardSupport(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE ucIIDCount,
IN LPBYTE IIDs,
IN BYTE ucLenCap,
IN LPBYTE pbPcdCap,
IN BYTE ucDuosCount,
IN CLSCRF_SAM_VC_DUO* Duos,
OUT LPBYTE pbRndQ,
OUT LPBYTE pucInfo,//(0x83 - 4-áàéò UID, 0x03 - 7-áàéò UID)
OUT LPBYTE pbPiccCap,
OUT LPBYTE pbUID,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		if ((ucIIDCount == 0) || (ucDuosCount == 0))
			return SCARD_E_INVALID_PARAMETER;

		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF2);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(ucIIDCount);
		pTr->m_TrBuf.AddBulk(IIDs, 16 * ucIIDCount);
		pTr->m_TrBuf.AddByte(ucLenCap);
		if (ucLenCap > 0)
			if (pbPcdCap)
				pTr->m_TrBuf.AddBulk(pbPcdCap, ucLenCap);
		pTr->m_TrBuf.AddByte(ucDuosCount);
		pTr->m_TrBuf.AddBulk((LPBYTE)Duos, 4 * ucDuosCount);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;

			DWORD dwSize = 12;
			pTr->m_RcBuf.GetBulk(3, pbRndQ, &dwSize);

			dwSize = 1;
			pTr->m_RcBuf.GetBulk(15, pucInfo, &dwSize);

			dwSize = 2;
			pTr->m_RcBuf.GetBulk(16, pbPiccCap, &dwSize);

			dwSize = 7;
			pTr->m_RcBuf.GetBulk(18, pbUID, &dwSize);
		}
		else if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_VirtualCardSelect(IN LPVOID pReader,
IN BYTE ucLC,
IN BYTE  ucKeyNumber,
IN BYTE  ucKeyVersion,
IN LPBYTE pbRndQ,
IN BYTE  ucLenDiv,
IN LPBYTE pbDiv,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF3);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0x01);
		pTr->m_TrBuf.AddByte(ucKeyNumber);
		pTr->m_TrBuf.AddByte(ucKeyVersion);
		pTr->m_TrBuf.AddBulk(pbRndQ, 12);
		pTr->m_TrBuf.AddByte(ucLenDiv);
		if (ucLenDiv > 0)
			if (pbDiv)
				pTr->m_TrBuf.AddBulk(pbDiv, ucLenDiv);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

//=====================================================================

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_SAM_MifarePlus_VirtualCardDeselect(IN LPVOID pReader,
IN BYTE ucLC,
OUT LPBYTE pucSW1,
OUT LPBYTE pucSW2)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD dwLen = 1;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();
		pTr->m_TrBuf.AddByte(0xF3);
		pTr->m_TrBuf.AddByte(ucLC);
		pTr->m_TrBuf.AddByte(0x00);

		Status = pTr->SingleTransceive();

		if (Status == SCARD_S_SUCCESS)
		{
			*pucSW1 = 0x90;
			*pucSW2 = 0x00;
		}
		if ((Status == SCARD_F_INTERNAL_ERROR) && (GetLastInternalError(pReader) == UEM_SAM_APDU_ERR))
		{
			if (
				(!pTr->m_RcBuf.GetBulk(1, pucSW1, &dwLen))
				|| (!pTr->m_RcBuf.GetBulk(2, pucSW2, &dwLen))
				)
				Status = SCARD_E_COMM_DATA_LOST;
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}



//=====================================================================

