#include "StdAfx.h"
#include "Clscrfl.h"
#include "Clscrfl_SAM.h"
#ifdef CLSCRF_DES3
	#include "Clscrfl_DES3.h"
#endif //CLSCRF_DES3
#include <fstream>
#include "SAM_Set.h"
#include "Types.h"

extern std::map<PVOID, READER_SETTINGS> m_ReaderSettings;

//extern std::map<PVOID, DF_SETTINGS> m_DF_Settings;

//extern std::map<PVOID, CRYPTO_SETTINGS> m_CryptoSettings;

void DFSubkeys(PVOID pReader)//, BYTE *key, BYTE KeySize, BYTE *K1, BYTE *K2)
{
	BYTE *key = m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey;
	//BYTE KeySize = m_DF_Settings[pReader].SessionKeySize;
	BYTE BlockSize = (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_AES)?16:8;
	BYTE *K1 = m_ReaderSettings[pReader].CurrentDFSettings()->K1;
	BYTE *K2 = m_ReaderSettings[pReader].CurrentDFSettings()->K2;

	BYTE const_Rb[16] = {0};
	const_Rb[BlockSize-1] = (BlockSize == 16)?0x87:0x1B;

	BYTE L[16] = {0};
	BYTE Z[16] = {0};
	BYTE tmp[16] = {0};

	switch (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType)
	{
	case CLSCRF_DF_AUTH_TYPE_DES:
		des_cbc_encrypt(
			key,
			Z,
			L,
			BlockSize );
		break;
	case CLSCRF_DF_AUTH_TYPE_3DES:
		des_3cbc_encrypt(
			key,
			key+8,
			Z,
			L,
			BlockSize );
		break;
	case CLSCRF_DF_AUTH_TYPE_3K3DES:
		des_3k3cbc_encrypt(
			key,
			key+8,
			key+16,
			Z,
			L,
			BlockSize );
		break;
	case CLSCRF_DF_AUTH_TYPE_AES:
		aes_128_encrypt_cbc( 
			Z,
			BlockSize,
			key,
			tmp,
			L );
		break;
	default:
		break;
	}

    if ( (L[0] & 0x80) == 0 ) { /* If MSB(L) = 0, then K1 = L << 1 */
        leftshift_onebit(L,K1,BlockSize);
    } else { /* Else K1 = ( L << 1 ) (+) Rb */

        leftshift_onebit(L,tmp,BlockSize);
        xor_size(tmp,const_Rb,K1,BlockSize);
    }

    if ( (K1[0] & 0x80) == 0 ) 
    {
        leftshift_onebit(K1,K2,BlockSize);
    } else 
    {
        leftshift_onebit(K1,tmp,BlockSize);
        xor_size(tmp,const_Rb,K2,BlockSize);
    }
    
    return;
}


void DFEncrypt(PVOID pReader, PBYTE pInput, PBYTE pOutput, int iSize, int* piResultSize, bool f80Padding, bool calc_mac, PBYTE crc32Start, int crc32Size)
{
	if ((!pInput)||(!pOutput)||(iSize<=0))//||(m_DF_Settings[pReader].SessionKeySize == 0))
		return;

	BYTE TmpInp[1024*16] = {0};
	BYTE TmpBuf[1924*16] = {0};
	
	BYTE IV[16] = {0};

	memcpy(IV, m_ReaderSettings[pReader].CurrentDFSettings()->IV, 16);

	int iMacSize = 0;
	//USHORT _crc16 = 0;

	ULONG crc32 = 0;
	CRC32 ccc32;

	memcpy(TmpInp, pInput, iSize);
/*
	if (inc_crc32)
	{
		DWORD dwCrc32 = CRC_32(TmpInp, iSize);
		memcpy(&TmpInp[iSize], &dwCrc32, 4); iSize += 4;
	}
*/
	PBYTE mac_key = NULL;

	//memcpy(IV, m_DF_Settings[pReader].IV, 24);
	if (calc_mac)
		f80Padding = true;
//USHORT wCrc;
//			crc16 crc;
	switch (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType)
	{
	case CLSCRF_DF_AUTH_TYPE_DES:
		/*
		if (f80Padding&&(iSize%8))
		{
			TmpInp[iSize] = 0x80;
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K1;
		}
		else
		{
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K2;
		}
*/
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			iMacSize = 8*(int)ceil((float)iSize/8);
			des_cbc_encrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				TmpInp,
				TmpBuf,
				iMacSize,
				IV,
				mac_key );
			memmove(&TmpBuf[iSize], &TmpBuf[iMacSize-8], 4);
			memcpy(TmpBuf, TmpInp, iSize);
			iSize += 4;
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			//for (int i = 0; i < 2; i++)
			//	crc.calc_crc16(TmpInp[i+10]);
			
			//wCrc = crc.return_crc16();

			//_crc16 = CRC_16(TmpInp+10, 2);
			//_crc16 = CRC_16(TmpInp, iSize);

			//memcpy(&TmpInp[iSize], (PVOID)&_crc16, 2);

			//for (int i = 0; i < iSize; i++)
			//	crc.calc_crc16(TmpInp[i]);
			
			//wCrc = crc.return_crc16();

			//memcpy(&TmpInp[iSize], (PVOID)&wCrc, 2);

			//ComputeCrc(CRC_A, (char*)TmpInp+10, 2, TmpInp+10, TmpInp+11);

			ComputeCrc(CRC_A, (char*)TmpInp, iSize, &TmpInp[iSize], &TmpInp[iSize+1]);

			iSize +=2;
			iSize = 8*(int)ceil((float)iSize/8);
			des_cbc_decrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				TmpInp,
				TmpBuf,
				iSize,
				IV );
			break;
		default:
			break;
		}

		break;
	case CLSCRF_DF_AUTH_TYPE_3DES:
		/*
		if (f80Padding&&(iSize%8))
		{
			TmpInp[iSize] = 0x80;
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K1;
		}
		else
		{
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K2;
		}
*/
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			iMacSize = 8*(int)ceil((float)iSize/8);
			des_3cbc_encrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				TmpInp,
				TmpBuf,
				iMacSize,
				IV,
				IV+8,
				mac_key );
			memmove(&TmpBuf[iSize], &TmpBuf[iMacSize-8], 4);
			memcpy(TmpBuf, TmpInp, iSize);
			iSize += 4;
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			//for (int i = 0; i < 2; i++)
			//	crc.calc_crc16(TmpInp[i+10]);
			
			//wCrc = crc.return_crc16();

			//_crc16 = CRC_16(TmpInp+10, 2);
			//_crc16 = CRC_16(TmpInp, iSize);

			//memcpy(&TmpInp[iSize], (PVOID)&_crc16, 2);

			//for (int i = 0; i < iSize; i++)
			//	crc.calc_crc16(TmpInp[i]);
			
			//wCrc = crc.return_crc16();

			//memcpy(&TmpInp[iSize], (PVOID)&wCrc, 2);

			//ComputeCrc(CRC_A, (char*)TmpInp+10, 2, TmpInp+10, TmpInp+11);

			ComputeCrc(CRC_A, (char*)TmpInp, iSize, &TmpInp[iSize], &TmpInp[iSize+1]);

			iSize +=2;
			iSize = 8*(int)ceil((float)iSize/8);
			des_3cbc_encipher_by_decrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				TmpInp,
				TmpBuf,
				iSize,
				IV,
				IV+8 );
			break;
		default:
			break;
		}

/*		if (f80Padding&&(iSize%8))
		{
			TmpInp[iSize] = 0x80;
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K1;
		}
		else
		{
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K2;
		}
		iSize = 8*ceil((float)iSize/8);
		des_3cbc_encrypt(
			m_DF_Settings[pReader].SessionKey,
			m_DF_Settings[pReader].SessionKey+8,
			TmpInp,
			TmpBuf,
			iSize,
			IV,
			IV+8,
			mac_key );*/
		break;
	case CLSCRF_DF_AUTH_TYPE_3K3DES:
		if (calc_mac)
		{
			if (f80Padding&&(iSize%8))
			{
				TmpInp[iSize] = 0x80;
				if (calc_mac)
					mac_key = m_ReaderSettings[pReader].CurrentDFSettings()->K2;
			}
			else
			{
				if (calc_mac)
					mac_key = m_ReaderSettings[pReader].CurrentDFSettings()->K1;
			}
			iSize = 8*(int)ceil((float)iSize/8);
			des_3k3cbc_encrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+16,
				TmpInp,
				TmpBuf,
				iSize,
				IV,
				IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				mac_key );
		}
		else
		{
			switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
			{
			case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
				memcpy(TmpBuf, TmpInp, iSize);
				break;
			case CLSCRF_DF_TRANSFER_TYPE_MACCED:
				memcpy(TmpBuf, crc32Start, crc32Size);
				memcpy(&TmpBuf[crc32Size], TmpInp, iSize);
				iMacSize = crc32Size + iSize;
				DFEncrypt(
					pReader, 
					TmpBuf, 
					m_ReaderSettings[pReader].CurrentDFSettings()->IV,
					iMacSize, 
					NULL, 
					true, 
					true );

				memcpy( TmpBuf, TmpInp, iSize );
				memcpy( &TmpBuf[iSize], m_ReaderSettings[pReader].CurrentDFSettings()->IV, 8 );
				iSize += 8;

				break;
			case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
				if (crc32Start&&crc32Size)
				{
					ccc32.Hash(crc32Start, crc32Size);
				}

				ccc32.Hash(TmpInp, iSize);
				crc32 = ccc32.Get();

				memcpy(&TmpInp[iSize], (PVOID)&crc32,4);

				iSize +=4;
				iSize = 8*(int)ceil((float)iSize/8);

				des_3k3cbc_encrypt(
					m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
					m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
					m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+16,
					TmpInp,
					TmpBuf,
					iSize,
					IV,
					IV,
					m_ReaderSettings[pReader].CurrentDFSettings()->IV
					);
					break;
			default:
				break;
			}
		}
		break;
	case CLSCRF_DF_AUTH_TYPE_AES:
		if (calc_mac)
		{
			if (f80Padding&&(iSize%16))
			{
				TmpInp[iSize] = 0x80;
				mac_key = m_ReaderSettings[pReader].CurrentDFSettings()->K2;
			}
			else
			{
				mac_key = m_ReaderSettings[pReader].CurrentDFSettings()->K1;
			}
			iSize = 16*(int)ceil((float)iSize/16);
			/*AES_128_CBC_En_MAC(
				TmpInp,
				iSize/16,
				m_DF_Settings[pReader].SessionKey,
				TmpBuf,
				IV );*/

			aes_128_encrypt_cbc(
				TmpInp,
				iSize,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				IV,
				TmpBuf,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				mac_key);
		}
		else
		{
			switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
			{
			case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
				memcpy(TmpBuf, TmpInp, iSize);
				break;
			case CLSCRF_DF_TRANSFER_TYPE_MACCED:
				memcpy(TmpBuf, crc32Start, crc32Size);
				memcpy(&TmpBuf[crc32Size], TmpInp, iSize);
				iMacSize = crc32Size + iSize;
				DFEncrypt(
					pReader, 
					TmpBuf, 
					m_ReaderSettings[pReader].CurrentDFSettings()->IV,
					iMacSize, 
					NULL, 
					true, 
					true );

				memcpy( TmpBuf, TmpInp, iSize );
				memcpy( &TmpBuf[iSize], m_ReaderSettings[pReader].CurrentDFSettings()->IV, 8 );
				iSize += 8;

				/*
				if (f80Padding&&(iSize%16))
				{
					TmpInp[iSize] = 0x80;
					if (calc_mac)
						mac_key = m_DF_Settings[pReader].K2;
				}
				else
				{
					if (calc_mac)
						mac_key = m_DF_Settings[pReader].K1;
				}
				iSize = 16*ceil((float)iSize/16);
				AES_128_CBC_En_MAC(
					TmpInp,
					iSize/16,
					m_DF_Settings[pReader].SessionKey,
					TmpBuf,
					IV );
					*/
	/*
				aes_128_encrypt_cbc(
					TmpInp,
					iSize,
					m_DF_Settings[pReader].SessionKey,
					IV,
					TmpBuf,
					m_DF_Settings[pReader].IV,
					mac_key);*/
//				break;
				break;
			case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
				if (crc32Start&&crc32Size)
				{
					ccc32.Hash(crc32Start, crc32Size);
				}

				ccc32.Hash(TmpInp, iSize);
				crc32 = ccc32.Get();

				memcpy(&TmpInp[iSize], (PVOID)&crc32,4);

				iSize +=4;
				iSize = 16*(int)ceil((float)iSize/16);

				aes_128_encrypt_cbc(
					TmpInp,
					iSize,
					m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
					IV,
					TmpBuf,
					m_ReaderSettings[pReader].CurrentDFSettings()->IV
					);
				break;
			default:
				break;
			}
		}
		break;
	default:
		break;
	}

	int iBlockSize = (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_AES)?16:8;

	if (mac_key)
		memcpy(pOutput, &TmpBuf[iSize-iBlockSize], iBlockSize);
	else
		memcpy(pOutput, TmpBuf, iSize);

	if (piResultSize)
		*piResultSize = iSize;
}

bool DFDecrypt(PVOID pReader, PBYTE pInput, PBYTE pOutput, int iSize, int iExpectedSize, int* piResultSize, bool f80Padding)
{
	bool bRes = true;

	if ((!pInput)||(!pOutput)||(iSize<=0))//||(m_DF_Settings[pReader].SessionKeySize == 0))
		return false;

	BYTE TmpInp[1024*16] = {0};
	BYTE TmpBuf[1024*16] = {0};

	BYTE IV[16] = {0};

	memcpy(IV, m_ReaderSettings[pReader].CurrentDFSettings()->IV, 16);

	int iMacSize = 0;
	USHORT crc16 = 0;

	ULONG crc32 = 0;
	CRC32 ccc32;

	BYTE ucZero = 0;

	memcpy(TmpInp, pInput, iSize);

	//memcpy(IV, m_DF_Settings[pReader].IV, 24);
	switch (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType)
	{
	case CLSCRF_DF_AUTH_TYPE_DES:
/*
		if (f80Padding&&(iSize%8))
		{
			TmpInp[iSize] = 0x80;
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K1;
		}
		else
		{
			if (calc_mac)
				mac_key = m_DF_Settings[pReader].K2;
		}
		*/
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			iSize -= 4;
			des_cbc_encrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				TmpInp,
				TmpBuf,
				iSize,
				IV );
			iMacSize = 8*(int)ceil((float)iSize/8);
			bRes = (memcmp(&TmpInp[iSize], &TmpBuf[iMacSize-8], 4) == 0);
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			des_cbc_decrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				TmpInp,
				TmpBuf,
				iSize,
				IV );

			if (iExpectedSize == 0)
			{
				for (iSize = iSize-1; iSize > 0; iSize--)
					if (TmpBuf[iSize] == 0x80)
						break;

				if (iSize <= 2)
					return false;

				iSize -= 2;
			}
			else
				iSize = iExpectedSize;


			ComputeCrc(CRC_A, (char*)TmpBuf, iSize, (PBYTE)&crc16, (PBYTE)&crc16+1);
			//crc16 = CRC_16(TmpBuf, iSize);
			bRes = (memcmp(&TmpBuf[iSize], (PVOID)&crc16, 2) == 0);
			break;
		default:
			break;
		}
		break;
	case CLSCRF_DF_AUTH_TYPE_3DES:
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			iSize -= 4;
			des_3cbc_encrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				TmpInp,
				TmpBuf,
				iSize,
				IV,
				IV);//,
				//m_DF_Settings[pReader].IV );
			iMacSize = 8*(int)ceil((float)iSize/8);
			bRes = (memcmp(&TmpInp[iSize], &TmpBuf[iMacSize-8], 4) == 0);
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			des_3cbc_decrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				TmpInp,
				TmpBuf,
				iSize,
				IV,
				IV);//,
				//m_DF_Settings[pReader].IV );
			if (iExpectedSize == 0)
			{
				for (iSize = iSize-1; iSize > 0; iSize--)
					if (TmpBuf[iSize] == 0x80)
						break;

				if (iSize <= 2)
					return false;

				iSize -= 2;
			}
			else
				iSize = iExpectedSize;


			ComputeCrc(CRC_A, (char*)TmpBuf, iSize, (PBYTE)&crc16, (PBYTE)&crc16+1);
			//crc16 = CRC_16(TmpBuf, iSize);
			bRes = (memcmp(&TmpBuf[iSize], (PVOID)&crc16, 2) == 0);
			break;
		default:
			break;
		}
		break;
	case CLSCRF_DF_AUTH_TYPE_3K3DES:
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			iSize = 8*(int)ceil((float)iSize/8);
			des_3k3cbc_decrypt(
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+8,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey+16,
				TmpInp,
				TmpBuf,
				iSize,
				IV,
				IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV );
			if (iExpectedSize == 0)
			{
				for (iSize = iSize-1; iSize > 0; iSize--)
					if (TmpBuf[iSize] == 0x80)
						break;

				if (iSize <= 2)
					return false;

				iSize -= 2;
			}
			else
				iSize = iExpectedSize;

			ccc32.Hash(TmpBuf, iSize);
			ccc32.Hash(&ucZero, 1);
			
			crc32 = ccc32.Get();
			bRes = (memcmp(&TmpBuf[iSize], (PVOID)&crc32, 4) == 0);
			break;
		default:
			break;
		}
		break;
	case CLSCRF_DF_AUTH_TYPE_AES:
		switch (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType)
		{
		case CLSCRF_DF_TRANSFER_TYPE_PLAIN:
		case CLSCRF_DF_TRANSFER_TYPE_MACCED:
			memcpy(TmpBuf, TmpInp, iSize);
			break;
		case CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED:
			iSize = 16*(int)ceil((float)iSize/16);
			aes_128_decrypt_cbc(
				TmpInp,
				iSize,
				m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
				IV,
				TmpBuf,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV );
			if (iExpectedSize == 0)
			{
				for (iSize = iSize-1; iSize > 0; iSize--)
					if (TmpBuf[iSize] == 0x80)
						break;

				if (iSize <= 2)
					return false;

				iSize -= 2;
			}
			else
				iSize = iExpectedSize;

			ccc32.Hash(TmpBuf, iSize);
			ccc32.Hash(&ucZero, 1);
			
			crc32 = ccc32.Get();
			bRes = (memcmp(&TmpBuf[iSize], (PVOID)&crc32, 4) == 0);
			break;
		default:
			break;
		}
		//int iSizeExpected = iSize;
		break;
	default:
		break;
	}

	memcpy(pOutput, TmpBuf, iSize);

	if (piResultSize)
		*piResultSize = iSize;

	return bRes;
}

void CryptoSubkeys(PVOID pReader)
{
	BYTE *key = m_ReaderSettings[pReader].CurrentCryptoSettings()->SessionKey;
	BYTE *K1 = m_ReaderSettings[pReader].CurrentCryptoSettings()->K1;
	BYTE *K2 = m_ReaderSettings[pReader].CurrentCryptoSettings()->K2;

	BYTE const_Rb[16] = {0};
	const_Rb[15] = 0x87;

	BYTE L[16] = {0};
	BYTE Z[16] = {0};
	BYTE tmp[16] = {0};

    //CTransfer* pTr = (CTransfer*)pReader;

	if (m_ReaderSettings[pReader].CurrentCryptoSettings()->fEncryptedMode)
		aes_128_encrypt_cbc( 
			Z,
			16,
			key,
			tmp,
			L );

    if ( (L[0] & 0x80) == 0 ) {
        leftshift_onebit(L,K1,16);
    } else {

        leftshift_onebit(L,tmp,16);
        xor_size(tmp,const_Rb,K1,16);
    }

    if ( (K1[0] & 0x80) == 0 ) 
    {
        leftshift_onebit(K1,K2,16);
    } else 
    {
        leftshift_onebit(K1,tmp,16);
        xor_size(tmp,const_Rb,K2,16);
    }
    
    return;
}


void CryptoEncrypt(PVOID pReader, 
				   PBYTE pInput, 
				   PBYTE pOutput, 
				   int iSize, 
				   int* piResultSize)
{
	if ((!pInput)||(!pOutput)||(iSize<=0))
		return;

	BYTE TmpInp[256] = {0};
	BYTE TmpBuf[256] = {0};
	
	BYTE IV[16] = {0};

	memcpy(IV, m_ReaderSettings[pReader].CurrentCryptoSettings()->IV, 16);

	ULONG crc32 = 0;
	CRC32 ccc32;

	memcpy(TmpInp, pInput, iSize);

    //CTransfer* pTr = (CTransfer*)pReader;

	if (m_ReaderSettings[pReader].CurrentCryptoSettings()->fEncryptedMode)
	{
		ccc32.Hash(TmpInp, iSize);
		crc32 = ccc32.Get();

		memcpy(&TmpInp[iSize], (PVOID)&crc32,4);

		iSize +=4;
		TmpInp[iSize++] = 0x80;
		iSize = 16*(int)ceil((float)iSize/16);

		aes_128_encrypt_cbc(
			TmpInp,
			iSize,
			m_ReaderSettings[pReader].CurrentCryptoSettings()->SessionKey,
			IV,
			TmpBuf,
			m_ReaderSettings[pReader].CurrentCryptoSettings()->IV
			);
	}
	else
		memcpy(TmpBuf, TmpInp, iSize);


	memcpy(pOutput, TmpBuf, iSize);

	if (piResultSize)
		*piResultSize = iSize;
}

bool CryptoDecrypt(
				   PVOID pReader, 
				   PBYTE pInput, 
				   PBYTE pOutput, 
				   int iSize, 
				   int iExpectedSize, 
				   int* piResultSize)
{
	bool bRes = true;

	if ((!pInput)||(!pOutput)||(iSize<=0))
		return false;

	BYTE TmpInp[256] = {0};
	BYTE TmpBuf[256] = {0};

	BYTE IV[16] = {0};

	memcpy(IV, m_ReaderSettings[pReader].CurrentCryptoSettings()->IV, 16);

	ULONG crc32 = 0;
	CRC32 ccc32;

	//BYTE ucZero = 0;

	memcpy(TmpInp, pInput, iSize);

    //CTransfer* pTr = (CTransfer*)pReader;

	if (m_ReaderSettings[pReader].CurrentCryptoSettings()->fEncryptedMode)
	{
		iSize = 16*(int)ceil((float)iSize/16);
		aes_128_decrypt_cbc(
			TmpInp,
			iSize,
			m_ReaderSettings[pReader].CurrentCryptoSettings()->SessionKey,
			IV,
			TmpBuf,
			m_ReaderSettings[pReader].CurrentCryptoSettings()->IV );
		//if (iExpectedSize == 0)
		//{
		for (iSize = iSize-1; iSize > 0; iSize--)
			if (TmpBuf[iSize] == 0x80)
				break;

		if (iSize <= 4)
			return false;

		iSize -= 4;
		//}
		//else
		//	iSize = iExpectedSize;

		ccc32.Hash(TmpBuf, iSize);
		//ccc32.Hash(&ucZero, 1);
		
		crc32 = ccc32.Get();
		bRes = (memcmp(&TmpBuf[iSize], (PVOID)&crc32, 4) == 0);
	}
	else
		memcpy(TmpBuf, TmpInp, iSize);

	memcpy(pOutput, TmpBuf, iSize);

	if (piResultSize)
		*piResultSize = iSize;

	return bRes;
}
