//==========================================================================//
//  SAM_Set.c : implementation file                                         //
//==========================================================================//


#include "StdAfx.h"


#include "Clscrfl_SAM.h"

#include "aes_128.h"
//#include "crc32.h"
#include <fstream>
#include <list>
#include <vector>
#include <map>
#include <math.h>
#include "Transfer_unix.h"

//#include <String.h>
//#include <MfErrNo.h>
//#include <SAM\SAM_AV2.h>
#include "SAM_Set.h"
#include "DESCrpt.h"
//#include "AES_128.h"
//#include <Crypto\CRC.h>

#define SAM_BUF_SIZE 256

//#define UEM_SAM_BYTECOUNTERR	SCARD_E_COMM_DATA_LOST
//#define UEM_SAM_APDU_ERR		SCARD_F_COMM_ERROR
//#define UEM_SAM_INVALID_CARD_MAC SCARD_E_INVALID_VALUE
//#define UEM_SAM_AUTHENTICATION_ERR 

BYTE SAM_Plain[SAM_BUF_SIZE];
BYTE SAM_Coded[SAM_BUF_SIZE];
BYTE SessionKeyEnc[16];
BYTE SessionKeyMAC[16];

BYTE KeyX[24];

//extern volatile DWORD SAM_Waiting_mks;


typedef struct _K64DES
{
	BYTE Key[8];
} K64DES;

typedef struct _K128AES
{
	BYTE Key[16];
} K128AES;

typedef struct _CRYPTO_SUITE
{
	DWORD f;
	union{
	DWORD   dw[8];
	K64DES  DesKey[4];
	K128AES AesKey[2];
	BYTE    b[32];
	} k;
} CRYPTO_SUITE;

char SAM_LockUnlock( LPVOID pReader, BYTE LogicalChannel, BYTE SubCmd, 
				 BYTE KeyNo, BYTE KeyVer, BYTE* KeyValue, 
				 BYTE UnlockKeyNo, BYTE UnlockKeyVer, 
				 DWORD MaxChainBlock,
				 BYTE* SW )
//	char SAM_LockUnlock( LPVOID pReader, BYTE LogicalChannel, BYTE SubCmd, BYTE KeyNo, BYTE KeyVer, BYTE* KeyValue, BYTE UnlockKeyNo, BYTE UnlockKeyVer, BYTE* SW )
{
	BYTE* ToSM;
	BYTE* From;
	int i, j;
	DWORD SendLen, RecvLen = SAM_BUF_SIZE;
	BYTE  SV1[16];
	BYTE  MAC16[16];
	BYTE  Kxe[16];
//    BYTE  Kx[16] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16 };
//    BYTE MaxChainBlock[3] = { 0x00, 0x00, 0x00 };
	char status = 0;

if( SubCmd > 3 )
	return (char)MI_WRONG_PARAMETER_VALUE;

	ToSM = SAM_Plain;
	From = SAM_Coded;
ToSM[0] = 0x80 | LogicalChannel;
ToSM[1] = 0x10;
ToSM[2] = SubCmd;   // P1
ToSM[3] = 0x00; // P2
SendLen = 5;
ToSM[SendLen++] = KeyNo;
ToSM[SendLen++] = KeyVer;
if( SubCmd == 0x02 )
{
	ToSM[SendLen++] = UnlockKeyNo;
	ToSM[SendLen++] = UnlockKeyVer;
}
else if( SubCmd == 0x03 )
{
	memcpy( &ToSM[7], &MaxChainBlock, 3 );
	SendLen = 10;
}
ToSM[4] = (BYTE)(SendLen - 5); // Lc
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );
/*
// Rnd2 from SAM                      // Debug
	RecvLen = 14;
	From[0]  = 0xC8;
	From[1]  = 0xE7;
	From[2]  = 0xF7;
	From[3]  = 0x07;
	From[4]  = 0x45;
	From[5]  = 0x5E;
	From[6]  = 0x19;
	From[7]  = 0x4F;
	From[8]  = 0x01;
	From[9]  = 0x5E;
	From[10] = 0x7F;
	From[11] = 0x35;
	From[12] = 0x90;
	From[13] = 0xAF;
*/
	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;

	memcpy( &SV1[ 5], &From[7], 5 );
	memcpy( &SV1[10], &From[0], 5 );

	From[12] = SubCmd;
if( SubCmd == 0x03 )
	memcpy( &From[13], &MaxChainBlock, 3 );
else if( SubCmd == 0x02 )
{
	From[13] = UnlockKeyNo;
	From[14] = UnlockKeyVer;
	From[15] = 0;
}
else
{
	From[13] = 0;
	From[14] = 0;
	From[15] = 0;
}

AES_128_CBC_En_MAC( From, 16, KeyValue, MAC16 );

	ToSM = &SAM_Plain[5];
	for( i = 0, j = 1; i < 8; i++, j+=2 )
		ToSM[i] = MAC16[j];

// Rnd1 p.6 to SAM
	srand(time(NULL));
	ToSM[8] = rand() & 0xFF;// 0x00;
	ToSM[9] = rand() & 0xFF;// 0x01;
	ToSM[10] = rand() & 0xFF;// 0x02;
	ToSM[11] = rand() & 0xFF;// 0x03;
	ToSM[12] = rand() & 0xFF;// 0x04;
	ToSM[13] = rand() & 0xFF;// 0x05;
	ToSM[14] = rand() & 0xFF;// 0x06;
	ToSM[15] = rand() & 0xFF;// 0x07;
	ToSM[16] = rand() & 0xFF;// 0x08;
	ToSM[17] = rand() & 0xFF;// 0x09;
	ToSM[18] = rand() & 0xFF;// 0x0A;
	ToSM[19] = rand() & 0xFF;// 0x0B;

	memcpy( &SV1[0], &ToSM[15], 5 );
	for( i = 0; i < 5; i++ )
		SV1[10+i] ^= ToSM[8+i];
	SV1[15] = 0x91;

	aes_128_encrypt16( SV1, KeyValue, Kxe );

// p.7
	ToSM = SAM_Plain;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0x10;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x14; // Lc
	SendLen = 5 + 20;
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;

	ToSM = &SAM_Plain[13];
	ToSM[12] = SubCmd;
if( SubCmd == 0x03 )
	memcpy( &ToSM[13], &MaxChainBlock, 3 );
else if( SubCmd == 0x02 )
{
	ToSM[13] = UnlockKeyNo;
	ToSM[14] = UnlockKeyVer;
	ToSM[15] = 0;
}
else
{
	ToSM[13] = 0;
	ToSM[14] = 0;
	ToSM[15] = 0;
}

	AES_128_CBC_En_MAC( ToSM, 16, KeyValue, MAC16 );

	for( i = 0, j = 1; i < 8; i++, j+=2 )
		ToSM[i] = MAC16[j];

	if( memcmp( ToSM, From, 8 ) )
		return (char)UEM_SAM_INVALID_CARD_MAC;
/*
// encrypted RndB from SAM
	From[8]  = 0xE8;
	From[9]  = 0x0D;
	From[10] = 0x04;
	From[11] = 0xF5;
	From[12] = 0x44;
	From[13] = 0xBE;
	From[14] = 0x8D;
	From[15] = 0x3B;
	From[16] = 0x43;
	From[17] = 0x56;
	From[18] = 0xB9;
	From[19] = 0xBF;
	From[20] = 0x9C;
	From[21] = 0x2F;
	From[22] = 0xA0;
	From[23] = 0x31;
*/
	aes_128_decrypt16( &From[8], Kxe, &From[24] );
	From[40] = From[24];
	From[41] = From[25];
	From = &SAM_Coded[26];

// RndA p.14 to SAM
	From = &SAM_Coded[10];
	From[0] = rand() & 0xFF;// 0x00;
	From[1] = rand() & 0xFF;// 0x01;
	From[2] = rand() & 0xFF;// 0x02;
	From[3] = rand() & 0xFF;// 0x03;
	From[4] = rand() & 0xFF;// 0x04;
	From[5] = rand() & 0xFF;// 0x05;
	From[6] = rand() & 0xFF;// 0x06;
	From[7] = rand() & 0xFF;// 0x07;
	From[8] = rand() & 0xFF;// 0x08;
	From[9] = rand() & 0xFF;// 0x09;
	From[10] = rand() & 0xFF;// 0x0A;
	From[11] = rand() & 0xFF;// 0x0B;
	From[12] = rand() & 0xFF;// 0x0C;
	From[13] = rand() & 0xFF;// 0x0D;
	From[14] = rand() & 0xFF;// 0x0E;
	From[15] = rand() & 0xFF;// 0x0F;

	for( i = 0; i < 16; i++ )
		SV1[i] = 0;
	ToSM = &SAM_Plain[5];
	aes_128_encrypt_cbc( 
		From, 
		32, 
		Kxe, 
		SV1, 
		ToSM );

	ToSM = &SAM_Plain[0];
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0x10;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x20; // Lc
	SendLen = 5 + 32;
	ToSM[SendLen++] = 0x00;    // Le
	From = &SAM_Coded[28];

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );
	if ( status == 0x01)
		return SCARD_S_SUCCESS;


	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0x00) )
		return (char)UEM_SAM_APDU_ERR;
/*
// encrypted RndA'' p.20 from SAM   Debug
	From[0]   = 0xDB;
	From[1]   = 0x60;
	From[2]   = 0x15;
	From[3]   = 0x2D;
	From[4]   = 0x02;
	From[5]   = 0xF5;
	From[6]   = 0x80;
	From[7]   = 0xE1;
	From[8]   = 0xDD;
	From[9]   = 0x2E;
	From[10]  = 0x9B;
	From[11]  = 0xFE;
	From[12]  = 0x39;
	From[13]  = 0xD1;
	From[14]  = 0x8C;
	From[15]  = 0x1A;
*/
	aes_128_decrypt16( From, Kxe, MAC16 );

	From = &SAM_Coded[10];
	From[16] = From[0];
	From[17] = From[1];

	if( memcmp( &From[2], MAC16, 16 ) )
		return (char)UEM_SAM_AUTHENTICATION_ERR;

	return status;
}//SAM_LockUnlock

char SAM_AV1_AuthenticateHost( LPVOID pReader, BYTE LogicalChannel, BYTE* SW )
{
	CRYPTO_SUITE m_CryptoSuite;
	int i;
	DWORD SendLen, RecvLen = SAM_BUF_SIZE;
//    volatile 
	BYTE* ToSM;
	BYTE* From;
	BYTE RndA[16];
	BYTE RndB[16];
	BYTE AuthMode = 0x00; // 
	BYTE KeyNo = 0;
	BYTE KeyVer = 0;
	char status = 0;
// 0 - Reset
// 1 - First
// 2 - SL2
// else - Following

	ToSM = SAM_Plain;
	From = SAM_Coded;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xa4;
	ToSM[2] = AuthMode; // P1
	ToSM[3] = 0x00;     // P2
	ToSM[5] = KeyNo;
	ToSM[6] = KeyVer;
	SendLen = 7;
	if( AuthMode == 0 )
	{
		ToSM[4] = 0x02; // Lc
	}
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;

//    printf("DES_KeySetting\n");

	m_CryptoSuite.k.dw[0] = 0;
	m_CryptoSuite.k.dw[1] = 0;
	m_CryptoSuite.k.dw[6] = 0;  // &m_CryptoSuite.k.b[24]
	m_CryptoSuite.k.dw[7] = 0;

	DES_KeySetting( &m_CryptoSuite.k.b[0] );
	memcpy( &ToSM[12], From, 8 );

	DES_CBC_De( &ToSM[12], 1, &m_CryptoSuite.k.b[24] );
// RndB
	memcpy( RndB, &ToSM[12], 8 );

	ToSM[20] = ToSM[12];

	srand(time(NULL));
	RndA[0] = rand() & 0xFF;// 0xf4;
	RndA[1] = rand() & 0xFF;// 0x4b;
	RndA[2] = rand() & 0xFF;// 0x0f;
	RndA[3] = rand() & 0xFF;// 0x1a;
	RndA[4] = rand() & 0xFF;// 0xe4;
	RndA[5] = rand() & 0xFF;// 0xf3;
	RndA[6] = rand() & 0xFF;// 0xf0;
	RndA[7] = rand() & 0xFF;// 0x2b;

	memcpy( &ToSM[5], RndA, 8 );

	DES_CBC_En( &ToSM[5], 2, &m_CryptoSuite.k.b[24] );

	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xa4;
	ToSM[2] = 0x00;         // P1
	ToSM[3] = 0x00;         // P2
	ToSM[4] = 16;           // Lc
	SendLen = 5 + 16;
	ToSM[SendLen++] = 0x00; // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								&From[1],
								&RecvLen );

	memcpy( SW, &From[RecvLen-1], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-1] != 0x90) || (From[RecvLen] != 0x00) )
		return (char)UEM_SAM_APDU_ERR;

	DES_CBC_De( &From[1], 1, &m_CryptoSuite.k.b[24] );
	From[0] = From[8];

	for( i = 0; i < 4; i++ )
	{
		SessionKeyEnc[i+ 0] = RndA[i];
		SessionKeyEnc[i+ 4] = RndB[i];
		SessionKeyEnc[i+ 8] = RndA[i];
		SessionKeyEnc[i+12] = RndB[i];
	}

	return status;
}//SAM_AV1_AuthenticateHost

char SAM_AV1_ChangeKeyEntry( LPVOID pReader,
							 BYTE LogicalChannel, BYTE  KeyNo, BYTE ProMas,
							 BYTE* NewPosAkey, BYTE  NewPosAkeyVer,
							 BYTE* NewPosBkey, BYTE  NewPosBkeyVer,
							 BYTE* NewPosCkey, BYTE  NewPosCkeyVer, BYTE* SW )
{
	CRYPTO_SUITE m_CryptoSuite;
	DWORD SendLen, RecvLen = SAM_BUF_SIZE;
	WORD i, nw, CRC;
	BYTE* ToSM;
	BYTE* From;
//BYTE SKey[16] = { 0xF4,0x4B,0x0F,0x1A,0xBC,0x1B,0x53,0x5F,0xF4,0x4B,0x0F,0x1A,0xBC,0x1B,0x53,0x5F };// Debug const check

	char status = 0;

	ToSM = &SAM_Plain[5];
	From = SAM_Coded;

	memcpy( &ToSM[00], NewPosAkey, 16 );
	memcpy( &ToSM[16], NewPosBkey, 16 );
	memcpy( &ToSM[32], NewPosCkey, 16 );
	SendLen = 48;
	ToSM[SendLen++] = 0x00;    // DF_AID 0
	ToSM[SendLen++] = 0x00;    // DF_AID 1
	ToSM[SendLen++] = 0x00;    // DF_AID 2
	ToSM[SendLen++] = 0x00;    // DF_KeyNo
	ToSM[SendLen++] = 0x00;    // KeyNoCEK
	ToSM[SendLen++] = 0x00;    // KeyVCEK
	ToSM[SendLen++] = 0xFF;    // RefNoKUC
	ToSM[SendLen++] = 0x20;    // SET LSB
	ToSM[SendLen++] = 0x00;    // SET MSB
	ToSM[SendLen++] = NewPosAkeyVer;    // Version of key PosA
	ToSM[SendLen++] = NewPosBkeyVer;    // Version of key PosB
	ToSM[SendLen++] = NewPosCkeyVer;    // Version of key PosC

	CRC = 0x6363;
	for( i = 0; i < SendLen; i++ )
		CTransfer::CalculateFCS( CRC, ToSM[i] );
//    printf("CRC = 0x%04X\n",CRC);

	memcpy( &ToSM[SendLen], &CRC, 2 );
	SendLen += 2;
	nw = (WORD)((SendLen + 7) / 8 * 8);
	for( i = (WORD)SendLen; i < nw; i++ )
	{
		ToSM[i] = 0;
	}
	SendLen = nw;

	m_CryptoSuite.k.dw[6] = 0;
	m_CryptoSuite.k.dw[7] = 0;
	memcpy( &m_CryptoSuite.k.b[0], SessionKeyEnc, 8 );
//    memcpy( &m_CryptoSuite.k.b[0], SKey, 8 );   // Debug const check
	DES_KeySetting( &m_CryptoSuite.k.b[0] );
	DES_CBC_En( ToSM, (SendLen>>3), &m_CryptoSuite.k.b[24] );

	ToSM = SAM_Plain;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xC1;
	ToSM[2] = KeyNo;        // P1
	ToSM[3] = ProMas;       // P2
	ToSM[4] = (BYTE)SendLen;      // Lc
	SendLen += 5;

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0x00) )
		return (char)UEM_SAM_APDU_ERR;

	return status;
}//SAM_AV1_ChangeKeyEntry

//char SAM_LockUnlock( BYTE LogicalChannel, BYTE KeyNo, BYTE KeyVer, BYTE* KeyValue, BYTE* SW )
char SAM_LockUnlock_Old( LPVOID pReader, BYTE LogicalChannel, BYTE SubCmd, BYTE KeyNo, BYTE KeyVer, BYTE* KeyValue, BYTE UnlockKeyNo, BYTE UnlockKeyVer, BYTE* SW )
{
	BYTE* ToSM;
	BYTE* From;
	int i, j;
	DWORD SendLen, RecvLen = SAM_BUF_SIZE;
	BYTE  SV1[16];
	BYTE  MAC16[16];
	BYTE  Kxe[16];
//    BYTE  Kx[16] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16 };
	BYTE MaxChainBlock[3] = { 0x00, 0x00, 0x00 };
	char status = 0;

	ToSM = SAM_Plain;
	From = SAM_Coded;
	SendLen = 0;
	ToSM[SendLen++] = 0x80 | LogicalChannel;
	ToSM[SendLen++] = 0x10;
	ToSM[SendLen++] = SubCmd;   // P1
	ToSM[SendLen++] = 0x00; // P2
	ToSM[SendLen++] = 0x00; // Lc
	ToSM[SendLen++] = KeyNo;
	ToSM[SendLen++] = KeyVer;
	if (SubCmd == 0x03)
	{
		memcpy( &ToSM[SendLen], MaxChainBlock, 3 ); SendLen += 3;
	}
	if (SubCmd == 0x02)
	{
		ToSM[SendLen++] = UnlockKeyNo;
		ToSM[SendLen++] = UnlockKeyVer;
	}
	ToSM[4] = (BYTE)(SendLen - 5);
	//SendLen = 10;
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );
/*
// Rnd2 from SAM                      // Debug
	RecvLen = 14;
	From[0]  = 0xC8;
	From[1]  = 0xE7;
	From[2]  = 0xF7;
	From[3]  = 0x07;
	From[4]  = 0x45;
	From[5]  = 0x5E;
	From[6]  = 0x19;
	From[7]  = 0x4F;
	From[8]  = 0x01;
	From[9]  = 0x5E;
	From[10] = 0x7F;
	From[11] = 0x35;
	From[12] = 0x90;
	From[13] = 0xAF;
*/
	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;

	memcpy( &SV1[ 5], &From[7], 5 );
	memcpy( &SV1[10], &From[0], 5 );

	From[12] = SubCmd;
	if (SubCmd == 0x03)
	{
		memcpy( &From[13], &ToSM[7], 3 );
		AES_128_CBC_En_MAC( From, 16, KeyValue, MAC16 );
	}
	else if (SubCmd == 0x02)
	{
		memcpy( &From[13], &ToSM[7], 2 );
		AES_128_CBC_En_MAC( From, 15, KeyValue, MAC16 );
	}
	else
	{
		AES_128_CBC_En_MAC( From, 13, KeyValue, MAC16 );
	}


	ToSM = &SAM_Plain[5];
	for( i = 0, j = 1; i < 8; i++, j+=2 )
		ToSM[i] = MAC16[j];

// Rnd1 p.6 to SAM
	srand(time(NULL));
	ToSM[8] = rand() & 0xFF;// 0x00;
	ToSM[9] = rand() & 0xFF;// 0x01;
	ToSM[10] = rand() & 0xFF;// 0x02;
	ToSM[11] = rand() & 0xFF;// 0x03;
	ToSM[12] = rand() & 0xFF;// 0x04;
	ToSM[13] = rand() & 0xFF;// 0x05;
	ToSM[14] = rand() & 0xFF;// 0x06;
	ToSM[15] = rand() & 0xFF;// 0x07;
	ToSM[16] = rand() & 0xFF;// 0x08;
	ToSM[17] = rand() & 0xFF;// 0x09;
	ToSM[18] = rand() & 0xFF;// 0x0A;
	ToSM[19] = rand() & 0xFF;// 0x0B;

	memcpy( &SV1[0], &ToSM[15], 5 );
	for( i = 0; i < 5; i++ )
		SV1[10+i] ^= ToSM[8+i];
	SV1[15] = 0x91;

	aes_128_encrypt16( SV1, KeyValue, Kxe );

// p.7
	ToSM = SAM_Plain;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0x10;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x14; // Lc
	SendLen = 5 + 20;
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;

	ToSM = &SAM_Plain[13];
	ToSM[12] = SubCmd;
	memcpy( &ToSM[13], MaxChainBlock, 3 );

	AES_128_CBC_En_MAC( ToSM, 16, KeyValue, MAC16 );

	for( i = 0, j = 1; i < 8; i++, j+=2 )
		ToSM[i] = MAC16[j];

	if( memcmp( ToSM, From, 8 ) )
		return (char)UEM_SAM_INVALID_CARD_MAC;
/*
// encrypted RndB from SAM
	From[8]  = 0xE8;
	From[9]  = 0x0D;
	From[10] = 0x04;
	From[11] = 0xF5;
	From[12] = 0x44;
	From[13] = 0xBE;
	From[14] = 0x8D;
	From[15] = 0x3B;
	From[16] = 0x43;
	From[17] = 0x56;
	From[18] = 0xB9;
	From[19] = 0xBF;
	From[20] = 0x9C;
	From[21] = 0x2F;
	From[22] = 0xA0;
	From[23] = 0x31;
*/
	aes_128_decrypt16( &From[8], Kxe, &From[24] );
	From[40] = From[24];
	From[41] = From[25];
	From = &SAM_Coded[26];

// RndA p.14 to SAM
	From = &SAM_Coded[10];
	From[0] = rand() & 0xFF;// 0x00;
	From[1] = rand() & 0xFF;// 0x01;
	From[2] = rand() & 0xFF;// 0x02;
	From[3] = rand() & 0xFF;// 0x03;
	From[4] = rand() & 0xFF;// 0x04;
	From[5] = rand() & 0xFF;// 0x05;
	From[6] = rand() & 0xFF;// 0x06;
	From[7] = rand() & 0xFF;// 0x07;
	From[8] = rand() & 0xFF;// 0x08;
	From[9] = rand() & 0xFF;// 0x09;
	From[10] = rand() & 0xFF;// 0x0A;
	From[11] = rand() & 0xFF;// 0x0B;
	From[12] = rand() & 0xFF;// 0x0C;
	From[13] = rand() & 0xFF;// 0x0D;
	From[14] = rand() & 0xFF;// 0x0E;
	From[15] = rand() & 0xFF;// 0x0F;

	for( i = 0; i < 16; i++ )
		SV1[i] = 0;
	ToSM = &SAM_Plain[5];
	aes_128_encrypt_cbc( 
		From, 
		32, 
		Kxe, 
		SV1, 
		ToSM );

	ToSM = &SAM_Plain[0];
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0x10;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x20; // Lc
	SendLen = 5 + 32;
	ToSM[SendLen++] = 0x00;    // Le
	From = &SAM_Coded[28];

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0x00) )
		return (char)UEM_SAM_APDU_ERR;
/*
// encrypted RndA'' p.20 from SAM   Debug
	From[0]   = 0xDB;
	From[1]   = 0x60;
	From[2]   = 0x15;
	From[3]   = 0x2D;
	From[4]   = 0x02;
	From[5]   = 0xF5;
	From[6]   = 0x80;
	From[7]   = 0xE1;
	From[8]   = 0xDD;
	From[9]   = 0x2E;
	From[10]  = 0x9B;
	From[11]  = 0xFE;
	From[12]  = 0x39;
	From[13]  = 0xD1;
	From[14]  = 0x8C;
	From[15]  = 0x1A;
*/
	aes_128_decrypt16( From, Kxe, MAC16 );

	From = &SAM_Coded[10];
	From[16] = From[0];
	From[17] = From[1];

	if( memcmp( &From[2], MAC16, 16 ) )
		return (char)UEM_SAM_AUTHENTICATION_ERR;

	return status;
}//SAM_LockUnlock

char SAM_SwitchToAV2Mode( LPVOID pReader, 
						 BYTE LogicalChannel, 
						 BYTE* NewPosAkey, BYTE  NewPosAkeyVer,
						 BYTE* NewPosBkey, BYTE  NewPosBkeyVer,
						 BYTE* NewPosCkey, BYTE  NewPosCkeyVer, DWORD dwMaxChainBlocks, BYTE* SW )
{
	char status = 0;

//printf("SAM_AV1_AuthenticateHost\n");
	status = SAM_AV1_AuthenticateHost( pReader, 0, SW );
//printf("SAM_AV1_AuthenticateHost status = %d\n",status);
	if( status )
		return status;

//printf("SAM_AV1_ChangeKeyEntry\n");
	status = SAM_AV1_ChangeKeyEntry( pReader, LogicalChannel, 0, 0xff,
									 NewPosAkey, NewPosAkeyVer,
									 NewPosBkey, NewPosBkeyVer,
									 NewPosCkey, NewPosCkeyVer, SW );
//printf("SAM_AV1_ChangeKeyEntry status = %d\n",status);
	if( status )
		return status;

//	SAM_Waiting_mks = 5000000;
//printf("SAM_LockUnlock\n",status);
	status = SAM_LockUnlock( pReader, LogicalChannel, 0x03, 0, NewPosAkeyVer, NewPosAkey, 0, 0, dwMaxChainBlocks, SW );
//printf("SAM_LockUnlock status = %d\n",status);
//	SAM_Waiting_mks = 50000;

	return status;
}//SAM_SwitchToAV2Mode

char SAM_AuthenticateHost( LPVOID pReader,
						   BYTE LogicalChannel, 
						   BYTE KeyKind,
						   BYTE* Kx, 
						   BYTE KeyNo, 
						   BYTE KeyVer, 
						   BYTE HostMode, 
						   BYTE* SW ) 
{
	int i, j;
	DWORD SendLen, RecvLen = SAM_BUF_SIZE;
//    volatile 
	BYTE* ToSM;
	BYTE* From;
	BYTE  SV1[16];
	BYTE  SV2[16];
	BYTE  MAC16[16];
	BYTE  MAC24[24];
	BYTE  Kxe[24];
	char status = 0;

	//BYTE MyKx[16] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F};
	//BYTE MyKeyNo = 0x05;
	//BYTE MyKeyVer = 0x01;
	//BYTE MyHostMode = 0x02;
	//BYTE MyRND2[12] = {0x25, 0x09, 0xC7, 0xB0, 0x9F, 0x2D, 0xA8, 0xFF, 0x6D, 0x76, 0x57, 0x8B};

	ToSM = SAM_Plain;
	From = SAM_Coded;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xa4;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x03; // Lc
	ToSM[5] = KeyNo;
	//ToSM[5] = MyKeyNo;
	ToSM[6] = KeyVer;
	//ToSM[6] = MyKeyVer;
	ToSM[7] = HostMode;
	//ToSM[7] = MyHostMode;
	SendLen = 8;
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;
	//memcpy(From, MyRND2, 12);

/*
// Debug Rnd2 from SAM
	From[ 0] = 0xCC;
	From[ 1] = 0x3E;
	From[ 2] = 0xE4;
	From[ 3] = 0x98;
	From[ 4] = 0x8E;
	From[ 5] = 0x60;
	From[ 6] = 0x2A;
	From[ 7] = 0xC1;
	From[ 8] = 0x3A;
	From[ 9] = 0x50;
	From[10] = 0x06;
	From[11] = 0x65;
*/
	memcpy( &SV1[ 5], &From[7], 5 );
	memcpy( &SV1[10], &From[0], 5 );

	From[12] = HostMode;
	//From[12] = MyHostMode;
	From[13] = 0;
	From[14] = 0;
	From[15] = 0;

	switch(KeyKind)
	{
	case 3:
	case 5:
		AES_192_CBC_En_MAC( From, 24, Kx, MAC24 );
		break;
	default:
		AES_128_CBC_En_MAC( From, 16, Kx, MAC16 );
		//AES_128_CBC_En_MAC( From, 16, MyKx, MAC16 );
	}


	ToSM = &SAM_Plain[5];

	switch(KeyKind)
	{
	case 3:
	case 5:
		for( i = 0, j = 1; i < 12; i++, j+=2 )
			ToSM[i] = MAC24[j];
		break;
	default:
		for( i = 0, j = 1; i < 8; i++, j+=2 )
			ToSM[i] = MAC16[j];
	}


// Rnd1 p.6 to SAM
	srand(time(NULL));
	ToSM[8] = rand() & 0xFF;// 0x00;
	ToSM[9] = rand() & 0xFF;// 0x01;
	ToSM[10] = rand() & 0xFF;// 0x02;
	ToSM[11] = rand() & 0xFF;// 0x03;
	ToSM[12] = rand() & 0xFF;// 0x04;
	ToSM[13] = rand() & 0xFF;// 0x05;
	ToSM[14] = rand() & 0xFF;// 0x06;
	ToSM[15] = rand() & 0xFF;// 0x07;
	ToSM[16] = rand() & 0xFF;// 0x08;
	ToSM[17] = rand() & 0xFF;// 0x09;
	ToSM[18] = rand() & 0xFF;// 0x0A;
	ToSM[19] = rand() & 0xFF;// 0x0B;

	memcpy( &SV1[0], &ToSM[15], 5 );
	for( i = 0; i < 5; i++ )
		SV1[10+i] ^= ToSM[8+i];
	SV1[15] = 0x91;

	switch(KeyKind)
	{
	case 3:
	case 5:
		aes_192_encrypt24( SV1, Kx, Kxe );
		break;
	default:
		aes_128_encrypt16( SV1, Kx, Kxe );
		//aes_128_encrypt16( SV1, MyKx, Kxe );
	}

// p.7
	ToSM = SAM_Plain;
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xa4;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x14; // Lc
	SendLen = 5 + 20;
	ToSM[SendLen++] = 0x00;    // Le

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0xAF) )
		return (char)UEM_SAM_APDU_ERR;
/*
// Debug Response-2 from SAM
	RecvLen = 26;
	From[ 0] = 0x21;
	From[ 1] = 0xC4;
	From[ 2] = 0xE0;
	From[ 3] = 0x6C;
	From[ 4] = 0x91;
	From[ 5] = 0xF8;
	From[ 6] = 0x7F;
	From[ 7] = 0x86;
	From[ 8] = 0x32;
	From[ 9] = 0x5A;
	From[10] = 0x37;
	From[11] = 0xF6;
	From[12] = 0x50;
	From[13] = 0x6E;
	From[14] = 0xC6;
	From[15] = 0x55;
	From[16] = 0x77;
	From[17] = 0x12;
	From[18] = 0x4E;
	From[19] = 0xAF;
	From[20] = 0xC5;
	From[21] = 0x99;
	From[22] = 0xA4;
	From[23] = 0xB4;
	From[24] = 0x90;
	From[25] = 0xAF;
*/

	ToSM = &SAM_Plain[13];
	ToSM[12] = HostMode;
	ToSM[13] = 0;
	ToSM[14] = 0;
	ToSM[15] = 0;

	switch(KeyKind)
	{
	case 3:
	case 5:
		AES_192_CBC_En_MAC( ToSM, 24, Kx, MAC24 );
		break;
	default:
		AES_128_CBC_En_MAC( ToSM, 16, Kx, MAC16 );

		for( i = 0, j = 1; i < 8; i++, j+=2 )
			ToSM[i] = MAC16[j];
	}


	if( memcmp( ToSM, From, 8 ) )
		return (char)UEM_SAM_INVALID_CARD_MAC;

	switch(KeyKind)
	{
	case 3:
	case 5:
		aes_192_decrypt24( &From[16], Kxe, &From[24] );
		break;
	default:
		aes_128_decrypt16( &From[8], Kxe, &From[24] );
	}


// Load RndB for Ke & Km
	memcpy( &SV1[ 5], &From[24], 5 );
	memcpy( &SV1[10], &From[31], 5 );
	memcpy( &SV2[ 5], &From[28], 5 );
	memcpy( &SV2[10], &From[35], 5 );

// RndB''
	From[40] = From[24];
	From[41] = From[25];
	From = &SAM_Coded[26];

// RndA p.14 to SAM
	From = &SAM_Coded[10];
	From[0] = rand() & 0xFF;// 0x00;
	From[1] = rand() & 0xFF;// 0x01;
	From[2] = rand() & 0xFF;// 0x02;
	From[3] = rand() & 0xFF;// 0x03;
	From[4] = rand() & 0xFF;// 0x04;
	From[5] = rand() & 0xFF;// 0x05;
	From[6] = rand() & 0xFF;// 0x06;
	From[7] = rand() & 0xFF;// 0x07;
	From[8] = rand() & 0xFF;// 0x08;
	From[9] = rand() & 0xFF;// 0x09;
	From[10] = rand() & 0xFF;// 0x0A;
	From[11] = rand() & 0xFF;// 0x0B;
	From[12] = rand() & 0xFF;// 0x0C;
	From[13] = rand() & 0xFF;// 0x0D;
	From[14] = rand() & 0xFF;// 0x0E;
	From[15] = rand() & 0xFF;// 0x0F;
// Load RndA for Ke & Km
	memcpy( SV1, &From[0], 5 );
	for( i = 0; i < 5; i++ )
		SV1[10+i] ^= From[7+i];
	SV1[15] = 0x81;
	memcpy( SV2, &From[4], 5 );
	for( i = 0; i < 5; i++ )
		SV2[10+i] ^= From[11+i];
	SV2[15] = 0x82;

	switch(KeyKind)
	{
	case 3:
	case 5:
		for( i = 0; i < 24; i++ )
			MAC24[i] = 0;
		ToSM = &SAM_Plain[5];
		aes_192_encrypt_cbc( 
			From, 
			32, 
			Kxe, 
			MAC24, 
			ToSM );
		break;
	default:
		for( i = 0; i < 16; i++ )
			MAC16[i] = 0;
		ToSM = &SAM_Plain[5];
		aes_128_encrypt_cbc( 
			From, 
			32, 
			Kxe, 
			MAC16, 
			ToSM );
	}


	ToSM = &SAM_Plain[0];
	ToSM[0] = 0x80 | LogicalChannel;
	ToSM[1] = 0xa4;
	ToSM[2] = 0x00; // P1
	ToSM[3] = 0x00; // P2
	ToSM[4] = 0x20; // Lc
	SendLen = 5 + 32;
	ToSM[SendLen++] = 0x00;    // Le
	From = &SAM_Coded[28];

	RecvLen = SAM_BUF_SIZE;
	status = (char)CLSCRF_SAM_APDU(	pReader,
								ToSM,
								SendLen,
								From,
								&RecvLen );

	memcpy( SW, &From[RecvLen-2], 2 );
	if( status )
		return status;
	if( RecvLen < 2 )
		return (char)UEM_SAM_BYTECOUNTERR;
	if( (From[RecvLen-2] != 0x90) || (From[RecvLen-1] != 0x00) )
		return (char)UEM_SAM_APDU_ERR;
/*
// encrypted RndA'' p.20 from SAM   Debug
	From[0]   = 0x3C;
	From[1]   = 0xB3;
	From[2]   = 0x66;
	From[3]   = 0x28;
	From[4]   = 0x17;
	From[5]   = 0xD8;
	From[6]   = 0xCE;
	From[7]   = 0xF6;
	From[8]   = 0x3A;
	From[9]   = 0xC1;
	From[10]  = 0x63;
	From[11]  = 0xBD;
	From[12]  = 0x6A;
	From[13]  = 0xBB;
	From[14]  = 0x0F;
	From[15]  = 0x6F;
*/
	switch(KeyKind)
	{
	case 3:
	case 5:
		aes_192_decrypt24( From, Kxe, MAC24 );
		break;
	default:
		aes_128_decrypt16( From, Kxe, MAC16 );
	}

	From = &SAM_Coded[10];
	From[16] = From[0];
	From[17] = From[1];

	switch(KeyKind)
	{
	case 3:
	case 5:
		if( memcmp( &From[2], MAC24, 24 ) )
			return (char)UEM_SAM_AUTHENTICATION_ERR;
		break;
	default:
		if( memcmp( &From[2], MAC16, 16 ) )
			return (char)UEM_SAM_AUTHENTICATION_ERR;
	}

	//if( HostMode == 0 )
	//	return status;

	switch(KeyKind)
	{
	case 3:
	case 5:
		aes_192_encrypt24( SV1, Kx, SessionKeyEnc );
		aes_192_encrypt24( SV2, Kx, SessionKeyMAC );
		memcpy(KeyX, Kx, 24);
		break;
	default:
		aes_128_encrypt16( SV1, Kx, SessionKeyEnc );
		aes_128_encrypt16( SV2, Kx, SessionKeyMAC );
		memcpy(KeyX, Kx, 16);
	}


	return status;
}//SAM_AuthenticateHost
