// pcscndef.h
#ifndef pcscndeff
#define pcscndeff

#ifdef CLSCRF_EXPORTS
#define CLSCRF_API __declspec(dllexport)
#else
#define CLSCRF_API __declspec(dllimport)
#endif

//Specific error code
#define TAG_NOT_SUPPORTED               ((DWORD)0x601FFF05L)
#define TAG_WRONG_CONFIG                ((DWORD)0x601FFF06L)
#define TAG_NOT_DEFAULT                  ((DWORD)0x601FFF0AL)
#define TAG_MEMORY_ERROR				((DWORD)0x601FFF0BL)
#define SCARD_EXPECT_RECVLEN_MISMATCH   ((DWORD)0x601FFF0CL)
#define INVALID_NDEF_STATE              ((DWORD)0x601FFF0DL)
#define NDEF_LENGTH_MISMATCH            ((DWORD)0x601FFF0EL)
//return code from PC/SC or card in case not successful 0x6FF0SW1SW2


typedef enum _NDEF_CARD_TYPE
{
/* NOT SUPPORTED: means card is not NFC Forum supported tag by the library */
NDEF_CARD_NOT_SUPPORTED = 0,
/* TYPE 1: means card is NFC Forum Type 1 tag */
NDEF_CARD_TYPE_1 = 1,
/* TYPE 2: means card is NFC Forum Type 2 tag */
NDEF_CARD_TYPE_2 = 2,
/* TYPE 2: means card is NFC Forum Type 3 tag */
NDEF_CARD_TYPE_3 = 3,
/* TYPE 4: means card is NFC Forum Type 4 tag */
NDEF_CARD_TYPE_4 = 4,
/* MIFARE: means card is NXP MIFARE standard (1k/4k) tag supporting NFC Forum NDEF specification */
NDEF_CARD_TYPE_MIFARE = 5,
/* ICODE: means card is NXP ICODE SLI tag supporting NFC Forum NDEF specification */
NDEF_CARD_TYPE_ICODE = 6
} NDEF_CARD_TYPE;

typedef enum _NDEF_CARD_STATE
{
/* INVALID: means card is not NFC Forum formatted tag, formatting can only be performed for the supported cards */
NDEF_CARD_INVALID = 0,
/* INITIALISED: means card doesnt contain any NDEF data */
NDEF_CARD_INITIALISED = 1,
/* READ WRITE: means user can use read and write operations */
NDEF_CARD_READ_WRITE = 2,
/* READ ONLY: means user can only use read operation */
NDEF_CARD_READ_ONLY = 3
} NDEF_CARD_STATE;

typedef struct _NDEF_CHK_NDEF_INFO
{
NDEF_CARD_TYPE NdefCardType;
NDEF_CARD_STATE NdefCardState;
DWORD Ndef_ActualMsgLength;
DWORD Ndef_MaxMsgLength;
} NDEF_CHK_NDEF_INFO;

//Public function
CLSCRF_API LONG NDEF_CheckNdef(__in SCARDHANDLE hCard, __out NDEF_CHK_NDEF_INFO *pNfcChkNdefInfo);
CLSCRF_API LONG NDEF_ReadNdef(__in SCARDHANDLE hCard, __in LPBYTE pbReadBuffer, __inout DWORD &cbReadBufferLength);
CLSCRF_API LONG NDEF_WriteNdef(__in SCARDHANDLE hCard, __in LPCBYTE pbWriteBuffer, __in DWORD cbWriteBufferLength);
CLSCRF_API LONG NDEF_FormatNdef(__in SCARDHANDLE hCard, __in DWORD cbMaxNDEFMessageSize);

#endif