//http://usb4java.org/
import org.usb4java.BufferUtils;
import org.usb4java.DeviceHandle;
import org.usb4java.LibUsb;
import org.usb4java.LibUsbException;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import java.util.Arrays;

public class testlibusb {
	static final short USB_VID = (short)0xc251;
	static final short USB_PID = (short)0x130a;
	static final byte USB_EP_WRITE = (byte)0x02;
	static final byte USB_EP_READ = (byte)0x82;
	static final int TIMEOUT = 1000;
	static final int SCR_BUFFER_SIZE = 320;
	static final int SCR_MAX_PACKET_SIZE = 64;
	static final int SCR_ALIGNED_BUFFER_SIZE =
		((SCR_BUFFER_SIZE + SCR_MAX_PACKET_SIZE - 1) & ~(SCR_MAX_PACKET_SIZE - 1));
    
    static byte Cmd_ID = 0;
    
	private static final char[] HEX_ARRAY = "0123456789ABCDEF".toCharArray();
	public static String bytesToHex(byte[] bytes) {
		char[] hexChars = new char[bytes.length * 2];
		for (int j = 0; j < bytes.length; j++) {
			int v = bytes[j] & 0xFF;
			hexChars[j * 2] = HEX_ARRAY[v >>> 4];
			hexChars[j * 2 + 1] = HEX_ARRAY[v & 0x0F];
		}
		return new String(hexChars);
	}

	private static boolean writeToUsb(DeviceHandle handle, byte[] message) {
		ByteBuffer writeBuffer = ByteBuffer.allocateDirect(message.length);   
		writeBuffer.put(message);                                             
		IntBuffer writeBufTransferred = IntBuffer.allocate(1);

		int result = LibUsb.bulkTransfer(handle, USB_EP_WRITE, writeBuffer, writeBufTransferred, TIMEOUT); 
		 
		switch (result){
		case LibUsb.SUCCESS:
			if (writeBufTransferred.get() == message.length) {
				break;
			} else {
				System.out.println("Data transfer issue [write] Requested: " + 
					message.length + "\nTransferred: " + writeBufTransferred.get());
			}
			break;
		case LibUsb.ERROR_TIMEOUT:
			return false;
		default:
			System.out.println("Data transfer issue [write] Returned: " + result);
			return false;
		}
		return true;
	}

	private static byte[] readFromUsb(DeviceHandle handle) {
		ByteBuffer readBuffer;
		readBuffer = ByteBuffer.allocateDirect(4096);
		IntBuffer readBufTransferred = IntBuffer.allocate(1);

		int result = LibUsb.bulkTransfer(handle, USB_EP_READ, readBuffer, readBufTransferred, TIMEOUT);

		switch (result) {
		case LibUsb.SUCCESS:
			int trans = readBufTransferred.get();
			byte[] receivedBytes = new byte[trans];
			readBuffer.get(receivedBytes);
			return receivedBytes;
		case LibUsb.ERROR_TIMEOUT:
			break;
		default:
			System.out.println("Data transfer issue [read] Returned: " + result);
		}
		return null;
	}

	private static int UEM_ByteStuff( byte[] TX_buf, int TX_cnt, byte b )	{
		if( b < 0xFD ) {
			TX_buf[TX_cnt++] = b;
		} else {      
			TX_buf[TX_cnt++] = (byte)0xFF;
			TX_buf[TX_cnt++] = (byte)((byte)0xFF - b);
		}
		return TX_cnt;
	}
	
	private static short UEM_CalculateFCS( short FCS, byte b ) {
		byte i;
		short w;

		w = (short)(( (int)b ^ (int)(FCS & 0x0000FFFF) ) & (int)0x000000FF);
		i = 8;
		while (i > 0) {
			if ( (int)((int)(w & 0x0000FFFF) & (int)0x00000001) == (int)0x00000001 ) {
				w = (short)(((int)(w & 0x0000FFFF) >> 1) & (int)0x0000FFFF);
				w = (short)((int)(w & 0x0000FFFF) ^ (int)0x00008408);
			} else {
				w = (short)(((int)(w & 0x0000FFFF) >> 1) & (int)0x0000FFFF);
			}
			i--;
		}
		FCS = (short)(((int)(w & 0x0000FFFF) ^ (int)( ((int)(FCS & 0x0000FFFF) >> 8) & (int)0x0000FFFF )) & (int)0x0000FFFF);
		return FCS;
	}

	public static byte[] transcieve(DeviceHandle handle, byte[] outBuffer) throws Exception {
		int res = 0;
		byte[] TX_buf = new byte[SCR_BUFFER_SIZE];
		int TX_cnt = 0;
		short TX_FCS = (short)0xFFFF;
		short RX_FCS = (short)0xFFFF;
		int i, j, total, rest, result;
		
		TX_buf[TX_cnt++] = (byte)0xFD;  // Frame begin
		TX_FCS = UEM_CalculateFCS( TX_FCS, (byte)0x00 );
		TX_cnt = UEM_ByteStuff( TX_buf, TX_cnt, (byte)0x00 );
		TX_FCS = UEM_CalculateFCS( TX_FCS, Cmd_ID );
		TX_cnt = UEM_ByteStuff( TX_buf, TX_cnt, Cmd_ID++ );
		for( i = 0; i < outBuffer.length; i++ ) {
			TX_FCS = UEM_CalculateFCS( TX_FCS, outBuffer[i] );
			TX_cnt = UEM_ByteStuff( TX_buf, TX_cnt, outBuffer[i] );
		}
		TX_FCS ^= 0xFFFF;
		TX_cnt = UEM_ByteStuff( TX_buf, TX_cnt, (byte)(TX_FCS & (short)0x00FF) );
		TX_cnt = UEM_ByteStuff( TX_buf, TX_cnt, (byte)((TX_FCS >> 8) & (short)0x00FF));
		TX_buf[TX_cnt++] = (byte)0xFE;  // Frame end
										
		TX_buf = Arrays.copyOfRange(TX_buf, 0, TX_cnt);
		System.out.println(bytesToHex(TX_buf));
		if (!writeToUsb(handle, TX_buf)) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
			
		byte[] RX_buf = readFromUsb(handle);
		if (RX_buf == null) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
		System.out.println(bytesToHex(RX_buf));

		if (RX_buf.length < 8)	{
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
		
		if (RX_buf[0] != 0xFD) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
		
		if (RX_buf[RX_buf.length-1] != 0xFE) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
		
		for (i = 1, j = 0; i < RX_buf.length - 1; i++) {
			if (RX_buf[i] == (byte)0xFF) {
				RX_buf[j++] = (byte)((byte)0xFF - RX_buf[++i]);
			} else {
				RX_buf[j++] = RX_buf[i];
			}
		}

		if (j < 6) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}

		for (i = 0; i < j - 2; i++) {
			RX_FCS = UEM_CalculateFCS(RX_FCS, RX_buf[i]);
		}
		RX_FCS ^= 0xFFFF;

		if (RX_FCS != (RX_buf[i] | (RX_buf[i+1] << 8))) {
			throw new Exception("IFD_COMMUNICATION_ERROR");
		}
		
		return Arrays.copyOfRange(RX_buf, 3, j - 2);
	}

	public static void main(String[] args) {
		int result = LibUsb.init(null);
		if (result != LibUsb.SUCCESS) {
			return;
		}

		DeviceHandle handle = LibUsb.openDeviceWithVidPid(null, USB_VID, USB_PID);
		if (handle == null) {
			return;
		}

		/*byte[] data0 = {(byte)0xFD, (byte)0x00, (byte)0x00, (byte)0x05, (byte)0x01, (byte)0xEF, (byte)0x93, (byte)0xFE};
		if (!writeToUsb(handle, data0)) {
			System.out.println("!writeToUsb 0");
			return;	
		}
		byte[] receivedBytes = readFromUsb(handle);
		if (receivedBytes != null) {
			System.out.println(bytesToHex(receivedBytes));
		}*/
		
		byte[] receivedBytes;
		try {
			receivedBytes = transcieve(handle, new byte[] {(byte)0x05, (byte)0x01});
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}

		try {
			Thread.sleep(200);
		} catch (Exception e) {
		}

		/*byte[] data1 = {(byte)0xFD, (byte)0x00, (byte)0x01, (byte)0x05, (byte)0x01, (byte)0x33, (byte)0xC9, (byte)0xFE};
		if (!writeToUsb(handle, data1)) {
			System.out.println("!writeToUsb 1");
			return;	
		}
		receivedBytes = readFromUsb(handle);
		if (receivedBytes != null) {
			System.out.println(bytesToHex(receivedBytes));
		}*/
		
		try {
			receivedBytes = transcieve(handle, new byte[] {(byte)0x05, (byte)0x01});
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}

		LibUsb.close(handle);

		LibUsb.exit(null);
	}
}
