/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * main.cpp
 * Copyright (C) 2021 MicroEM JSC <support@microem.ru>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name ``MicroEM JSC'' nor the name of any other
 *    contributor may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * CLESCAR IS PROVIDED BY MicroEM JSC ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL MicroEM JSC OR ANY OTHER CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <string.h>

#include "../libClscrfl/Clscrfl.h"
#include "../libClscrfl/Clscrfl_SAM.h"


int
main (int argc, char *argv[])
{
	void* m_pReader;
    LONG Status;
    unsigned char Buf[300];
    DWORD i, n;
    char Err;
    bool  InterfaceRS = true;
    DWORD PortIndex = 0;
//========================================================================
//  ...>Demo.exe       - äëÿ ðàáîòû ñî ñ÷èòûâàòåëåì UEM RS íà COM1
//  ...>Demo.exe RS 1  - äëÿ ðàáîòû ñî ñ÷èòûâàòåëåì UEM RS íà COM2
//  ...>Demo.exe RS 2  - äëÿ ðàáîòû ñî ñ÷èòûâàòåëåì UEM RS íà COM3 è ò.ä.
//  ...>Demo.exe USB   - äëÿ ðàáîòû ñ åäèíñòâåííûì ñ÷èòûâàòåëåì UEM USB
//  ...>Demo.exe USB 1 - äëÿ ðàáîòû ñî âòîðûì ñ÷èòûâàòåëåì UEM USB
//  ...>Demo.exe USB 2 - äëÿ ðàáîòû ñ òðåòüèì ñ÷èòûâàòåëåì UEM USB è ò.ä.
//========================================================================
    if( argc > 1 && strcmp( argv[1], "USB" ) == 0 )
            InterfaceRS = false;

    if( argc > 2 )
        PortIndex = atoi( argv[2] );

    printf("\nCONTACTLESS CARD READER's FUNCTION LIBRARY's Demo Program\n");
    printf("\n");

    Status = CLSCRF_Create( &m_pReader );

    if( Status == 0 )
        printf("Interface is created successfully\n");
    else
        printf("Interface is not created\n");

    if( InterfaceRS )
        Status = CLSCRF_OpenRS( m_pReader, PortIndex, 9600, 0 );
    else
        Status = CLSCRF_OpenUSB( m_pReader, PortIndex, 0 );

    if( Status == 0 )
    {
        Status = CLSCRF_Sound( m_pReader, 2 );
        if( InterfaceRS )
            printf("Interface COM%d is opened successfully\n", PortIndex+1);
        else
            printf("Interface USB(%d) is opened successfully\n", PortIndex);
    }
    else if( InterfaceRS )
        printf("Interface COM%d is not opened\n", PortIndex+1);
    else
        printf("Interface USB(%d) is not opened\n", PortIndex);

//    if( InterfaceRS )
    {
        Status = CLSCRF_Mfrc_On( m_pReader );

        if( Status == 0 )
            printf("Reader-IC is turn on successfully\n");
        else
            printf("Reader-IC is not turn on\n");
    }

    printf("Mfrc_Version        ");
    Status = CLSCRF_Get_Mfrc_Version( m_pReader, Buf );
    if( Status == 0 )
        for( i = 0; i < 6; i++ )
            printf("%02X", Buf[i] );
    else
        printf("Sorry, error 0x%08lX", Status );
    printf("\n");

    printf("Mfrc_Serial_Number  ");
    Status = CLSCRF_Get_Mfrc_Serial_Number( m_pReader, Buf );
    if( Status == 0 )
        for( i = 0; i < 4; i++ )
            printf("%02X", Buf[3-i] );
    else
        printf("Sorry, error 0x%08lX", Status );
    printf("\n");

    //Sleep( 1000 );

    CLSCRF_Mfrc_Rf_Off_On( m_pReader, 1 );

    printf("Activate_Idle_A\n");
    Status = CLSCRF_Activate_Idle_A( m_pReader,
                                     &Buf[0],// pbATQ 2 bytes
                                     &Buf[2],// pbSAK 1 byte
                                     &Buf[3],// pbUID max 10 bytes
                                     &n );
    if( Status == 0 )
    {
        CLSCRF_Sound( m_pReader, 1 );
        printf("ATQ   %02X%02X\n", Buf[1], Buf[0] );
        printf("SAK   %02X\n", Buf[2] );
        printf("UID   " );
        for( i = 0; i < n; i++ )
            printf("%02X", Buf[i+3] );
        printf("\n");
        CLSCRF_Halt_A( m_pReader );
    }
    else
    {
        printf("Sorry, error 0x%08lX\n", Status );
        if( Status == 0x80100001 )
        {
            CLSCRF_GetLastInternalError( m_pReader, (BYTE*)&Err );
            printf("Internal Error = %d\n", Err );
        }
    }

//    if( InterfaceRS )
    {
    Status = CLSCRF_Mfrc_Off( m_pReader );

    if( Status == 0 )
        printf("Reader-IC is turn off successfully\n");
    else
        printf("Reader-IC is not turn off\n");
    }

    Status = CLSCRF_Close( m_pReader );

    if( Status == 0 )
        printf("Interface is closed successfully\n");
    else
        printf("Interface is not closed\n");

    Status = CLSCRF_Destroy( &m_pReader );

    if( Status == 0 )
        printf("Interface is destroyed successfully\n");
    else
        printf("Interface is not destroyed\n");

    printf("\n");
    printf( "Press Enter key ..." );
    getchar();

	return 0;
}

