package ru.microem.uembletest;

import android.Manifest;
import android.app.Activity;
import android.bluetooth.le.ScanFilter;
import android.bluetooth.le.ScanSettings;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Build;
import android.os.Looper;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothManager;
import android.bluetooth.BluetoothProfile;
import android.bluetooth.le.BluetoothLeScanner;
import android.bluetooth.le.ScanCallback;
import android.bluetooth.le.ScanResult;
import android.util.Log;
import android.view.View;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.Vector;

import static android.bluetooth.BluetoothDevice.PHY_LE_1M_MASK;
import static android.bluetooth.BluetoothDevice.PHY_LE_CODED_MASK;
import static android.bluetooth.BluetoothDevice.TRANSPORT_LE;

public class MainActivity extends AppCompatActivity implements AdapterView.OnItemSelectedListener {

    private static final String TAG = "123";
    private static Activity _MainActivity = null;
    // Bluetooth's variables
    BluetoothAdapter bluetoothAdapter;
    BluetoothLeScanner bluetoothLeScanner;
    BluetoothManager bluetoothManager;
    BluetoothScanCallback  bluetoothScanCallback;
    BluetoothGatt gattClient;

    BluetoothGattCharacteristic characteristicFIFO; // To get Value
    BluetoothGattCharacteristic characteristicCREDITS; // To get Value

    int finalMTU = 0;

    // UUID's (set yours)
    final UUID SERVICE_UUID_SERIAL = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d701");//UUID.fromString("ab0828b1-198e-4351-b779-901fa0e0371e");
    final UUID CHARACTERISTIC_UUID_FIFO = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d703");//UUID.fromString("1a220d0a-6b06-4767-8692-243153d94d85");
    final UUID CHARACTERISTIC_UUID_CREDITS = UUID.fromString("2456e1b9-26e2-8f83-e744-f34f01e9d704");//UUID.fromString("1a220d0a-6b06-4767-8692-243153d94d85");
    //final UUID DESCRIPTOR_UUID_ID = UUID.fromString("ec6e1003-884b-4a1c-850f-1cfce9cf6567");

    public void checkBTPermissions(){
        if(Build.VERSION.SDK_INT > Build.VERSION_CODES.LOLLIPOP){
            int permissionCheck = this.checkSelfPermission("Manifest.permission.ACCESS_FINE_LOCATION");
            permissionCheck += this.checkSelfPermission("Manifest.permission.ACCESS_COARSE_LOCATION");
            if (permissionCheck != 0) {

                this.requestPermissions(new String[]{Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION}, 1001); //Any number
            }
        }else{
            Log.d(TAG, "checkBTPermissions: No need to check permissions. SDK version < LOLLIPOP.");
        }
    }

    public static byte[] hexStringToByteArray(String s) {
        int len = s.length();
        byte[] data = new byte[len/2];

        for(int i = 0; i < len; i+=2){
            data[i/2] = (byte) ((Character.digit(s.charAt(i), 16) << 4) + Character.digit(s.charAt(i+1), 16));
        }

        return data;
    }

    final protected static char[] hexArray = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
    public static String byteArrayToHexString(byte[] bytes) {
        char[] hexChars = new char[bytes.length*2];
        int v;

        for(int j=0; j < bytes.length; j++) {
            v = bytes[j] & 0xFF;
            hexChars[j*2] = hexArray[v>>>4];
            hexChars[j*2 + 1] = hexArray[v & 0x0F];
        }

        return new String(hexChars);
    }

    private static final char[] HEX_ARRAY = "0123456789ABCDEF".toCharArray();
    public static String bytesToHex(byte[] bytes) {
        char[] hexChars = new char[bytes.length * 2];
        for (int j = 0; j < bytes.length; j++) {
            int v = bytes[j] & 0xFF;
            hexChars[j * 2] = HEX_ARRAY[v >>> 4];
            hexChars[j * 2 + 1] = HEX_ARRAY[v & 0x0F];
        }
        return new String(hexChars);
    }

    private static byte[] CRC16(byte[] buf, int start, int count)
    {
        //int len = buf.length;
        int crc = 0xFFFF;

        for (int pos = start; pos < start + count; pos++) {
            crc ^= (int)buf[pos] & 0xFF;   // XOR byte into least sig. byte of crc

            for (int i = 8; i != 0; i--) {    // Loop over each bit
                if ((crc & 0x0001) != 0) {      // If the LSB is set
                    crc >>= 1;                    // Shift right and XOR 0x8408
                    crc ^= 0x8408;
                }
                else                            // Else LSB is not set
                    crc >>= 1;                    // Just shift right
            }
        }

        return new byte[]{(byte)((crc^0xFFFF) & 0xFF), (byte)(((crc^0xFFFF) >> 8) & 0xFF)};
    }

    private static byte[] byteStuff(byte[] data) {
        ByteArrayOutputStream stuffeddata = new ByteArrayOutputStream();
        for (byte databyte : data) {
            if (databyte < 0xFD) {
                stuffeddata.write(databyte);
            } else {
                stuffeddata.write(0xFF);
                stuffeddata.write(0xFF - databyte);
            }
        }
        return stuffeddata.toByteArray();
    }

    private static byte[] unbyteStuff(byte[] rawdata) {
        ByteArrayOutputStream unstuffeddata = new ByteArrayOutputStream();
        boolean fInvertNext = false;
        for (byte databyte : rawdata) {
            if (databyte == 0xFF) {
                fInvertNext = true;
                continue;
            }
            if (fInvertNext) {
                unstuffeddata.write(0xFF - databyte);
                fInvertNext = false;
            } else {
                unstuffeddata.write(databyte);
            }
        }
        return unstuffeddata.toByteArray();
    }

    private static byte nCommand = 0x00;
    public static byte[] wrapCommand(byte[] data) {
        ByteArrayOutputStream rawdata = new ByteArrayOutputStream();
        //rawdata.write(0xFD);
        rawdata.write(0x00);
        rawdata.write(nCommand++);
        //byte[] tmpData = byteStuff(new byte[] {0x00, nCommand++});
        //rawdata.write(tmpData, 0, tmpData.length);

        rawdata.write(data, 0, data.length);
        //tmpData = byteStuff(data);
        //rawdata.write(tmpData, 0, tmpData.length);

        byte[] FCS = CRC16(rawdata.toByteArray(), 0, rawdata.size());
        rawdata.write(FCS, 0, FCS.length);
        //tmpData = byteStuff(FCS);
        //rawdata.write(tmpData, 0, tmpData.length);

        byte[] tmpData = byteStuff(rawdata.toByteArray());
        rawdata = new ByteArrayOutputStream();
        rawdata.write(0xFD);
        rawdata.write(tmpData, 0, tmpData.length);
        rawdata.write(0xFE);

        return rawdata.toByteArray();
    }

    public static byte[] unwrapResponse(byte[] rawdata){
        rawdata = unbyteStuff(rawdata);
        assert (rawdata[0] == 0xFD);
        assert (rawdata[rawdata.length - 1] == 0xFE);
        byte[] FCS = CRC16(rawdata, 1, rawdata.length - 4);
        assert(FCS[0] == rawdata[rawdata.length - 3]);
        assert(FCS[1] == rawdata[rawdata.length - 2]);

        return Arrays.copyOfRange(rawdata, 3, rawdata.length-3);
    }

    public ArrayList<String> _listBLE = new ArrayList<String>();

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        _MainActivity = this;
        // onResume()
        checkBTPermissions();
        /*if (ContextCompat.checkSelfPermission(this.getApplicationContext(),
                android.Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_GRANTED) {
        } else {
            //ActivityCompat.requestPermissions(this, new String[]{android.Manifest.permission.ACCESS_FINE_LOCATION},
            //        101);
            ActivityCompat.requestPermissions(this,new String[]{Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION}, 1001); //Any number
        }*/
        // Bluetooth

        Spinner spinner = (Spinner) findViewById(R.id.cb_command);
        spinner.setOnItemSelectedListener(this);

        final ListView listBLE = findViewById(R.id.bleList);

        ArrayAdapter<String> adapter = new ArrayAdapter<String>(_MainActivity,
                android.R.layout.simple_list_item_1, android.R.id.text1, _listBLE);
        listBLE.setAdapter(adapter);
        listBLE.setOnItemClickListener(new AdapterView.OnItemClickListener() {

            @Override
            public void onItemClick(AdapterView<?> parent, View view,
                                    int position, long id) {

                // ListView Clicked item index
                int itemPosition     = position;

                // ListView Clicked item value
                String  itemValue    = (String) listBLE.getItemAtPosition(position);

                // Show Alert
                Toast.makeText(getApplicationContext(),
                        "Connecting " +itemValue+ " ..." , Toast.LENGTH_LONG)
                        .show();


                /*for (Map.Entry<String, BluetoothDevice> e : _bleDevices.entrySet()) {
                    //A key    = e.getKey();
                    //B value  = e.getValue();
                    //e.getValue().
                    gattClient.disconnect();
                }*/
                //if (gattClient != null)
                //    gattClient.disconnect();
                connectDevice(_bleDevices.get(itemValue));
            }

        });

        bluetoothManager = (BluetoothManager) getSystemService(BLUETOOTH_SERVICE);
        bluetoothAdapter = BluetoothAdapter.getDefaultAdapter();//bluetoothManager.getAdapter();


        if (!bluetoothAdapter.isEnabled())
        {
            Intent enableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            //startActivityForResult(enableIntent, REQUEST_ENABLE_BLUETOOTH);
            //Intent discoverableIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_DISCOVERABLE);
            //discoverableIntent.putExtra(BluetoothAdapter.EXTRA_DISCOVERABLE_DURATION, 60);
            //startActivity(discoverableIntent);
            startActivityForResult(enableIntent, 0);
        }

        startScan();
    }

    // BLUETOOTH SCAN
    private void startScan(){
        Log.i(TAG,"startScan()");
        bluetoothScanCallback = new BluetoothScanCallback();
        //BluetoothAdapter.getDefaultAdapter().disable();
        //BluetoothAdapter.getDefaultAdapter().enable();
        IntentFilter filter = new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED);
        BroadcastReceiver broadcastReceiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                Object action = intent.getAction();
                if (action == BluetoothAdapter.ACTION_STATE_CHANGED) {
                    int state = intent.getIntExtra(BluetoothAdapter.EXTRA_STATE, BluetoothAdapter.ERROR);
                    switch (state) {
                        case BluetoothAdapter.STATE_ON: {
                            if (bluetoothAdapter.isEnabled()) {
                                bluetoothLeScanner = bluetoothAdapter.getBluetoothLeScanner();

                                ScanFilter.Builder builder = new ScanFilter.Builder();
                                /*switch(scanType) {
                                    case SCAN_BY_NAME:
                                        builder.setDeviceName(data);
                                        break;
                                }*/
                                Vector<ScanFilter> filter = new Vector<ScanFilter>();
                                filter.add(builder.build());
                                ScanSettings.Builder builderScanSettings = new ScanSettings.Builder();
                                //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_BALANCED);
                                //builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_POWER);
                                builderScanSettings.setScanMode(ScanSettings.SCAN_MODE_LOW_LATENCY);
                                builderScanSettings.setReportDelay(0);
                                bluetoothLeScanner.startScan(filter, builderScanSettings.build(),bluetoothScanCallback);
                                //bluetoothLeScanner.startScan(bluetoothScanCallback);
                            }
                        }
                        case BluetoothAdapter.STATE_OFF: {
                            bluetoothAdapter.enable();
                        }
                    }
                }
            }
        };
        registerReceiver(broadcastReceiver, filter);
        bluetoothAdapter.disable();
        //bluetoothAdapter.enable();
    }

    // BLUETOOTH CONNECTION
    private void connectDevice(BluetoothDevice device) {
        if (device == null) Log.i(TAG,"Device is null");
        GattClientCallback gattClientCallback = new GattClientCallback();
        //gattClient = device.connectGatt(this,false,gattClientCallback);

        if (gattClient != null && bluetoothManager.getConnectedDevices(BluetoothProfile.GATT).size() > 0)
            return;

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            gattClient = device.connectGatt(this, false, gattClientCallback, TRANSPORT_LE,PHY_LE_CODED_MASK ^ PHY_LE_1M_MASK);
        } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            gattClient = device.connectGatt(this, false, gattClientCallback, TRANSPORT_LE);
        } else {
            gattClient = device.connectGatt(this, false, gattClientCallback);
        }
    }

    public HashMap<String, BluetoothDevice> _bleDevices = new HashMap<String, BluetoothDevice>();

    // BLE Scan Callbacks
    private class BluetoothScanCallback extends ScanCallback {

        @Override
        public void onScanResult(int callbackType, ScanResult result) {
            Log.i(TAG, "onScanResult");
            if (result.getDevice().getName() != null){//ANNA-B1-F08B84
                if (result.getDevice().getName().contains("ANNA-") || result.getDevice().getName().contains("uEM Reader")) {
                    if(!_bleDevices.containsKey(result.getDevice().getName())) {
                        _bleDevices.put(result.getDevice().getName(), result.getDevice());
                        ListView listBLE = findViewById(R.id.bleList);
                        _listBLE.add(result.getDevice().getName());
                        ((ArrayAdapter<String>) listBLE.getAdapter()).notifyDataSetChanged();
                    }
                    // When find your device, connect.
                    connectDevice(result.getDevice());
                    //bluetoothLeScanner.stopScan(bluetoothScanCallback); // stop scan
                }
            }
        }

        @Override
        public void onBatchScanResults(List<ScanResult> results) {
            Log.i(TAG, "onBathScanResults");

            bluetoothLeScanner.startScan(bluetoothScanCallback);
        }

        @Override
        public void onScanFailed(int errorCode) {
            Log.i(TAG, "ErrorCode: " + errorCode);
        }
    }

    // Bluetooth GATT Client Callback
    private class GattClientCallback extends BluetoothGattCallback {
        @Override
        public void onMtuChanged(BluetoothGatt gatt, int mtu, int status) {
            super.onMtuChanged(gatt, mtu, status);
            finalMTU = mtu;
            gatt.discoverServices();
        }

        @Override
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            super.onConnectionStateChange(gatt, status, newState);
            Log.i(TAG,"onConnectionStateChange");

            if (status == BluetoothGatt.GATT_FAILURE) {
                Log.i(TAG, "onConnectionStateChange GATT FAILURE");
                return;
            } else if (status != BluetoothGatt.GATT_SUCCESS) {
                (MainActivity._MainActivity).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(getApplicationContext(),
                                "Failed to change connection state...", Toast.LENGTH_LONG)
                                .show();
                    }});
                Log.i(TAG, "onConnectionStateChange != GATT_SUCCESS");
                return;
            }

            if (newState == BluetoothProfile.STATE_CONNECTED) {
                Log.i(TAG, "onConnectionStateChange CONNECTED");
                (MainActivity._MainActivity).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(getApplicationContext(),
                                "BLE Device connected!", Toast.LENGTH_LONG)
                                .show();
                    }});
                gatt.requestMtu(512);
            } else if (newState == BluetoothProfile.STATE_DISCONNECTED) {
                Log.i(TAG, "onConnectionStateChange DISCONNECTED");
                (MainActivity._MainActivity).runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        Toast.makeText(getApplicationContext(),
                                "BLE Device disconnected!", Toast.LENGTH_LONG)
                                .show();
                    }});
            }
        }

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            //final BluetoothGatt g = gatt;
            super.onServicesDiscovered(gatt, status);
            Log.i(TAG,"onServicesDiscovered");
            if (status != BluetoothGatt.GATT_SUCCESS) return;

            // Reference your UUIDs
            characteristicFIFO = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_FIFO);
            characteristicCREDITS = gatt.getService(SERVICE_UUID_SERIAL).getCharacteristic(CHARACTERISTIC_UUID_CREDITS);
            //gatt.setCharacteristicNotification(characteristicFIFO,true);

            //BluetoothGattDescriptor descriptor = characteristicFIFO.getDescriptor(DESCRIPTOR_UUID_ID);
            //descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
            //gatt.writeDescriptor(descriptor);

            //characteristicCREDITS.setValue(new byte[]{0x33});
            //gatt.writeCharacteristic(characteristicCREDITS);

            //gatt.connect();
            //gatt.setCharacteristicNotification(characteristicFIFO, true);



            String UUID_MASK = "0000%s-0000-1000-8000-00805f9b34fb";
            String CLIENT_CHARACTERISTIC_CONFIG = String.format(UUID_MASK, "2902");

            if ((characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_NOTIFY) > 0) {
                BluetoothGattDescriptor descriptor = characteristicFIFO.getDescriptor(UUID.fromString(CLIENT_CHARACTERISTIC_CONFIG));
                if (descriptor == null || gatt == null) return;

                // Set write type to WRITE_TYPE_DEFAULT to fix an issue seen on some phone models and
                // Android versions that enable notification with a write_command message instead of a
                // write_request message (as specified by the Bluetooth specification).
                characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT);
                descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
                gatt.setCharacteristicNotification(characteristicFIFO, true);
                gatt.writeDescriptor(descriptor);
            }

            //gatt.
            //characteristicFIFO.setValue(new byte[]{0x33});
            //gatt.writeCharacteristic(characteristicFIFO);

        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            super.onCharacteristicRead(gatt, characteristic, status);
            Log.i(TAG,"onCharacteristicRead");
        }

        @Override
        public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            super.onCharacteristicWrite(gatt, characteristic, status);
            Log.i(TAG,"onCharacteristicWrite");
        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            super.onCharacteristicChanged(gatt, characteristic);
            Log.i(TAG,"onCharacteristicChanged");
            // Here you can read the characteristc's value
            final BluetoothGattCharacteristic ch = characteristic;
            (MainActivity._MainActivity).runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    TextView textView = (TextView) findViewById(R.id.outtext);
                    String out = new String(ch.getValue());
                    try {
                        out = bytesToHex(unwrapResponse(ch.getValue()));
                    } catch (Exception e) {
                        System.out.println("Error " + e.getMessage());
                        return;
                    }
                    textView.setText(out);
                }
            });
        }

        @Override
        public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            super.onDescriptorRead(gatt, descriptor, status);
            Log.i(TAG,"onDescriptorRead");
        }

        @Override
        public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            super.onDescriptorWrite(gatt, descriptor, status);
            Log.i(TAG,"onDescriptorWrite");
            if ((characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) > 0) {
                //byte[] test = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39};
                byte[] test = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39};
                //byte[] testBeep = {(byte)0xFD, 0x00, 0x02, 0x05, 0x01, 0x57, 0x26, (byte)0xFE};//   (Beep)
                //byte[] crc = CRC16(new byte[] {0x00, 0x02, 0x05, 0x01}, 4);
                byte[] testBeep = wrapCommand(new byte[] {0x05, 0x01});
                //byte[] testBlinkRed = {(byte)0xFD, 0x00, 0x40, 0x07, 0x01, 0x01, 0x00, (byte)0x88, 0x22, (byte)0xFE};//   (Blink Red)
                byte[] testBlinkRed = wrapCommand(new byte[] {0x07, 0x01, 0x01, 0x00});//   (Blink Red)
                //byte[] testGetVersion = {(byte)0xFD, 0x00, 0x25, 0x64, 0x65, (byte)0xBE, (byte)0xFE};//   (Get Version)
                byte[] testGetVersion = wrapCommand(new byte[] {0x64});//   (Get Version)
                //byte[] testBeep2 = {(byte)0xFD, 0x00, 0x19, 0x05, 0x01, 0x64, (byte)0x8A, (byte)0xFE};//   (Beep)
                byte[] testBeep2 = wrapCommand(new byte[] {0x05, 0x01});//   (Beep)
                //byte[] testBlinkGreen = {(byte)0xFD, 0x00, 0x42, 0x07, 0x02, 0x01, 0x00, 0x64, (byte)0xDB, (byte)0xFE};//   (Blink Green)
                byte[] testBlinkGreen = wrapCommand(new byte[] {0x07, 0x02, 0x01, 0x00});//   (Blink Green)
                byte[] testISO15693 = wrapCommand(new byte[] {0x75, 0x40, 0x55, 0x00, 0x00});// Complex ISO15693 search
                if (gatt == null) return;
                characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
                characteristicFIFO.setValue(testISO15693);
                gatt.writeCharacteristic(characteristicFIFO);
            }
        }

    }

    byte[][] arrCommands = new byte[][]{
        new byte[] {},
        new byte[] {0x05, 0x01},
        new byte[] {0x07, 0x01, 0x01, 0x00},
        new byte[] {0x64},
        new byte[] {0x05, 0x02},
        new byte[] {0x07, 0x02, 0x01, 0x00},
        new byte[] {0x10},
        new byte[] {0x51, 0x40},
        new byte[] {0x30, 0x00, 0x00, 0x00},
        new byte[] {0x04, (byte)0x80, 0x01},
        new byte[] {0x75, 0x40, 0x55, 0x00, 0x00}
    };

    public void onItemSelected(AdapterView<?> parent, View view,
                               int pos, long id) {
        // parent.getItemAtPosition(pos)
        if (pos > 0 && gattClient != null && characteristicFIFO != null && (characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) > 0) {
            characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
            characteristicFIFO.setValue(wrapCommand(arrCommands[pos]));
            gattClient.writeCharacteristic(characteristicFIFO);
            TextView textView = (TextView) findViewById(R.id.customCmd);
            String cmd;
            try {
                cmd = bytesToHex(arrCommands[pos]);
            } catch (Exception e) {
                System.out.println("Error " + e.getMessage());
                return;
            }
            textView.setText(cmd);
        }
    }

    public void onNothingSelected(AdapterView<?> parent) {
        // Another interface callback
    }

    /** Called when the user touches the button */
    public void onSendCommand(View view) {
        EditText textView = (EditText) findViewById(R.id.customCmd);
        if (gattClient != null && characteristicFIFO != null && (characteristicFIFO.getProperties() & BluetoothGattCharacteristic.PROPERTY_WRITE) > 0) {
            characteristicFIFO.setWriteType(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
            String s = (String)textView.getText().toString();
            characteristicFIFO.setValue(wrapCommand(hexStringToByteArray(s)));
            gattClient.writeCharacteristic(characteristicFIFO);
        }
    }

    public void onDisconnect(View view) {
        if (gattClient != null)
            gattClient.disconnect();
    }
}
